/* comm/admin.c */
/* Administrative commands */

#include "externs.h"


/* find flags */
static int convert_flags(player, wiz, s, mask, type)
dbref player;
int wiz, *mask, *type;
char *s;
{
  int i, last_id=' ';
  
  for(;*s;s++) {
    if(*s == 'c' && !wiz)
      continue;

    for(i=0;flaglist[i].flag;i++)
      if(flaglist[i].flag == *s) {
        if(flaglist[i].objtype != NOTYPE) {
          /* check that we're not mixing up object types */
          if(*type != NOTYPE && *type != flaglist[i].objtype) {
            notify(player, tprintf("Flag '%c' conflicts with '%c'.",
                   last_id, flaglist[i].flag));
            return 0;
          }

          *type=flaglist[i].objtype;
          last_id=*s;
        }

        /* add flag to result bitmask */
        *mask|=flaglist[i].bits;
        break;
      }
    
    if(!flaglist[i].flag) {
      notify(player, tprintf("Unknown flag '%c'.", *s));
      return 0;
    }
  }
  
  return 1;
}

/* finds objects which match specific categories in the database */
void do_search(player, arg1, arg3)
dbref player;
char *arg1, *arg3;
{
  char buf[4096], *arg2, *string="";
  int owner, category=0, type=NOTYPE, value=0;
  int i, thing, flag=1, count=0;

  /* available categories */
  enum { SEARCH_ATTR=1, SEARCH_CLASS, SEARCH_FLAGS, SEARCH_LINKS,
         SEARCH_INZONE };

  if(!charge_money(player, SEARCH_COST, 1))
    return;
  
  /* determine arguments */
  if((arg2=strchr(arg1, ' ')))   /* @search <arg1> <arg2>=<arg3> */
    *arg2++='\0';
  else if(!*arg3)                /* @search <arg1> */
    arg2="";
  else {                         /* @search <arg2>=<arg3> */
    arg2=arg1;
    arg1="";
  }
  
  /* search a specific player? */
  if(*arg1) {
    if((owner=lookup_player(arg1)) == AMBIGUOUS)
      owner=db[player].owner;
    else if(owner == NOTHING) {
      notify(player, tprintf("No such player '%s'.", arg1));
      return;
    } else if(!controls(player, owner, POW_STATS)) {
      notify(player, "You need a search warrant to do that.");
      return;
    }
  } else  /* default restriction */
    owner=power(player, POW_STATS)?AMBIGUOUS:db[player].owner;

  /* check category restrictions */

  if(!*arg2) {
    /* No Class Restrictions */
  } else if(string_prefix("attributes", arg2)) {
    /* Matches <attr>:<value> */

    category=SEARCH_ATTR;
    if((string=strchr(arg3, ':')))
      *string++='\0';
    else
      string="";
  } else if(string_prefix("class", arg2)) {
    /* Matches Class Name */

    category=SEARCH_CLASS;
    type=TYPE_PLAYER;
    for(value=0;value<NUM_CLASSES;value++)
      if(!strcasecmp(arg3, classnames[value]))
        break;
    if(value == NUM_CLASSES) {
      notify(player, "No such class name.");
      return;
    }
  } else if(string_prefix("flags", arg2)) {
    /* Matches Flags */

    category=SEARCH_FLAGS;
    if(!convert_flags(player, power(player, POW_WHO), arg3,
       &value, &type))
      return;
  } else if(string_prefix("inzone", arg2)) {
    /* Matches Flags */

    category=SEARCH_INZONE;
    type=TYPE_ROOM;
    if((value=match_thing(player, arg3, MATCH_EVERYTHING)) == NOTHING)
      return;
  } else if(string_prefix("links", arg2)) {
    /* Matches Links to <object> */

    category=SEARCH_LINKS;
    if((value=match_thing(player, arg3, MATCH_EVERYTHING)) == NOTHING)
      return;
  } else if(string_prefix("name", arg2)) {
    /* Matches Name of any object */
  } else {
    /* Matches Type Category */

    if(string_prefix("type", arg2)) {
      arg2=arg3;
      arg3="";
    }

    for(type=0;type < 8;type++)
      if(!strcasecmp(arg2, typenames[type]))
        break;
    if(type == 8) {
      notify(player, "Unknown type restriction.");
      return;
    }
  }

  /* main search loop */

  /* check through all valid object types */
  for(i=0;i<5;i++) {
    if(type != NOTYPE && type != i)
      continue;
    if(!flag)
      flag=2;

    for(thing=0;thing<db_top;thing++) {
      if(Typeof(thing) != i || (owner != AMBIGUOUS && db[thing].owner != owner))
        continue;

      /* check value restrictions */
      switch(category) {
      case SEARCH_ATTR:
        if(!match_attr(player, thing, arg3, 1, string)) continue;
        break;
      case SEARCH_CLASS:
        if(class(thing) != value) continue;
        break;
      case SEARCH_FLAGS:
        if(value && !(db[thing].flags & value)) continue;
        break;
      case SEARCH_INZONE:
        if(!inlist(db[thing].zone, value)) continue;
        break;
      case SEARCH_LINKS:
        if(db[thing].link != value) continue;
        break;
      default:
        if(*arg3 && !string_match(db[thing].name, arg3)) continue;
      }

      /* print type header */
      if(flag) {
        if(flag == 2)
          notify(player, "--");
        notify(player, tprintf("%s%s:", typenames[i],
	       (i == TYPE_GARBAGE)?"":"s"));
        notify(player, "");
        flag=0;
      }

      strcpy(buf, unparse_object(player, thing));

      if(controls(player, thing, POW_EXAMINE)) {
        if(i == TYPE_EXIT) {
          sprintf(buf+strlen(buf), "  [from %s to ", unparse_object(player,
                  db[thing].location));
          sprintf(buf+strlen(buf), "%s]", unparse_object(player,
                 db[thing].link));
        } else if(i == TYPE_PLAYER || (i == TYPE_THING && owner != AMBIGUOUS))
          sprintf(buf+strlen(buf), "  [location: %s]", unparse_object(player,
                  db[thing].location));
        else if(owner == AMBIGUOUS)
            sprintf(buf+strlen(buf), "  [owner: %s]", unparse_object(player,
                    db[thing].owner));
      }

      notify(player, buf);
      count++;
    }
  }

  if(count)
    notify(player, tprintf("%d object%s found.", count, count==1?"":"s"));
  else
    notify(player, "Nothing found.");
}

/* find owned objects in database */
void do_find(player, name)
dbref player;
char *name;
{
  char buf[2400];
  int i, count=0;

  /* check money */
  if(!charge_money(player, FIND_COST, 1))
    return;

  /* search database */
  for(i=0;i<db_top;i++)
    if(Typeof(i) != TYPE_EXIT && Typeof(i) != TYPE_GARBAGE && controls(player,
       i, POW_STATS) && (!*name || string_match(db[i].name, name))) {
      strcpy(buf, unparse_object(player, i));
      if(Typeof(i) == TYPE_THING && db[i].location != i)
        sprintf(buf+strlen(buf), "  [location: %s]", unparse_object(player,
               db[i].location));
      notify(player, buf);
      ++count;
    }

  if(!count)
    notify(player, "Nothing found.");
  else
    notify(player, tprintf("%d object%s found.", count, count==1?"":"s"));
}

/* @stats database checking command */
void do_stats(player, name)
dbref player;
char *name;
{
  int a, b, c, i, total, obj[8], pla[NUM_CLASSES];
  int count, owner=AMBIGUOUS, cols=get_cols(NULL, player)-26;
  char buf[512], buf2[32];
  
  /* find player to stat on */
  if(*name && (owner = lookup_player(name)) == AMBIGUOUS)
    owner=db[player].owner;
  else if(owner == NOTHING) {
    notify(player, tprintf("No such player '%s'.", name));
    return;
  } if(owner != AMBIGUOUS && !controls(player, owner, POW_STATS)) {
    notify(player, "You need a search warrant to do that.");
    return;
  }
  
  /* calculate the number of objects the player (or database) has */
  calc_stats(owner, &total, obj, pla);

  /* lets print it out */
  notify(player, tprintf("\e[1;32m%s Database Breakdown:\e[m",
         owner==AMBIGUOUS?MUD_NAME:db[owner].name));
  i=(owner==AMBIGUOUS)?db_top:total;
  notify(player, tprintf("\e[1m%7d \e[36mTotal Object%s\e[m", i, i==1?"":"s"));
  if(owner == AMBIGUOUS)
    notify(player, tprintf("\e[1m%7d \e[36mFully Active\e[m", total));

  for(i=0;i<8;i++)
    if(obj[i] > 0 && (i != TYPE_PLAYER || obj[i] > 1))
      notify(player, tprintf("\e[1m%7d \e[36m%s%s\e[m", obj[i],
             typenames[i], (i == TYPE_GARBAGE || obj[i]==1)?"":"s"));
  if(owner != AMBIGUOUS || (!VISIBLE_ADMIN && !Immortal(player)))
    return;

  notify(player, "\e[1;34m\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\e[m");
  for(i=0;i<NUM_CLASSES;i++) {
    if(!pla[i])
      continue;
    if(!i || pla[i] > cols/2) {
      notify(player, tprintf("\e[1m%7d \e[36m%s%s\e[m", pla[i],
             classnames[i], pla[i]==1?"":"s"));
      continue;
    }
    count=0;
    *buf=0;
    for(c=0;c<db_top;c++)
      if(Typeof(c) == TYPE_PLAYER && class(c) == i) {
        if(*buf)
          strcat(buf, ", ");
        if(strlen(buf)+strlen(db[c].name) > cols) {
          buf[strlen(buf)-1]=0;
          if(count++)
            notify(player, tprintf("%22s\e[1;33m%s\e[m", "", buf));
          else {
            sprintf(buf2, "%s%s", classnames[i], pla[i]==1?"":"s");
            b=13-strlen(buf2);
            strcat(buf2, "\e[34m");
            for(a=0;a<b;a++)
              strcat(buf2, "\371");
            notify(player, tprintf("\e[1m%7d \e[36m%s \e[33m%s\e[m", pla[i],
                   buf2, buf));
          }
          *buf=0;
        }
        strcat(buf, db[c].name);
      }

    if(count)
      notify(player, tprintf("%22s\e[1;33m%s\e[m", "", buf));
    else {
      sprintf(buf2, "%s%s", classnames[i], pla[i]==1?"":"s");
      c=13-strlen(buf2);
      strcat(buf2, "\e[34m");
      for(a=0;a<c;a++)
        strcat(buf2, "\371");
      notify(player, tprintf("\e[1m%7d \e[36m%s \e[33m%s\e[m", pla[i],
             buf2, buf));
    }
  }
}

void calc_stats(owner, total, obj, pla)
dbref owner;
int *total;
int obj[8], pla[NUM_CLASSES];
{
  dbref thing, i;
  
  /* zero out count stats */
  *total=0;
  for(i=0;i<8;i++)
    obj[i]=0;
  for(i=0;i<NUM_CLASSES;i++)
    pla[i]=0;

  /* find objects */
  for(thing=0;thing<db_top;thing++)
    if(owner == AMBIGUOUS || owner == db[thing].owner) {
      ++obj[Typeof(thing)];
      if(Typeof(thing) == TYPE_PLAYER)
        ++pla[class(thing)];
      if(Typeof(thing) != TYPE_GARBAGE)
        ++*total;
    }
}

/* delete all objects owned by player */
void do_wipeout(player, who)
dbref player;
char *who;
{
  int a, thing=lookup_player(who);
  
  /* find player to destroy stuff on */
  if(thing <= NOTHING) {
    notify(player, tprintf("No such player '%s'.", who));
    return;
  } if(!controls(player,thing,POW_MODIFY) || !controls(player,thing,POW_NUKE)) {
    notify(player, "Permission denied.");
    return;
  } if(class(thing) >= class(db[player].owner)) {
    notify(player, "You may only wipeout players of lower rank.");
    return;
  } if(thing == GOD) {
    notify(player, "God cannot be wiped out.");
    return;
  }

  for(a=2;a<db_top;a++)
    if(db[a].owner == thing && a != thing)
      recycle(a);
  validate_links();

  notify(player, tprintf("Destroyed everything %s owned.", db[thing].name));
}

/* command to change ownership with everything from <from> to <rcpt> */
void do_chownall(player, arg1, arg2)
dbref player;
char *arg1, *arg2;
{
  dbref from, rcpt, a;
  
  if(!*arg1 || !*arg2) {
    notify(player, "Usage: @chownall <from player>=<to player>");
    return;
  }

  /* find players */
  if((from = lookup_player(arg1)) == AMBIGUOUS)
    from=db[player].owner;
  if(from == NOTHING) {
    notify(player, tprintf("No such player '%s'.", arg1));
    return;
  }
  if((rcpt = lookup_player(arg2)) == AMBIGUOUS)
    rcpt=db[player].owner;
  if(rcpt == NOTHING) {
    notify(player, tprintf("No such destination player '%s'.", arg2));
    return;
  } if(Guest(rcpt)) {
    notify(player, "Guests cannot own any objects.");
    return;
  }

  if(from == rcpt) {
    notify(player, "Nothing to do.");
    return;
  }

  /* check permissions */
  if(!controls(player, from, POW_CHOWN) || !controls(player,rcpt,POW_CHOWN)) {
    notify(player, "Permission denied.");
    return;
  }

  /* chown everything */
  for(a=0;a<db_top;a++)
    if(db[a].owner == from && a != from) {
      halt_object(a);
      db[a].owner = rcpt;
    }

  notify(player, tprintf("Objects owned by %s are now owned by %s.",
         db[from].name, db[rcpt].name));
}

/* send high-priority message to connection-id */
void do_cpage(player, name, msg)
dbref player;
char *name, *msg;
{
  DESC *d;
  int concid=atoi(name);

  /* check concid */
  if(concid < 0) {
    notify(player, "That is an invalid connection id number.");
    return;
  }

  /* search for the correct concid */
  for(d=descriptor_list;d;d=d->next) {
    if(d->concid == concid) {
      if(d->flags & (C_OUTBOUND|C_SOCKET|C_DNS|C_PARENT|C_SERVICE)) {
        notify(player, "You cannot send messages to sockets or services.");
        return;
      }
      output2(d, tprintf("Incoming message from %s(#%d): %s\n",
              db[player].name, player, msg));
      notify(player, tprintf("Message sent to Connection ID %d: %s", concid,
             conc_title(d)));
      return;
    }
  }
  notify(player, "No such connection id number online.");
}

/* boots a connection id */
void do_cboot(player, name)
dbref player;
char *name;
{
  DESC *d;
  int concid=atoi(name);

  /* check concid */
  if(concid < 0) {
    notify(player, "That is an invalid connection id number.");
    return;
  }

  /* search for the correct concid */
  for(d=descriptor_list;d;d=d->next) {
    if(d->concid == concid) {
      if((d->flags & (C_OUTBOUND|C_SOCKET|C_DNS|C_PARENT|C_SERVICE)) &&
         !power(player, POW_CONSOLE)) {
        notify(player, "Permission denied.");
        return;
      }
      if(d->flags & C_DNS) {
        notify(player, "You cannot kill the domain name service resolver.");
        return;
      }
      if(!Invalid(d->player) && !controls(player, d->player, POW_BOOT)) {
        notify(player, "Permission denied.");
        return;
      }
      /* knock em offline */
      notify(player, tprintf("Connection ID %d: %s - Booted.", concid,
             conc_title(d)));
      log_main(tprintf("BOOT: %s(#%d) has booted Connection Id %d: %s", db[player].name,
		    player, concid, conc_title(d)));
      d->flags|=C_LOGOFF;
      return;
    }
  }
  notify(player, "No such connection id number online.");
}

/* regular player boot */
void do_boot(player, name, msg)
dbref player;
char *name, *msg;
{
  DESC *d;
  dbref thing;

  /* boot off all players from the game */
  if(!strcasecmp(name, "All")) {
    for(d=descriptor_list;d;d=d->next)
      if(d->player > 1 && !Immortal(d->player)) {
        notify(d->player, tprintf("Going down - %s.", *msg?msg:"Bye"));
        d->flags|=C_LOGOFF;
      }
    notify(player, "All players booted from the game.");
    log_main(tprintf("BOOT: %s(#%d) boots all players from the game.%s%s",
             db[player].name, player, *msg?" Reason: ":"", msg));
    return;
  }

  /* check permissions */
  if((thing=lookup_player(name)) <= NOTHING) {
    notify(player, tprintf("No such player '%s'.", name));
    return;
  } if(!controls(player, thing, POW_BOOT)) {
    notify(player, "Permission denied.");
    return;
  } if(thing == GOD) {
    notify(player, "You can't boot God!");
    return;
  } if(!boot_off(thing)) {
    notify(player, "That player is not connected.");
    return;
  }

  /* notify people */
  notify(player, tprintf("%s - Booted.", db[thing].name));
  log_main(tprintf("BOOT: %s(#%d) has booted %s%s%s.", db[player].name,
           player, db[thing].name, *msg?" for ":"", msg));
  if(*msg)
    notify(thing, tprintf("You have been booted by %s for %s.",
           db[player].name, msg));
}

/* joins another player */
void do_join(player, arg1)
dbref player;
char *arg1;
{
  dbref thing=lookup_player(arg1);
  
  /* check permissions */
  if(thing <= NOTHING) {
    notify(player, tprintf("No such player '%s'.", arg1));
    return;
  } if(!controls(player, thing, POW_JOIN)) {
    notify(player, "Permission denied.");
    return;
  } if(Typeof(player) != TYPE_PLAYER && !power(player, POW_BUILD)) {
    notify(player, "Only players may join others.");
    return;
  } if(!IS(thing, TYPE_PLAYER, PLAYER_CONNECT) && !power(player, POW_WHO)) {
    notify(player, "But that player isn't even connected!");
    return;
  }

  enter_room(player, db[thing].location);
}

/* summons other players to your location */
void do_summon(player, arg1)
dbref player;
char *arg1;
{
  dbref thing=lookup_player(arg1);
  
  /* check permissions */
  if(thing <= NOTHING) {
    notify(player, tprintf("No such player '%s'.", arg1));
    return;
  } if(!controls(player, thing, POW_SUMMON)) {
    notify(player, "Permission denied.");
    return;
  } if(Typeof(player) != TYPE_PLAYER && !power(player, POW_BUILD)) {
    notify(player, "Only players may summon others.");
    return;
  } if(!IS(thing, TYPE_PLAYER, PLAYER_CONNECT) && !power(player, POW_WHO)) {
    notify(player, "But that player isn't even connected!");
    return;
  }

  speaker=thing;
  enter_room(thing, db[player].location);
}

/* sends a player to the dungeon */
void do_capture(player, arg1, arg2)
dbref player;
char *arg1, *arg2;
{
  dbref thing=lookup_player(arg1);
  
  /* check permissions */
  if(thing <= NOTHING) {
    notify(player, tprintf("No such player '%s'.", arg1));
    return;
  } if(player == thing) {
    notify(player, "You can't jail yourself, silly!");
    return;
  } if(!controls(player, thing, POW_CAPTURE)) {
    notify(player, "Permission denied.");
    return;
  } if(DUNGEON_CELL < 1) {
    notify(player, "Sorry, no Dungeon Cell room is currently configured.");
    return;
  } if(db[thing].location == DUNGEON_CELL) {
    notify(player, tprintf("%s is already in the dungeon.", db[thing].name));
    return;
  }

  if(!IS(DUNGEON_CELL, TYPE_ROOM, ROOM_XZONE))
    notify(player, tprintf("Warning: %s is not set X-Zone!",
           unparse_object(player, DUNGEON_CELL)));

  notify(player, tprintf("%s - Jailed.", db[thing].name));
  log_main(tprintf("CAPTURE: %s(#%d) has jailed %s%s%s.", db[player].name,
           player, db[thing].name, *arg2?" for ":"", arg2));
  notify(thing, tprintf("You have been jailed by %s%s%s.", db[player].name,
         *arg2?" for ":"", arg2));
  enter_room(thing, DUNGEON_CELL);
}

/* frees someone from the dungeon cell */
void do_uncapture(player, arg1)
dbref player;
char *arg1;
{
  dbref thing=lookup_player(arg1);
  
  /* check permissions */
  if(thing <= NOTHING) {
    notify(player, tprintf("No such player '%s'.", arg1));
    return;
  } if(!controls(player, thing, POW_CAPTURE)) {
    notify(player, "Permission denied.");
    return;
  } if(DUNGEON_CELL < 1) {
    notify(player, "Sorry, no Dungeon Cell room is currently configured.");
    return;
  } if(db[thing].location != DUNGEON_CELL) {
    notify(player, "Sorry, that person is not in jail.");
    return;
  }

  notify(player, tprintf("You call upon the executioner to free %s%s soul from this realm.",
         db[thing].name, possname(db[thing].name)));
  log_main(tprintf("UNCAPTURE: %s(#%d) has freed %s.", db[player].name,
           player, db[thing].name));

  enter_room(thing, HOME);
}

/* check if a player is on the inactivity list */
int inactive(dbref player)
{
  long then;

  /* administrators are never considered inactive */
  if(Immortal(player) || (db[player].flags & PLAYER_INACTIVE_OK) ||
     (db[player].flags & PLAYER_CONNECT))
    return 0;

  /* determine last logoff time */
  if((then=db[player].data->last) < db[player].data->lastoff)
    then=db[player].data->lastoff;
  if(!then)
    then=db[player].create_time;  /* player has never connected */

  /* normal characters get 6 months */
  return (int)((now-then)/86400L) >= ((INACTIVENUKE > 1)?INACTIVENUKE:180);
}

/* command to nuke inactive players */
void do_inactivity(dbref player, char *arg1)
{
  int count=0, owned, i, thing;
  long tt;

  /* Perform inactivity nuke */
  if(!strcasecmp(arg1, "All")) {
    if(!power(player, POW_NUKE)) {
      notify(player, "Permission denied.");
      return;
    }
    log_main(tprintf("%s(#%d) does @inactivity. Processing...",
             db[player].name, player));
    if((count=inactive_nuke()))
      notify(player, tprintf("%d player%s deleted.", count, count==1?"":"s"));
    else
      notify(player, "There are no players scheduled for destruction.");
    return;
  }

  /* check inactivity for a specific player */
  if(*arg1) {
    if((thing=lookup_player(arg1)) == AMBIGUOUS)
      thing=player;
    if(thing == NOTHING || Typeof(thing) != TYPE_PLAYER) {
      notify(player, tprintf("No such player '%s'.", arg1));
      return;
    } if(!inactive(thing)) {
      notify(player, tprintf("%s is not inactive.", db[thing].name));
      return;
    }

    /* determine last logoff time */
    if((tt=db[player].data->last) < db[player].data->lastoff)
      tt=db[player].data->lastoff;
    if(!tt)
      tt=db[player].create_time;  /* player has never connected */

    notify(player, tprintf("%s is %ld day%s inactive.", db[thing].name,
           (now-tt)/86400L, ((now-tt)/86400L)==1L?"":"s"));
    return;
  }

  /* read and/or wipeout inactive player list */
  for(thing=0;thing<db_top;thing++) {
    if(Typeof(thing) != TYPE_PLAYER || !inactive(thing))
      continue;

    /* determine last logoff time */
    if((tt=db[thing].data->last) < db[thing].data->lastoff)
      tt=db[thing].data->lastoff;
    if(!tt)
      tt=db[thing].create_time;  /* player has never connected */

    for(i=0,owned=0;i<db_top;i++)
      if(db[i].owner == db[thing].owner && i != thing)
        owned++;
    ++count;
    notify(player, tprintf("%s is %ld day%s inactive: %d objects",
           db[thing].name, (now-tt)/86400L, ((now-tt)/86400L)==1L?"":"s",
           owned));
  }

  notify(player, tprintf("%d Player%s scheduled for destruction.", count,
         count==1?"":"s"));
}

int inactive_nuke()
{
  dbref thing;
  int i, owned, count=0;

  /* read and/or wipeout inactive player list */
  for(thing=0;thing<db_top;thing++) {
    if(Typeof(thing) != TYPE_PLAYER || !inactive(thing))
      continue;

    for(i=0,owned=0;i<db_top;i++)
      if(db[i].owner == db[thing].owner && i != thing)
        owned++;

    log_main(tprintf("%s(#%d): %d objects - Nuked.",
             db[thing].name, thing, owned));

    destroy_player(thing);
    ++count;
  }

  /* log result */
  log_main(tprintf("%d Player%s nuked.", count, count==1?"":"s"));
  return count;
}

/* find the last time another player logged onto the mud */
void do_laston(dbref player, char *name)
{
  dbref thing;
  long tt;

  if(!*name) {
    notify(player, "You must specify a valid player name.");
    return;
  } if((thing = lookup_player(name)) == NOTHING) {
    notify(player, tprintf("No such player '%s'.", name));
    return;
  } if(thing == AMBIGUOUS)
    thing=player;

  /* check if the player is hidden */
  if(!controls(player, thing, POW_WHO) && !could_doit(player,thing,A_HLOCK)) {
    notify(player, tprintf("%s wishes to have some privacy.", db[thing].name));
    return;
  }

  if(!(tt=db[thing].data->last)) {
    notify(player, tprintf("%s has never logged on.", db[thing].name));
    return;
  }
  notify(player, tprintf("%s was last logged on: %s.", db[thing].name,
         mktm(tt, player)));

  if(controls(player, thing, POW_WHO) && !IS(thing, TYPE_PLAYER,
     PLAYER_CONNECT) && (tt=db[thing].data->lastoff))
    notify(player, tprintf("%s last logged off at: %s.", db[thing].name,
           mktm(tt, player)));
}
