/* comm/look.c */
/* commands that display (not modify) rooms, contents and object examination */

#include "externs.h"

/* display contents of object or room */
static void look_contents(player, loc, title)
dbref player, loc;
char *title;
{
  int a, thing, count=0;
  
  if(IS(loc, TYPE_ROOM, ROOM_XZONE))
    return;

  /* search contents list */
  DOLIST(thing, db[loc].contents)
    if((a=can_see(player, thing))) {
      if(*title && !count++)
        notify(player, tprintf("\e[1;37m%s\e[m", title));
      if(IS(player, TYPE_PLAYER, PLAYER_TERSE))
        notify(player, tprintf("%s%s\e[m", unparse_object(player, thing),
               unparse_objheader(player, thing, a)));
      else
        notify(player, unparse_caption(player, thing, a));
    }
}

/* display exits in room */
static void look_exits(player, loc, title)
dbref player, loc;
char *title;
{
  char buf[8520];
  int a, thing;

  /* search exits list */
  *buf=0;
  DOLIST(thing, db[loc].exits)
    if((a=can_see(player, thing))) {
      if(!*buf)
        notify(player, tprintf("\e[1;37m%s\e[m", title));
      else
        strcat(buf, "  ");
      if(strlen(buf) > 8000) {
        strcat(buf, "...");
        break;
      }
      strcat(buf, unparse_exitname(player, thing, 0));
      if(a == 2)
        strcat(buf, "(dark)");
      else if(a == 4)
        strcat(buf, "(secret)");
    }

  /* process and display */
  if(!*buf)
    notify(player, "\e[1;37mNo obvious exits.\e[m");
  else
    notify(player, buf);
}

/* expand exitlist for PLAYER_TERSE flag */
static char *exitlist(player, loc)
dbref player, loc;
{
  static char buf[1024];
  char buf2[40];
  int thing, count=0;

  /* search exits list */
  strcpy(buf, "   \e[1m(");
  DOLIST(thing, db[loc].exits)
    if(can_see(player, thing) == 1) {
      if(count++)
        strcat(buf, ",");
      if(strlen(buf) > 970) {
        strcat(buf, ",...");
        break;
      }
      strncpy(buf2, db[thing].name, 39);
      buf2[39]=0;
      strcat(buf, buf2);
    }

  if(!count)
    return "";
  strcat(buf, ")\e[m");
  return buf;
}

/* looks in the room or at an object */
void do_look(player, arg1)
dbref player;
char *arg1;
{
  dbref thing=db[player].location;
  char buff[8192], *s=NULL;
  int seethru=0, lit;

  if(db[thing].flags & LIGHT) {
    notify(player, "It's too bright to see here.");
    return;
  }

  if(*arg1 && (thing=match_thing(player, arg1, MATCH_POSSESSIVE|
     (power(player,POW_REMOTE)?MATCH_EVERYTHING:MATCH_RESTRICT))) == NOTHING)
    return;

  /* look in depth at a room */
  if(!*arg1 || Typeof(thing) == TYPE_ROOM || thing == db[player].location) {
    if(IS(player, TYPE_PLAYER, PLAYER_TERSE)) {
      db[player].flags &= ~PLAYER_TERSE; /* unset the flag for normal look */
      look_room(player, thing, 0);
      db[player].flags |= PLAYER_TERSE; /* then reset it */
    } else
      look_room(player, thing, 0);
    if(*arg1 && controls(player, thing, POW_EXAMINE))
      look_atrs(player, thing, 0);
    return;
  }

  /* display object header */
  if(Typeof(thing) != TYPE_EXIT)
    notify(player, unparse_caption(player, thing, 0));

  /* display description messages */
  *buff=0;
  if(Typeof(thing) == TYPE_EXIT && !(db[thing].flags & OPAQUE) &&
     ((db[thing].flags & EXIT_TRANSPARENT) || could_doit(player,thing,A_LOCK)))
    seethru=1;
  if(!TIMEOFDAY)
    strcpy(buff, atr_get(thing, A_NDESC));
  if(!*buff)
    strcpy(buff, atr_get(thing, A_DESC));
  if(!*buff && !seethru)
    strcpy(buff, Typeof(thing) == TYPE_PLAYER ? (Guest(thing)?
           "A guest player!  Please be courteous and helpful.":
           "You see someone special and unique."):"You see nothing special.");
  if(*(s=parse_function(thing, player, buff)))
    notify(player, s);

  if(player == thing)
    sprintf(buff, "looks at %sself.", (Typeof(thing) != TYPE_PLAYER)?"it":
            db[thing].data->gender?"her":"him");
  else
    sprintf(buff, "looks %s%s.", (Typeof(thing) != TYPE_EXIT)?"at ":"",
            db[thing].name);
  did_it(player,thing,0,NULL,A_ODESC,buff,(player == thing)?0:A_ADESC);

  /* we are viewing a general object */
  if(!seethru || Invalid(db[thing].link)) {
    look_contents(player, thing, "Carrying:");
    return;
  }

  /* see through exits, as long as you pass its @lock */
  notify(player, tprintf("You peer through to %s...",
         unparse_object(player, db[thing].link)));
  *buff=0;
  lit=can_see(player, db[thing].link);
  if(!TIMEOFDAY || ((!lit || lit == 2) && (db[db[thing].link].flags & DARK)))
    strcpy(buff, atr_get(db[thing].link, A_NDESC));
  if(!*buff)
    strcpy(buff, atr_get(db[thing].link, A_DESC));

  default_color=unparse_color(player, db[thing].link, 4);
  if(*(s=parse_function(db[thing].link, player, buff))) {
    sprintf(buff, "%s%s", textcolor(7, default_color), s);
    notify(player, tprintf("%s%s", buff, textcolor(default_color, 7)));
  } else
    notify(player, "You see nothing descriptive on the other side.");
  default_color=7;

  look_contents(player, db[thing].link, "You also notice:");
}

/* general look in current room */
void look_room(player, loc, typ)
dbref player, loc;
int typ;
{
  char buf[8192], music[100], *s;
  int lit;

  if(Invalid(loc)) {
    notify(player, unparse_object(player, loc));
    return;
  }

  /* when moving rooms, change music */
  *music=0;
  if(typ && terminfo(player, TERM_SOUND)) {
    strcpy(buf, atr_get(loc, A_SOUND));
    if(!*buf)
      strcpy(buf, atr_zone(loc, A_SOUND));
    strncpy(buf, parse_function(loc, player, buf), 90);
    buf[90]=0;
    if(*buf)
      sprintf(music, "\e[%st", buf);
  }

  /* execute non-player description messages and actions */
  if(Typeof(loc) != TYPE_ROOM)
    did_it(player,loc,0,NULL,A_OIDESC,NULL,(player == loc)?0:A_AIDESC);
  else {
    did_it(player,loc,0,NULL,A_ODESC,NULL,(player == loc)?0:A_ADESC);
    if(Typeof(loc) == TYPE_ROOM) {
      if(could_doit(player, loc, A_LOCK))
        did_it(player,loc,0,NULL,A_OSUCC,NULL,A_ASUCC);
      else
        did_it(player,loc,0,NULL,A_OFAIL,NULL,A_AFAIL);
    }
  }

  /* check if player is blind */
  if(db[player].location == loc) {
    if(db[loc].flags & LIGHT) {
      notify(player, tprintf("%s\e[1;37mThe light here is too bright for you to see anything.\e[m",
             music));
      return;
    }
  }

  lit=can_see(player, loc);

  /* display room title and description */
  if(IS(player, TYPE_PLAYER, PLAYER_TERSE))
    notify(player, tprintf("%s%s%s", music, unparse_object(player, loc),
           exitlist(player, loc)));
  else {
    /* display room title */
    notify(player, tprintf("%s%s%s", music, unparse_object(player, loc),
           IS(player, TYPE_PLAYER, PLAYER_TERSE)?exitlist(player, loc):""));

    *buf=0;
    if(Typeof(loc) != TYPE_ROOM)
      strcpy(buf, atr_get(loc, A_IDESC));
    if(!*buf && (!TIMEOFDAY || ((!lit || lit == 2) && (db[loc].flags & DARK))))
      strcpy(buf, atr_get(loc, A_NDESC));
    if(!*buf)
      strcpy(buf, atr_get(loc, A_DESC));
    default_color=unparse_color(player, loc, 4);
    if(*(s=parse_function(loc, player, buf)))
      notify(player, tprintf("%s%s", textcolor(7, default_color), s));

    if(Typeof(loc) == TYPE_ROOM) {
      if(could_doit(player, loc, A_LOCK))
        did_it(player,loc,A_SUCC,NULL,0,NULL,0);
      else
        did_it(player,loc,A_FAIL,NULL,0,NULL,0);
    }
    default_color=7;
  }

  /* display contents */
  look_contents(player, loc, "Contents:");
  if(Typeof(loc) != TYPE_PLAYER) {
    if(!IS(player, TYPE_PLAYER, PLAYER_TERSE))
      look_exits(player, loc, "Obvious exits:");
  }
}

/* utility to test if <player> can see <thing> in normal conditions */
/* rslt=0: player cannot see the object
        1: player can see object
        2: wizard looking at (hidden) object
        3: player with infravision sees red objects
        4: player sees hidden exit in cave */
int can_see(dbref player, dbref thing)
{        
  int dark=controls(player, thing, POW_DARK)?2:0;
  dbref loc=db[thing].location;

  /* Normal look doesn't include player himself, disconnected players,
     or those that aren't in the same dimensional plane. If the plane is
     NOTHING (-1), then the object exists in all planes. */
     
  if(player == thing ||
     (db[thing].plane != NOTHING && db[thing].plane != db[player].plane) ||
     (Typeof(thing) == TYPE_PLAYER && (!(db[thing].flags & PLAYER_CONNECT))))
  return 0;

  /* Check bright room and visual-lock */
  if((db[loc].flags & LIGHT) || !could_doit(player, thing, A_VLOCK))
    return dark;

  /* Objects in inventory and things set Light are always visible. */
  if(loc == player || (Typeof(thing) != TYPE_PLAYER &&
     (db[thing].flags & LIGHT)))
    return 1;

  /* A player's torch lights up the whole room for everyone */

  /* Dark objects are hidden to normal eye. */
  if((db[thing].flags & DARK) && (Typeof(thing) == TYPE_THING ||
     Typeof(thing) == TYPE_EXIT))
    return dark;

  /* Rooms/caves set dark, objects set opaque */
  if(IS(db[thing].location, TYPE_ROOM, DARK) || (db[player].location !=
     db[thing].location && (db[db[thing].location].flags & OPAQUE)))
    return dark;

  /* everything else is a normal, bright room */
  return 1;
}

/* show visible exits and their destinations */
void do_exits(dbref player)
{
  char buf[2100];
  int a, thing, count=0, loc=db[player].location;

  /* players cannot see in a dark (or very bright) room */
  if(db[loc].flags & LIGHT) {
    notify(player, "You can't see the exits!");
    return;
  }

  /* search exits list */
  DOLIST(thing, db[loc].exits)
    if((a=can_see(player, thing))) {
      if(!count++)
        notify(player, "\e[1mVisible exits:\e[m");
      if(!(db[thing].flags & EXIT_TRANSPARENT) &&
         !could_doit(player, thing, A_LOCK))
        strcpy(buf, "You can't tell.");
      else if(db[thing].link == NOTHING)
        strcpy(buf, "Unlinked.");
      else
        strcpy(buf, unparse_object(player, db[thing].link));
      notify(player, tprintf("  %s -> %s%s", unparse_exitname(player,thing,1),
             buf, a==2?"  \e[1;30m(dark)\e[m":
             a==4?"  \e[1;30m(secret)\e[m":""));
    }

  /* process and display */
  if(!count)
    notify(player, "\e[1;37mNo obvious exits.\e[m");
}


static struct all_atr_list *all_attrs_internal(thing, myop, dep)
dbref thing;
struct all_atr_list *myop;
int dep;
{
  ALIST *a;
  ATTR *atr;
  dbref *ptr;
  struct all_atr_list *tmp, *last;

  for(a=db[thing].attrs;a;a=a->next) {
    if(!(atr=atr_num(a->obj, a->num)))
      continue;
    if(!dep || (atr->flags & INH)) {
      /* check to see if there's already something with this type */
      last=NULL;
      for(tmp=myop;tmp;tmp=tmp->next) {
        if(tmp->type == atr)
          break;
        last=tmp;
      }
      if(tmp)
        continue;

      tmp=malloc(sizeof(struct all_atr_list));
      tmp->next=NULL;
      tmp->type=atr;
      tmp->value=a->text;
      tmp->numinherit=dep;
      if(last)
        last->next=tmp;
      else
        myop=tmp;
    }
  }

  for(ptr=db[thing].parents;ptr && *ptr != NOTHING;ptr++)
    myop=all_attrs_internal(*ptr, myop, dep+1);
  return myop;
}

struct all_atr_list *all_attributes(dbref thing)
{
  /* start with 0 depth */
  return all_attrs_internal(thing, NULL, 0);
}

/* display attribute with #dbref's unparsed */
char *unparse_dbref(player, arg, type)
dbref player;
char *arg;
int type;
{
  static char buf[8192];
  char temp[strlen(arg)+1], *s;
  int a, num, count=0, dep=0;

  strcpy(temp, arg);

  for(a=0;temp[a] && count < 8000;a++) {
    if(!type && temp[a] == '[')
      dep++;
    else if(dep > 0 && temp[a] == ']')
      dep--;

    /* parse #dbref */
    if(!dep && temp[a] == '#' && isdigit(temp[a+1])) {
      num=atoi(temp+a+1);
      if(Invalid(num) || Typeof(num) == TYPE_GARBAGE) {
        buf[count++]=temp[a];  /* print #dbref as is */
        continue;
      }

      buf[count]='\0';
      s=unparse_object(player, num);
      if(strlen(s)+count < 8000) {
        strcpy(buf+count, s);
        count=strlen(buf);
        while(temp[a+1] && isdigit(temp[a+1]))  /* skip over rest of number */
          a++;
        continue;
      }
    }

    /* just add character to string */
    buf[count++]=temp[a];
  }

  buf[count]='\0';
  return buf;
}

/* examine a single attribute */
static void look_atr(player, thing, allatrs)
dbref player, thing;
struct all_atr_list *allatrs;
{
  char buf[16], hex[16], *s=uncompress(allatrs->value);
  int flags=allatrs->type->flags;

  *buf=0;
  if(allatrs->numinherit)
    sprintf(buf, "%d>", allatrs->numinherit);
  notify(player, tprintf("\e[1m%s%s%s:\e[m%s", buf, allatrs->type->name,
         (flags & AF_FUNC)?"()":"", (flags & AF_DATE)?mktm(atol(s), player):
         (flags & AF_LOCK)?unparse_dbref(player, s, 0):  /* top level only */
         (flags & AF_DBREF)?unparse_dbref(player, s, 1): /* in functions too */
         (flags & AF_BITMAP)?(sprintf(hex, "0x%08x", atoi(s)), hex):s));
}

/* examine all visible attributes on a thing */
void look_atrs(player, thing, inherit)
dbref player, thing;
int inherit;
{
  struct all_atr_list *allatrs, *anext;
  int count=0;

  for(allatrs=all_attributes(thing);allatrs;free(allatrs),allatrs=anext) {
    anext=allatrs->next;
    if(!(allatrs->type->obj == NOTHING && allatrs->type->num == A_DESC) &&
       can_see_atr(player, thing, allatrs->type) && (inherit || (!inherit &&
       !allatrs->numinherit))) {
      if(!count++)
        notify(player, "\e[1mAttribute list:\e[m");
      look_atr(player, thing, allatrs);
    }
  }
}

ATTR *match_attr(player, thing, attrname, typ, attrvalue)
dbref player, thing;
char *attrname, *attrvalue;
int typ;
{
  struct all_atr_list *allatrs, *anext;
  ATTR *found=NULL;

  for(allatrs=all_attributes(thing);allatrs;free(allatrs),allatrs=anext) {
    anext=allatrs->next;
    if(!found && (!typ || can_see_atr(player, thing, allatrs->type)) &&
       wild_match(allatrs->type->name, attrname, 0, 0))
      if(!attrvalue || !*attrvalue || wild_match(uncompress(allatrs->value),
         attrvalue, 0, 0))
        found=allatrs->type;
  }
  return found;
}

/* display attributes on an object */
void do_examine(player, name, arg2, argv, pure, cause, option)
dbref player, cause;
char *name, *arg2, **argv, *pure;
int option;
{
  dbref *ptr, content, exit, thing=db[player].location;
  char buf[2400], *s;
  int i=0, count=0;

  if((s=strchr(name, '/')))
    *s++='\0';

  if(*name)
    if((thing=match_thing(player, name, MATCH_EVERYTHING|((player == GOD)?
       MATCH_DESTROYED:0))) == NOTHING)
      return;

  /* display specialized techniques only */
  if(option == 3) {
    notify(player, "Function not implemented.");
    return;
  }

  /* searching for a pattern in attributes */
  if(strchr(arg2, '*') || strchr(arg2, '?')) {
    if(!s)
      s="*";
    i=1;  /* use wildcard value matching */
  }

  /* are we examining only one attribute on object? */
  if(s) {
    struct all_atr_list *allatrs, *anext;

    /* search attribute list */
    for(allatrs=all_attributes(thing);allatrs;free(allatrs),allatrs=anext) {
      anext=allatrs->next;
      if(!can_see_atr(player, thing, allatrs->type))
        continue;
      if(!wild_match(allatrs->type->name, s, 0, 0))
        continue;
      if(i && !wild_match(uncompress(allatrs->value), arg2, 0, 0))
        continue;

      look_atr(player, thing, allatrs);
      count++;
    }

    if(!count)
      notify(player, "No match.");
    return;
  }

  /* display object header */
  notify(player, unparse_object(player, thing));
  if(*(s=atr_get(thing, A_DESC)) && can_see_atr(player, thing,
     atr_num(NOTHING, A_DESC)))
    notify(player, tprintf("\e[1mDesc:\e[m%s", s));
  else if(Typeof(thing) == TYPE_PLAYER && *(s=atr_parse(thing,player,A_PLAN)))
    notify(player, s);

  if(thing != db[thing].owner)
    notify(player, tprintf("\e[1mOwner:\e[m %s", db[db[thing].owner].name));
  *buf=0;
  if(*(s=unparse_flags(player, thing, 1)))
    sprintf(buf, "  \e[1mFlags:\e[m %s", s);
  notify(player, tprintf("\e[1mType:\e[m %s%s", typenames[Typeof(thing)],buf));
  if(db[thing].zone) {
    *buf=0;
    for(ptr=db[thing].zone;ptr && *ptr != NOTHING;ptr++) {
      if(strlen(buf) > 1000) {
        strcat(buf, " ...");
        break;
      }
      strcat(buf, ", ");
      strcat(buf, unparse_object(player, *ptr));
    }
    notify(player, tprintf("\e[1mZone:\e[m%s\e[m", buf+1));
  }

  /* show them short description of object if owned by someone else */
  if(!controls(player, thing, POW_EXAMINE) && !(db[thing].flags & VISIBLE)) { 
    look_atrs(player, thing, 0);  /* only show osee attributes */

    /* check to see if there is anything there owned by other player */
    count=0;
    DOLIST(content, db[thing].contents)
      if(controls(player, content, POW_EXAMINE)) {
        if(!count++)
          notify(player, "\e[1mContents:\e[m");
        notify(player, unparse_object(player, content));
      }
    return;
  }
  
  /* begin regular examine routines and show stats on the object */
  
  /* display monetary value */
  if(Typeof(thing) == TYPE_PLAYER) {
    *buf=0;
    notify(player, tprintf("\e[1m%s:\e[m %d%s", MONEY_PLURAL,db[thing].pennies,
           buf));
    notify(player, tprintf("\e[1mSessions:\e[m %d  \e[1mSteps:\e[m %d",
           db[thing].data->sessions, db[thing].data->steps));
  }

  /* display memory usage */
  if(Typeof(thing) == TYPE_PLAYER)
    sprintf(buf, "  \e[1mTotal:\e[m %d", playmem_usage(thing));
  else
    *buf=0;
  notify(player, tprintf("\e[1mBytes:\e[m %d%s", mem_usage(thing), buf));

  /* display dimensional plane value if non-default */
  if(power(player, POW_DARK) &&
     (db[thing].plane != (Typeof(thing) == TYPE_EXIT ? NOTHING : 0)))
    notify(player, tprintf("\e[1mPlane:\e[m %d", db[thing].plane));

  /* display creation/modification dates */
  if(db[thing].create_time > 0L)
    notify(player, tprintf("\e[1mCreation Date:\e[m %s",
           mktm(db[thing].create_time, player)));
  if(db[thing].mod_time > 0L)
    notify(player, tprintf("\e[1mModification Time:\e[m %s",
           mktm(db[thing].mod_time, player)));

  /* list parents & relay rooms */
  if(db[thing].parents) {
    *buf=0;
    for(ptr=db[thing].parents;ptr && *ptr != NOTHING;ptr++) {
      if(strlen(buf) > 1000) {
        strcat(buf, " ...");
        break;
      }
      strcat(buf, ", ");
      strcat(buf, unparse_object(player, *ptr));
    }
    notify(player, tprintf("\e[1mParents:\e[m%s", buf+1));
  }
  if(db[thing].relay) {
    *buf=0;
    for(ptr=db[thing].relay;ptr && *ptr != NOTHING;ptr++) {
      if(strlen(buf) > 1000) {
        strcat(buf, " ...");
        break;
      }
      strcat(buf, ", ");
      strcat(buf, unparse_object(player, *ptr));
    }
    notify(player, tprintf("\e[1mRelay:\e[m%s", buf+1));
  }

  /* view attributes */
  if(db[thing].atrdefs) {
    ATRDEF *k;

    count=0;
    for(k=db[thing].atrdefs;k;k=k->next) {
      if((k->atr->flags & AF_HIDDEN) && player != GOD)
        continue;
      if(!count++)
        notify(player, "\e[1mAttribute definitions:\e[m");
      notify(player, tprintf("  %s%s%s%s", k->atr->name,
             (k->atr->flags & AF_FUNC)?"()":"",
             (k->atr->flags)?": ":"", unparse_attrflags(k->atr->flags)));
    }
  }

  /* 'All' option: View attributes set in parents also */
  look_atrs(player, thing, (option == 1)?1:0);
  
  if(option == 2)  /* Brief option. No contents or links displayed */
    return;

  /* show him the contents */
  if(db[thing].contents != NOTHING) {
    notify(player, "\e[1mContents:\e[m");
    DOLIST(content, db[thing].contents)
      notify(player, unparse_object(player, content));
  }
  
  /* lastly, display link/location factors */

  /* rooms show entrances and exits */
  if(Typeof(thing) == TYPE_ROOM) {
    for(i=0,count=0;i<db_top;i++)
      if(Typeof(i) == TYPE_EXIT && db[i].link == thing &&
         (controls(player, i, POW_EXAMINE) ||
         controls(player, thing, POW_EXAMINE))) {
        if(!count++)
          notify(player, "\e[1mEntrances:\e[m");
        notify(player, unparse_object(player, i));
      }
    if(!count)
      notify(player, "\e[1mNo entrances.\e[m");

    if(db[thing].exits != NOTHING) {
      notify(player, "\e[1mExits:\e[m");
      DOLIST(exit, db[thing].exits)
	notify(player, unparse_object(player, exit));
    } else
      notify(player, "\e[1mNo exits.\e[m");

    if(db[thing].link != NOTHING)
      notify(player, tprintf("\e[1mDropto:\e[m %s",
             unparse_object(player, db[thing].link)));
    return;
  }

  /* exits show source and destination */
  if(Typeof(thing) == TYPE_EXIT) {
    notify(player, tprintf("\e[1mSource:\e[m %s", unparse_object(player,
           db[thing].location)));
    if(db[thing].link != NOTHING)
      notify(player, tprintf("\e[1mDestination:\e[m %s",
             unparse_object(player, db[thing].link)));
    return;
  }

  /* players, zones, and things have link/location */
  if(db[thing].location != NOTHING && Typeof(thing) != TYPE_ZONE)
    notify(player, tprintf("\e[1mLocation:\e[m %s", unparse_object(player,
           db[thing].location)));
  if(db[thing].link != NOTHING)
    notify(player, tprintf("\e[1m%s:\e[m %s", Typeof(thing) == TYPE_ZONE?
           "Origin Room":"Home", unparse_object(player, db[thing].link)));
}

/* list items and things in your inventory */
void do_inventory(dbref player)
{
  look_contents(player, player, "You are carrying:");

  /* display wealth */
  if(Typeof(player) == TYPE_PLAYER)
    notify(player, tprintf("You have \e[1;32m%d\e[m %s.", db[player].pennies,
           db[player].pennies==1?MONEY_SINGULAR:MONEY_PLURAL));
}

/* print the commands required to best recreate an object */
void do_decompile(player, arg1)
dbref player;
char *arg1;
{
  ATTR *atr;
  ALIST *a;
  ATRDEF *k;
  char objname[512], *s;
  int thing;

  if((thing=match_thing(player, arg1, MATCH_EVERYTHING)) == NOTHING)
    return;
  if(!controls(player, thing, POW_EXAMINE)) {
    notify(player, "Permission denied.");
    return;
  }

  switch(Typeof(thing)) {
    case TYPE_ROOM: strcpy(objname, "here"); break;
    case TYPE_ZONE: sprintf(objname, "#%d", thing); break;
    default:
      if(!strcasecmp(arg1, "me"))
        strcpy(objname, "me");
      else
        strcpy(objname, db[thing].name);
  }

  if(*(s=unparse_flags(player, thing, 1)))
    notify(player, tprintf("@set %s=%s", objname, s));

  for(k=db[thing].atrdefs;k;k=k->next) {
    if((k->atr->flags & AF_HIDDEN) && player != GOD)
      continue;
    notify(player, tprintf("@defattr %s/%s%s%s", objname, k->atr->name,
           (k->atr->flags)?"=":"", unparse_attrflags(k->atr->flags)));
  }

  for(a=db[thing].attrs;a;a=a->next)
    if((atr=atr_num(a->obj, a->num)))
      if(!(atr->flags & AF_UNIMP) && can_see_atr(player, thing, atr))
        notify(player, tprintf("@set %s=%s:%s", objname, atr->name,
               uncompress(a->text)));
}

/* scope out the location of a distant player */
void do_whereis(player, name)
dbref player;
char *name;
{
  dbref thing;
  
  if((thing = lookup_player(name)) <= NOTHING) {
    notify(player, tprintf("No such player '%s'.", name));
    return;
  } if((!(db[thing].flags & PLAYER_CONNECT) || !could_doit(player, thing,
       A_HLOCK)) && !controls(player, thing, POW_WHO)) {
    notify(player, "That player is not connected.");
    return;
  }

  if(!controls(player, thing, POW_EXAMINE) && (db[thing].flags & DARK))
    notify(player, tprintf("%s wishes to have some privacy.", db[thing].name));
  else if(!locatable(player, thing))
    notify(player, tprintf("%s can't be located.", db[thing].name));
  else
    notify(player, tprintf("%s is at: %s", db[thing].name,
           unparse_object(player, db[thing].location)));
}

/* view personal stats on a player */
void do_finger(player, arg1, arg2, argv, pure, cause, option)
dbref player, cause;
char *arg1, *arg2, **argv, *pure;
int option;
{
  char buf[9000], buff[40], *s, *gend;
  dbref thing;
  long a;

  if(!*arg1) {
    notify(player, "You must specify a valid player name.");
    return;
  }

  if((thing=lookup_player(arg1)) == AMBIGUOUS)
    thing=player;
  if(Invalid(thing) || Typeof(thing) != TYPE_PLAYER) {
    notify(player, tprintf("No such player '%s'.", arg1));
    return;
  }
  gend=(db[thing].data->gender)?"She":"He";

  notify(player, unparse_caption(player, thing, 0));

  /* people using +finger on God do not see anything but his level & plan */
  if(thing == GOD) {
    if(*(s=atr_parse(thing, player, A_PLAN)))
      notify(player, s);
    return;
  }

  if(!could_doit(player,thing,A_PLOCK))
    notify(player, tprintf("%s is not accepting pages.", db[thing].name));

  /* Display either player's location (if locatable) and idle time, or display
   * when the player last logged on if the player is hidden or offline. */

  if(controls(player, thing, POW_WHO) || could_doit(player, thing, A_HLOCK)) {
    if(db[thing].flags & PLAYER_CONNECT) {
      if(locatable(player, thing))
        notify(player, tprintf("%s can be found at: %s", gend,
               unparse_object(player, db[thing].location)));
      sprintf(buf, "is %s idle", time_format(TMA, getidletime(thing)));
      if(*(s=atr_get(thing, A_IDLE)))
        strcat(buf, tprintf(": %s", s));
      else
        strcat(buf, ".");
    } else {
      if(!(a=db[thing].data->last))
        sprintf(buf, "has never connected.");
      else
        sprintf(buf, "hasn't connected in %s.", time_format(TMA, now-a));
    }
    notify(player, tprintf("%s is %s old, and %s", gend,
           time_format(TMA, db[thing].data->age), buf));
    if(!(db[thing].flags & PLAYER_CONNECT) && *(s=atr_get(thing, A_AWAY)))
      notify(player, tprintf("Away message from %s: %s", db[thing].name, s));
  }

  if(power(player, POW_WHO) && controls(player, thing, POW_WHO) &&
     *(s=atr_get(thing, A_LASTFROM)))
    notify(player, tprintf("%s last connected from the address %s.", gend, s));

  strcpy(buff, gender_subst(thing, POSS));
  *buff=toupper(*buff);
  if(db[thing].flags&PLAYER_CONNECT && (controls(player,thing,POW_WHO) ||
     could_doit(player,thing,A_HLOCK))) {
    if(*(s=atr_parse(thing, player, A_DOING)))
      notify(player, tprintf("%s character is currently doing: %s", buff, s));
    else
      notify(player, tprintf("%s doesn't seem to be doing anything at the moment.", gend));
  }
  if(Immortal(player) || terminfo(thing, TERM_REGINFO)) {
    if(*(s=atr_get(thing, A_RLNAME)))
      notify(player, tprintf("%s Real Life Name is %s.", buff, s));
    if(*(s=atr_get(thing, A_EMAIL)))
      notify(player, tprintf("%s Email Address is %s", buff, s));
  }

  /* 'brief' command option disables printing the user's @plan */
  if(option == 1)
    return;

  /* show plan */
  if(*(s=atr_parse(thing, player, A_PLAN))) {
    notify(player, "Plan:");
    notify(player, s);
  } else
    notify(player, "No plan.");
}
