/* db/compress.c */
/* Handles RAM compression routines for swift database loading and saving */

#include "externs.h"

/* Top 154 bigrams in the WindsMARE database as of 7/19/1995 */ 
char *tokens[154] = {
  "qu",
  "e ", "th", " t", "he", "s ", "ou", "in", " a",
  "st", " s", "t ", "er", "es", "n ", "al", "d ",
  "to", "an", "ar", "ea", "ro", "o ", "re", "or",
  "nd", " o", " w", " f", "on", " i", " b", "wa",
  "r ", "ll", "as", "y ", " c", "at", "u ", "om",
  "ng", "l ", "le", " d", "ut", "ow", ". ", "k ",
  "h ", "nt", "f ", "en", "se", "me", "ne", "g ",  /* codes 127-255 */
  ", ", "rt", "it", "of", "no", "hi", "is", "ve",
  "il", "te", "ra", " h", "lo", " m", " l", "li",
  "ge", " n", " r", "ac", "ee", "ai", "ck", "ma",
  "el", "we", "co", "a ", " y", "so", "de", "m ",
  "yo", "t.", "oo", "lk", "Yo", " p", "do", "un",
  "  ", "la", "et", "ha", "ti", " e", "ri", "ta",
  "ad", "Th", "tr", "gh", "ed", "ur", "ca", "fr",
  "wn", "be", " T", "pe", "ch", "ks", "id", "ba",
  "ce", "rs", " g", "ay", "up", " u", "em", "ot",
  
  "wi", "e.", "p ", "ds", "ss", "fo", "ic", "pa", "us",  /* codes 1-9 */
 
  "mi", "rn", "si", "ag", "ly", "ig", "ns", "ir",  /* codes 11-26 */
  "ho", "ld", "h.", "im", "da", "sh", "ie", "s." };

/* code 10 = DB Newline */
/* code 27 = ESC */
/* code 28 = Run-Length Encoding */
/* code 29 = TAB */
/* code 30 = Bell */
/* code 31 = Linefeed */

char token_table[128][128];
char highfont[256]=
  "                                "
  "   First 128 chars Initialized  "
  "  by the following function.    "
  "                                "
  "FUTORKCGWHNIJPZS><         c#   "
  "  **+   ?C   !<>  :||          ."
  "`^.|-+NPWBTEWFR|CND|DI|~_'. ||||"
  "olaiiahoeuorc> |=  u   >^.1*32-";

void init_compress()
{
  int i;
  
  memset(token_table, 0, 128*128);

  for(i=0;i<129;i++)
    token_table[(int)tokens[i][0]][(int)tokens[i][1]] = i+127;
  for(; i < 138;i++)
    token_table[(int)tokens[i][0]][(int)tokens[i][1]] = i-128;
  for(; i < 154;i++)
    token_table[(int)tokens[i][0]][(int)tokens[i][1]] = i-127;

  for(i=0;i<128;i++)
    highfont[i]=i;
}

char *compress(unsigned char *s)
{
  static char buf[10000];
  char *to;
  char token=0;
  int a, old=0, count=0, bmatch;

  if(!s)
    return NULL;

  /* tokenize the first characters */
  for(to = buf;s[0];) {
    bmatch = (s[1] && (token = token_table[(int)s[0]][(int)s[1]]))?1:0;
    if(s[0] == old && count < 257 && s[1] && !bmatch &&
       s[0] != 9 && s[0] != 7 && s[0] != '\n')
      ++count;
    else {
      if(count > 0) {
        if(count < 4) {
          for(a=0;a<count;a++)
            *to++ = old;
        } else {
          *to++ = 28;
          if(count < 13)
            *to++ = count-3;
          else
            *to++ = count-2;
          *to++ = old;
        }
      }
      if(s[0] == 9) {
        *to++ = 29;
      } else if(s[0] == 7) {
        *to++ = 30;
      } else if(s[0] == '\n') {
        *to++ = 31;
      } else if(bmatch) {
        *to++ = token;
        s++;
      } else if(!s[1]) {
        *to++ = s[0];
      } else {
        old=s[0];
        count=1;
        s++;
        continue;
      }
      old=0;
      count=0;
    }
    s++;
  }
  
  /* copy the last character (if any) and null */
  while((*to++ = *s++));
  
  return buf;
}

char *uncompress(unsigned char *s)
{
  static char buf[10000];
  char *to;
  char *token;
  int a, unv, rep;

  if(!s)
    return NULL;
  for(to = buf; *s; s++) {
    if(*s > 126) {
      token = tokens[*s - 127];
      *to++ = *token++;
      *to++ = *token;
      continue;
    } else if(*s > 0 && *s < 10) {
      token = tokens[*s + 128];
      *to++ = *token++;
      *to++ = *token;
      continue;
    } else if(*s > 10 && *s < 27) {
      token = tokens[*s + 127];
      *to++ = *token++;
      *to++ = *token;
      continue;
    }

    switch(*s) {
      case 28:
        if(!*(++s)) {
          *to++ = 0;
          return buf;
        }
        unv=*s;
        if(unv > 0 && unv < 10)
          rep=unv+3;
        else if(unv > 10)
          rep=unv+2;
        else
          break;
        s++;
        for(a=0;a<rep;a++)
          *to++ = *s;
        break;
      case 29: *to++ = 9; break;
      case 30: *to++ = 7; break;
      case 31: *to++ = '\n'; break;
      default: *to++ = *s;
    }
  }
  
  *to++ = *s;
  
  return buf;
}
