/* db/destroy.c */

#include "externs.h"

/* free list statistics */
dbref purge_notify=NOTHING;
int numrooms=0;
int numfloat=0;
int numgoing=0;
int ndisrooms=0;
int ndestroyed=0;
int ninactive=0;
int treedepth=0;

dbref next_free=0;
dbref *freelist=NULL;

struct goingobj {
  struct goingobj *next;
  dbref obj;
  long when;
} *goinglist=NULL;

int purge_loop=0;


/* clears an object's contents and recycles it for later use */
void do_destroy(player, arg1, arg2)
dbref player;
char *arg1, *arg2;
{
  struct goingobj *ptr=NULL;
  long delay=now;
  dbref thing;
  
  /* check permissions */
  if((thing=match_thing(player, arg1, MATCH_EVERYTHING)) == NOTHING)
    return;
  if(!controls(player, thing, POW_MODIFY) && (Typeof(thing) != TYPE_THING ||
     db[thing].location != player || !(db[thing].flags & THING_DESTROY_OK))) {
    notify(player, "Permission denied.");
    return;
  } if(thing < 2) {
    notify(player, "Don't you think that's sorta an odd thing to destroy?");
    return;
  } if(Typeof(thing) == TYPE_PLAYER) {
    notify(player, "Destroying players isn't allowed. Try @nuke instead.");
    return;
  }

  /* find object in Going list */
  if(db[thing].flags & GOING)
    for(ptr=goinglist;ptr;ptr=ptr->next)
      if(ptr->obj == thing)
        break;

  /* check if object is already set Going */
  if(!*arg2 && ptr) {
    notify(player, tprintf("%s is already scheduled for destruction in %s.",
           unparse_object(player, thing), time_format(TMF, ptr->when-now)));
    return;
  }

  /* set destruction time (in seconds) */
  if(*arg2) {
    if((delay=get_date(arg2, player, now)) == -1L) {
      notify(player, "Invalid time format string.");
      return;
    } if(delay > now+604800L) {
      notify(player, "Sorry, the maximum delay for destruction is 1 week.");
      return;
    }
  } else {
    /* rooms and zones have extended periods of time just in case */
    if(!power(player, POW_SECURITY) && db[player].owner != db[thing].owner &&
       !IS(thing, TYPE_THING, THING_DESTROY_OK)) {
      notify(player, "Warning: You do not own that object.");
      delay+=900L;
    } else if(Typeof(thing) == TYPE_ROOM || Typeof(thing) == TYPE_ZONE)
      delay+=600L;
    else if(db[thing].children) {
      notify(player, "Warning: It has children.");
      delay+=60L;
    }
  }
  
  /* print destruction message.. */
  if(delay > now) {
    if(!ptr) {
      ptr=malloc(sizeof(struct goingobj));
      ptr->next=goinglist;
      ptr->obj=thing;
      goinglist=ptr;
    }
    ptr->when=delay;

    db[thing].flags |= GOING;
    if(Typeof(thing) == TYPE_ROOM)
      notify_except(thing, NOTHING, "The room shakes and begins to crumble.");
    notify(player, tprintf("%s is now scheduled for destruction in %s.",
           unparse_object(player, thing), time_format(TMF, delay-now)));
    return;
  }

  /* ..or immediately destroy object */
  notify(player, tprintf("%s has been destroyed.",
         unparse_object(player, thing)));
  recycle(thing);
  validate_links();
}

/* undestroy an object already scheduled for destruction */
void do_undestroy(dbref player, char *arg1)
{
  struct goingobj *last=NULL, *ptr;
  dbref thing;
  
  if((thing=match_controlled(player, arg1, POW_MODIFY)) == NOTHING)
    return;
  if(!(db[thing].flags & GOING)) {
    notify(player, "Sorry, that object is not scheduled for destruction.");
    return;
  }

  /* remove destruction schedule from going-list */
  for(ptr=goinglist;ptr;ptr=ptr->next) {
    if(ptr->obj == thing) {
      if(last)
        last->next=ptr->next;
      else
        goinglist=ptr->next;
      free(ptr);
      break;
    }
    last=ptr;
  }

  db[thing].flags &= ~GOING;
  notify(player, tprintf("%s has been saved from destruction.",
         unparse_object(player, thing)));
}

// Send contents of object home & destroy its exits. Clear object completely.
//
// WARNING: validate_links() MUST be called after all recycle() functions
// are processed and before any other command is executed.
//
void recycle(dbref thing)
{
  eqtech *tech, *tnext;
  eqlist *mgk, *mgk2;
  eqdef *def, *dnext;
  eqlist *entry, *eqnext;
  struct goingobj *dest;
  ATRDEF *k, *k2;
  ALIST *atr, *atr2;
  dbref i, next, *ptr;
  
  speaker=thing;

  /* mask object in destruction list */
  for(dest=goinglist;dest;dest=dest->next)
    if(dest->obj == thing) {
      dest->obj=NOTHING;
      break;
    }

  /* display room messages */
  if(Typeof(thing) == TYPE_ROOM) {
    notify_except(thing,NOTHING,"The floor disappears beneath your feet..!");
    notify_except(thing,NOTHING,"It is the end of the world.. as we know it.");
  }

  /* first, clear all commands in the object's queue */
  db[thing].flags |= HAVEN;
  halt_object(thing);

  /* clear out exits in room */
  while(db[thing].exits != NOTHING)
    recycle(db[thing].exits);

  /* clear out contents */
  for(i=db[thing].contents;i != NOTHING;i=next) {
    next=db[i].next;
    if(i == thing)
      continue;

    if(db[i].link == thing || Typeof(db[i].link) == TYPE_GARBAGE)
      db[i].link=(Typeof(i) == TYPE_PLAYER)?PLAYER_START:db[i].owner;

    if(Typeof(thing) != TYPE_ROOM && (Typeof(i) != TYPE_PLAYER ||
       Typeof(db[thing].location) != TYPE_PLAYER))
      enter_room(i, db[thing].location);
    else
      enter_room(i, HOME);
  }

  /* clear all rooms zoned to <thing> */
  if(Typeof(thing) == TYPE_ZONE)
    for(i=0;i<db_top;i++)
      PULL(db[i].zone, thing);

  /* move object to location NOTHING */
  if(Typeof(thing) != TYPE_ROOM && Typeof(thing) != TYPE_ZONE) {
    if(!game_online && nstat[NS_EXEC] == 1) {
      if(Typeof(thing) == TYPE_EXIT)
        db[db[thing].location].exits =
             remove_first(db[db[thing].location].exits, thing);
      else
        db[db[thing].location].contents =
             remove_first(db[db[thing].location].contents, thing);
    } else
      enter_room(thing, NOTHING);
  }

  /* Now lets clean up the object completely. */

  WPTR(db[thing].name, "");
  db[thing].owner=GOD;
  db[thing].flags=TYPE_GARBAGE;
  db[thing].location=NOTHING;
  db[thing].link=NOTHING;
  db[thing].exits=NOTHING;
  db[thing].contents=NOTHING;
  db[thing].bitmaplen=0;
  free(db[thing].bitmap);
  db[thing].bitmap=NULL;

  /* clear attributes and attribute definitions */
  for(atr=db[thing].attrs;atr;atr=atr2) {
    atr2=atr->next;
    free(atr);
  }
  for(k=db[thing].atrdefs;k;k=k2) {
    k2=k->next;
    free(k->atr->name);
    free(k->atr);
    free(k);
  }
  db[thing].attrs=NULL;
  db[thing].atrdefs=NULL;

  /* clear player data */
  if(Typeof(thing) == TYPE_PLAYER)
    free(db[thing].data);
  db[thing].data=NULL;

  /* clear out all techniques */
  for(tech=db[thing].tech;tech;tech=tnext) {
    tnext=tech->next;
    for(mgk=tech->list;mgk;mgk=mgk2) {
      mgk2=mgk->next;
      free(mgk2);
    }
    free(tech);
  }
  db[thing].tech=NULL;

  /* remove equipment definitions */
  for(def=db[thing].equip;def;def=dnext) {
    dnext=def->next;
    for(entry=def->item;entry;entry=eqnext) {
      eqnext=entry->next;
      free(entry);
    }
    free(def);
  }
  db[thing].equip=NULL;

  /* remove parents and defined attributes on children */
  for(ptr=db[thing].children;ptr && *ptr != NOTHING;ptr++) {
    atr_clean(*ptr, thing);
    PULL(db[*ptr].parents, thing);
  }
  for(ptr=db[thing].parents;ptr && *ptr != NOTHING;ptr++)
    PULL(db[*ptr].children, thing);

  CLEAR(db[thing].children);
  CLEAR(db[thing].parents);
  CLEAR(db[thing].relay);
  CLEAR(db[thing].zone);

  /* set the marker to the next object# to create */
  if(thing < next_free)
    next_free=thing;
}

/* check all links for validity. destroy unlinked exits. */
void validate_links()
{
  int i;

  for(i=0;i<db_top;i++)
    if(!Invalid(db[i].link) && Typeof(db[i].link) == TYPE_GARBAGE) {
      if(Typeof(i) == TYPE_EXIT)
        recycle(i);
      else {
        db[i].link=db[db[i].owner].link;
        if(Typeof(db[i].link) == TYPE_GARBAGE)
          db[i].link=PLAYER_START;
      }
    }
}

int db_check()
{
  int amt=0, i;

  for(i=0;i<db_top;i++) {
    log_dbref=i;
    amt+=mem_usage(i);
  }
  return amt;
}

/* returns the #dbref of the next object number to be created */
dbref get_upfront()
{
  if(freelist)
    return *freelist;

  while(next_free < db_top) {
    if(Typeof(next_free) == TYPE_GARBAGE)
      break;
    next_free++;
  }

  return next_free;
}

/* moves an object# to the top of the free list */
void do_upfront(dbref player, char *arg1)
{
  dbref thing;

  if(Guest(player)) {
    notify(player, "Guests cannot use that command.");
    return;
  }
  
  if((thing=match_thing(player, arg1, MATCH_ABS|MATCH_DESTROYED)) == NOTHING)
    return;
  if(Typeof(thing) != TYPE_GARBAGE) {
    notify(player, "That object does not exist in the free list.");
    return;
  } if(freelist && *freelist == thing) {
    notify(player, "That object is already at the top of the free list.");
    return;
  }

  PULL(freelist, thing);
  PUSH_F(freelist, thing);
  notify(player, tprintf("Pushed #%d to the front.", thing));
}

/* return a cleaned up object off the free list or NOTHING */
dbref free_get()
{
  dbref newobj;

  /* search freelist for specific objects via @upfront */
  while(freelist) {
    newobj=*freelist;
    PULL(freelist, newobj);

    if(Typeof(newobj) == TYPE_GARBAGE)
      return newobj;
  }

  /* search database for destroyed objects */
  while(next_free < db_top) {
    if(Typeof(next_free) == TYPE_GARBAGE)
      return next_free++;
    next_free++;
  }

  return NOTHING;
}

void do_purge(dbref player)
{
  purge_loop=0;
  purge_notify=player;
  notify(player, tprintf("%s Purge Statistics\304\304 %s", MUD_NAME,
         mktm(now, player)));
  fix_free_list();
  purge_notify=NOTHING;

  notify(player, "");
  notify(player, tprintf("Rooms Checked: %d", numrooms));
  notify(player, tprintf("Rooms Marked Floating: %d", numfloat));
  notify(player, tprintf("Currently Disconnected Rooms: %d", ndisrooms));
  notify(player, tprintf("Exit Tree Depth from %s: %d",
         unparse_object(player, PLAYER_START), treedepth));
  notify(player, tprintf("Next Object on Freelist: #%d", get_upfront()));
  notify(player, "");
  notify(player, tprintf("Objects Now Destroyed: %d", ndestroyed));
  notify(player, tprintf("Objects Scheduled for Destruction: %d", numgoing));
  notify(player, tprintf("Total Inactive Objects: %d", ninactive));
}

/* called at game startup before any queue commands */
void destroy_old_objs()
{
  dbref thing;

  /* destroy unsaved objects and unlinked exits */
  for(thing=0;thing<db_top;thing++)
    if((Typeof(thing) == TYPE_EXIT && db[thing].link == NOTHING) ||
       (IS(thing, TYPE_THING, THING_UNSAVED) &&
        Typeof(db[thing].location) != TYPE_PLAYER))
      db[thing].flags |= GOING;

  fix_free_list();
}

/* rebuild free list and count disconnected rooms */
void fix_free_list()
{
  struct goingobj *ptr, *next, *last=NULL;
  char buf[2100];
  dbref thing;

  ndestroyed=0;
  ninactive=0;
  numgoing=0;

  if(!game_online) {
    /* purge all objects previously scheduled for destruction */
    for(thing=0;thing<db_top;thing++) {
      if(Typeof(thing) == TYPE_GARBAGE)
        ++ninactive;
      else if(db[thing].flags & GOING) {
        recycle(thing);
        ++ninactive;
        ++ndestroyed;
      }
    }
  } else {
    /* destroy objects scheduled for destruction */
    for(ptr=goinglist;ptr;ptr=next) {
      next=ptr->next;
      if(now >= ptr->when) {
        if(ptr->obj != NOTHING) {
          recycle(ptr->obj);
          ++ndestroyed;
        }
        free(ptr);
        if(last)
          last->next=next;
        else
          goinglist=next;
      } else
        last=ptr;
    }

    /* count inactive objects */
    for(thing=0;thing<db_top;thing++)
      if(Typeof(thing) == TYPE_GARBAGE)
        ++ninactive;
      else if(db[thing].flags & GOING)
        ++numgoing;
  }

  if(ndestroyed)
    validate_links();

  if(purge_loop++ % 10)
    return;

  /* Now to count disconnected rooms (every 10 minutes). */

  ndisrooms=0;
  numrooms=0;
  numfloat=0;
  treedepth=0;

  /* first, mark all rooms connected to PLAYER_START and #0 */
  dbmark(PLAYER_START);
  if(PLAYER_START != 0)
    dbmark(0);

  /* next, mark all floating rooms and exits from them */
  for(thing=0;thing<db_top;thing++)
    if(IS(thing, TYPE_ROOM, ROOM_FLOATING)) {
      dbmark(thing);
      ++numfloat;
    }

  /* now, inform owners of disconnected rooms */
  for(thing=0;thing<db_top;thing++) {
    if(Typeof(thing) != TYPE_ROOM)
      continue;

    if(db[thing].flags & ROOM_MARKED) {
      db[thing].flags &= ~ROOM_MARKED;
      ++numrooms;
    } else if(!(db[thing].flags & GOING)) {
      if(game_online) {
        if(purge_notify != NOTHING) {
          if(!ndisrooms) {
            notify(purge_notify, "");
            notify(purge_notify, "Disconnected Rooms:");
            notify(purge_notify, "");
          }
          strcpy(buf, unparse_object(purge_notify, thing));
          sprintf(buf+strlen(buf),"  [owner: %s]", unparse_object(purge_notify,
                  db[thing].owner));
          notify(purge_notify, buf);
        } else
          notify(db[thing].owner, tprintf(
              "You own a disconnected room, %s(#%d)", db[thing].name, thing));
      }
      ++ndisrooms;
    }
  }
}

/* stack-based check for disconnected rooms */
void dbmark(dbref loc)
{
  struct dbmark {
    struct dbmark *next;
    dbref i;
  } *stack=NULL, *ptr;

  int depth=0;
  dbref link;  

  if(Invalid(loc) || Typeof(loc) != TYPE_ROOM || (db[loc].flags & ROOM_MARKED))
    return;

  /* initial room entry */
  stack=malloc(sizeof(struct dbmark));
  stack->next=NULL;
  stack->i=db[loc].exits;

  db[loc].flags |= ROOM_MARKED;

  /* enter tree search */
  while(stack) {
    if(stack->i != NOTHING) {  /* New exit to check */
      link=db[stack->i].link;
      stack->i=db[stack->i].next;
      if(!Invalid(link) && Typeof(link) == TYPE_ROOM &&
         !(db[link].flags & ROOM_MARKED)) {
        ptr=malloc(sizeof(struct dbmark));
        ptr->next=stack;
        ptr->i=db[link].exits;
        stack=ptr;
        if(++depth > treedepth)
          treedepth=depth;
        db[link].flags |= ROOM_MARKED;
      }
    } else {  /* Backtrack a level and continue to next exit */
      ptr=stack->next;
      free(stack);
      stack=ptr;
      --depth;
    }
  }
}
