/* db/rule.c */
/* procedures for loading/saving configuration entries */

#include "externs.h"

struct townlist *towns=NULL;
struct chapter *chapters=NULL;
struct lquest *levelquests=NULL;


/* Field Data - Saved even while combat is turned off */

void save_towns(FILE *f, int rule)
{
  struct townlist *ptr;
  int count=0;

  if(!towns)
    return;
  putchr(f, rule);

  /* count number of town sectors to be saved */
  for(ptr=towns;ptr;ptr=ptr->next,count++);
  putnum(f, count, 2);

  for(ptr=towns;ptr;ptr=ptr->next) {
    putstring(f, ptr->name);
    putnum(f, ptr->bitpos, 2);
    putref(f, ptr->zone);
  }
}

int load_towns(FILE *f)
{
  struct townlist *new, *last=NULL;
  int i, count=getnum(f, 2);

  for(i=count;i;i--) {
    new=malloc(sizeof(struct townlist));
    new->next=NULL;
    getstring(f, new->name);
    new->bitpos=getnum(f, 2);
    new->zone=getref(f);

    if(last)
      last->next=new;
    else
      towns=new;
    last=new;
  }

  return count;
}

void save_chapters(FILE *f, int rule)
{
  struct chapter *ptr;
  int count=0;

  if(!chapters)
    return;
  putchr(f, rule);

  /* count number of chapters to be saved */
  for(ptr=chapters;ptr;ptr=ptr->next,count++);
  putnum(f, count, 2);

  for(ptr=chapters;ptr;ptr=ptr->next) {
    putchr(f, ptr->num);
    putstring(f, ptr->name);
  }
}

int load_chapters(FILE *f)
{
  struct chapter *new, *last=NULL;
  int i, count=getnum(f, 2);

  for(i=count;i;i--) {
    new=malloc(sizeof(struct chapter));
    new->next=NULL;
    new->num=fgetc(f);
    getstring(f, new->name);

    if(last)
      last->next=new;
    else
      chapters=new;
    last=new;
  }

  return count;
}

void save_levelquests(FILE *f, int rule)
{
  struct lquest *ptr;
  int count=0;

  if(!levelquests)
    return;
  putchr(f, rule);

  /* count number of chapters to be saved */
  for(ptr=levelquests;ptr;ptr=ptr->next,count++);
  putnum(f, count, 2);

  for(ptr=levelquests;ptr;ptr=ptr->next) {
    putnum(f, ptr->level, 2);
    putnum(f, ptr->bitpos, 2);
    putchr(f, ptr->flags);
    putstring(f, ptr->name);
    putstring(f, ptr->desc);
  }
}

int load_levelquests(FILE *f)
{
  struct lquest *new, *last=NULL;
  int i, count=getnum(f, 2);

  for(i=count;i;i--) {
    new=malloc(sizeof(struct lquest));
    new->next=NULL;
    new->level=getnum(f, 2);
    new->bitpos=getnum(f, 2);
    new->flags=fgetc(f);
    getstring(f, new->name);
    getstring(f, new->desc);

    if(last)
      last->next=new;
    else
      levelquests=new;
    last=new;
  }

  return count;
}


/* Temporary Data - Unsaved outside of combat system */

void save_parties(FILE *f, int rule) {}
int load_parties(FILE *f) {
  int a, size;

  while((size=getnum(f, 2))) {
    getstring_noalloc(f);
    getlong(f);
    getnum(f, 4);
    fgetc(f);

    for(a=0;a<size;a++) {
      getref(f);
      getnum(f, 4);
      fgetc(f);
    }
  }

  return 0;
}
