/* db/save.c */

#include "externs.h"

#define DB_VERSION 1	// TinyMare Compressed Ram v1

extern int db_flags;


void putref(FILE *f, dbref num)
{
  if(num == HOME)
    putnum(f, db_top, dbref_len);
  else if(num < 0)
    fputc(0xFF, f);
  else
    putnum(f, num, dbref_len);
}

void putstring(FILE *f, char *s)
{
  if(s)
    fputs(s, f);
  fputc(0, f);
}


dbref db_write(FILE *f, char *filename)
{
  int i;
  
  /* determine database modifier flags */
  db_flags=0;
  if(sizeof(long) == 8)
    db_flags |= DB_LONGINT;

  /* write database header */
  fprintf(f, "MARE");
  putchr(f, DB_VERSION);
  putnum(f, db_top, 4);
  putnum(f, db_flags, 4);
  putchr(f, NUM_POWS);

  /* select database reference number byte-length */
  dbref_len=(db_top < 0xFF)?1:(db_top < 0xFF00)?2:(db_top < 0xFF0000)?3:4;

  /* save configuration */
  for(i=0;db_rule[i].type;i++)
    db_rule[i].save(f, db_rule[i].type);
  fputc(0, f);

  /* write database */
  for(i=0;i<db_top;i++) {
    if(Typeof(i) == TYPE_GARBAGE)
      continue;
    db_write_object(f, i);
  }

  /* write closing -1 as end of list marker. check for out-of-disk errors. */
  if(fputc(0xFF, f) == EOF) {
    perror(filename);
    return 0;
  }

  return db_top;
}

void db_write_object(FILE *f, dbref i)
{
  ALIST *attr;
  eqtech *eqp;
  eqlist *entry;
  char *ptr, *max;
  int a, b;
  
  /* save object header */
  putref(f, i);
  putstring(f, db[i].name);
  putref(f, db[i].location);
  putref(f, db[i].contents);
  putref(f, db[i].exits);
  putref(f, db[i].link);
  putref(f, db[i].next);
  putref(f, db[i].owner);
  putnum(f, db[i].flags, 4);
//putnum(f, db[i].plane, 4);
  putnum(f, db[i].pennies, 4);

  if(db_flags & DB_LONGINT) {
    putlong(f, db[i].create_time);
    putlong(f, db[i].mod_time);
  } else {
    putnum(f, db[i].create_time, 4);
    putnum(f, db[i].mod_time, 4);
  }

  /* save object lists */
  putlist(f, db[i].parents);
  putlist(f, db[i].children);
  putlist(f, db[i].relay);
  putlist(f, db[i].zone);

  /* write out special player information */
  if(Typeof(i) == TYPE_PLAYER) {
    fputc(db[i].data->class-1, f);
    fputc(db[i].data->rows, f);
    fputc(db[i].data->cols, f);
    fputc((db[i].data->tzdst << 6) + ((db[i].data->tz & 0x1F) << 1) +
          db[i].data->gender, f);
    putnum(f, db[i].data->term, 2);
    putnum(f, db[i].data->steps, 4);
    putnum(f, db[i].data->sessions, 4);
    putnum(f, db[i].data->age, 4);

    if(db_flags & DB_LONGINT) {
      putlong(f, db[i].data->last);
      putlong(f, db[i].data->lastoff);
    } else {
      putnum(f, db[i].data->last, 4);
      putnum(f, db[i].data->lastoff, 4);
    }

    fwrite(db[i].data->pass, (NUM_POWS+3)/4+10, 1, f);
  }

  /* write bitmap to file */
  ptr=db[i].bitmap;
  max=ptr+db[i].bitmaplen;
  while(max > ptr && !*(max-1))
    max--;
  putnum(f, max-ptr, 2);
  while(ptr < max)
    fputc(*(ptr++), f);

  /* write techniques */
  for(a=0,eqp=db[i].tech;eqp;eqp=eqp->next)
    if(eqp->list && ++a == 255)
      break;
  putchr(f, a);

  for(a=0,eqp=db[i].tech;eqp;eqp=eqp->next) {
    if(!eqp->list)
      continue;
    if(++a > 255)
      break;

    putchr(f, eqp->type);
    for(b=0,entry=eqp->list;entry;entry=entry->next,b++);
    putnum(f, b, 2);  /* number of techniques in table */

    for(entry=eqp->list;entry;entry=entry->next) {
      putnum(f, entry->num, 2);
      putchr(f, entry->level);
      putnum(f, entry->value, 4);
    }
  }

  /* write the attribute list */
  for(attr=db[i].attrs;attr;attr=attr->next) {
    fputc(attr->num, f);
    putref(f, attr->obj);
    putstring(f, attr->text);
  }
  fputc(0, f);

  put_atrdefs(f, i);
}

void putlist(FILE *f, dbref *list)
{
  int a=0, *ptr;

  /* save list count */
  for(ptr=list;ptr && *ptr != NOTHING;ptr++,a++);
  putref(f, a-1);

  for(ptr=list;ptr && *ptr != NOTHING;ptr++)
    putref(f, *ptr);
}

void put_atrdefs(FILE *f, dbref thing)
{
  ATRDEF *k;

  for(k=db[thing].atrdefs;k;k=k->next) {
    putchr(f, k->atr->num);
    putnum(f, k->atr->flags, 2);
    putstring(f, k->atr->name);
  }
  fputc(0, f);
}
