/* game/command.c */

#include "externs.h"

extern char log_text[8192];

struct cmdlist {
  char *name;
  void (*cmd)();
  unsigned int arg:5;
  unsigned int alias:1;
  char perms;
  char *opt;
};


/* list the commands */
/* A=#args.. 0=none parsed, 1=arg1 parsed, 2=arg2 parsed, 3=argv parsed,
             &4=combat mode only, &8=direct mode only, &16=don't parse arg2. */
/* B=alias.. 0=command typed in full, 1=alias for command */

struct cmdlist lcmd[]={
  /* command name         function name         A,B     perms */
  {"+c",		do_com,			2,1,	0},
  {"+ch",		do_channel,		1,1,	0},
  {"+chan",		do_channel,		1,1,	0},
  {"+channel",		do_channel,		1,0,	0},
  {"+co",		do_com,			2,1,	0},
  {"+com",		do_com,			2,0,	0},
  {"+f",		do_finger,		1,1,	0, "brief"},
  {"+finger",		do_finger,		1,0,	0, "brief"},
  {"+i",		do_info,		1,1,	0},
  {"+in",		do_info,		1,1,	0},
  {"+inf",		do_info,		1,1,	0},
  {"+info",		do_info,		1,0,	0},
  {"+l",		do_laston,		1,1,	0},
  {"+la",		do_laston,		1,1,	0},
  {"+last",		do_laston,		1,1,	0},
  {"+laston",		do_laston,		1,0,	0},
  {"+m",		do_mail,		18,1,	0},
  {"+ma",		do_mail,		18,1,	0},
  {"+mail",		do_mail,		18,0,	0},
  {"+mo",		do_motd,		0,1,	0},
  {"+motd",		do_motd,		0,0,	0},
  {"+t",		do_term,		2,1,	0},
  {"+term",		do_term,		2,0,	0},
  {"+tz",		do_tzone,		1,1,	0},
  {"+tzone",		do_tzone,		1,0,	0},
  {"+u",		do_uptime,		0,1,	0},
  {"+up",		do_uptime,		0,1,	0},
  {"+uptime",		do_uptime,		0,0,	0},
  {"+v",		do_version,		0,1,	0},
  {"+ver",		do_version,		0,1,	0},
  {"+vers",		do_version,		0,1,	0},
  {"+version",		do_version,		0,0,	0},
  {"@add",		do_addparent,		2,1,	0},
  {"@addp",		do_addparent,		2,1,	0},
  {"@addparent",	do_addparent,		2,0,	0},
  {"@addr",		do_addrelay,		2,1,	0},
  {"@addrelay",		do_addrelay,		2,0,	0},
  {"@addz",		do_addzone,		2,1,	0},
  {"@addzone",		do_addzone,		2,0,	0},
  {"@boot",		do_boot,		2,0,	POW_BOOT},
  {"@br",		do_broadcast,		0,1,	POW_BROADCAST},
  {"@broadcast",	do_broadcast,		0,0,	POW_BROADCAST},
  {"@capture",		do_capture,		2,0,	POW_CAPTURE},
  {"@cboot",		do_cboot,		1,0,	POW_BOOT},
  {"@cemit",		do_cemit,		2,0,	POW_SPOOF},
  {"@ch",		do_chown,		2,1,	0},
  {"@cho",		do_chown,		2,1,	0},
  {"@chown",		do_chown,		2,0,	0},
  {"@chownall",		do_chownall,		2,0,	POW_CHOWN},
  {"@class",		do_class,		2,0,	POW_SECURITY},
  {"@clone",		do_clone,		2,0,	0},
  {"@config",		do_config,		10,0,	POW_CONFIG},
  {"@console",		do_console,		2,0,	POW_CONSOLE},
  {"@cpage",		do_cpage,		2,0,	POW_CONSOLE},
  {"@crash",		do_crash,		0,0,	POW_SHUTDOWN},
  {"@create",		do_create,		2,0,	0},
  {"@ct",		do_ctrace,		0,1,	POW_CONSOLE},
  {"@ctrace",		do_ctrace,		0,0,	POW_CONSOLE},
  {"@cy",		do_cycle,		2,1,	0},
  {"@cyc",		do_cycle,		2,1,	0},
  {"@cycle",		do_cycle,		2,0,	0},
  {"@db",		do_dbtop,		2,1,	POW_DB},
  {"@dbt",		do_dbtop,		2,1,	POW_DB},
  {"@dbtop",		do_dbtop,		2,0,	POW_DB},
  {"@decompile",	do_decompile,		1,0,	0},
  {"@def",		do_defattr,		2,1,	0},
  {"@defattr",		do_defattr,		2,0,	0},
  {"@del",		do_delparent,		2,1,	0},
  {"@delp",		do_delparent,		2,1,	0},
  {"@delparent",	do_delparent,		2,0,	0},
  {"@delr",		do_delrelay,		2,1,	0},
  {"@delrelay",		do_delrelay,		2,0,	0},
  {"@delz",		do_delzone,		2,1,	0},
  {"@delzone",		do_delzone,		2,0,	0},
  {"@destroy",		do_destroy,		2,0,	0},
  {"@dig",		do_dig,			3,0,	0},
  {"@dump",		do_dump,		0,0,	POW_DB},
  {"@echo",		do_echo,		1,0,	0},
  {"@ed",		do_edit,		3,1,	0},
  {"@edit",		do_edit,		3,0,	0},
  {"@emit",		do_emit,		1,0,	0},
  {"@empower",		do_empower,		2,0,	POW_SECURITY},
  {"@find",		do_find,		1,0,	0},
  {"@flush",		do_flush,		2,0,	POW_CONSOLE},
  {"@fo",		do_force,		2,1,	0},
  {"@for",		do_force,		2,1,	0},
  {"@force",		do_force,		2,0,	0},
  {"@fore",		do_foreach,		2,1,	0},
  {"@foreach",		do_foreach,		2,0,	0},
  {"@halt",		do_halt,		2,0,	0},
  {"@inactivity",	do_inactivity,		1,0,	0},
  {"@input",		do_input,		2,0,	POW_CONSOLE},
  {"@it",		do_iterate,		3,1,	0},
  {"@iter",		do_iterate,		3,1,	0},
  {"@iterate",		do_iterate,		3,0,	0},
  {"@link",		do_link,		2,0,	0},
  {"@list",		do_list,		2,0,	0},
  {"@log",		do_log,			2,0,	POW_SECURITY},
  {"@mem",		do_memory,		1,1,	0},
  {"@memory",		do_memory,		1,0,	0},
  {"@misc",		do_misc,		0,0,	0},
  {"@name",		do_name,		2,0,	0},
  {"@necho",		do_necho,		0,0,	0},
  {"@nemit",		do_nemit,		0,0,	0},
  {"@np",		do_npemit,		2,1,	0},
  {"@npe",		do_npemit,		2,1,	0},
  {"@npem",		do_npemit,		2,1,	0},
  {"@npemit",		do_npemit,		2,0,	0},
  {"@nuke",		do_nuke,		1,0,	POW_NUKE},
  {"@oemit",		do_oemit,		2,0,	0},
  {"@open",		do_open,		2,0,	0},
  {"@oremit",		do_oremit,		2,0,	0},
  {"@output",		do_output,		2,0,	POW_CONSOLE},
  {"@passwd",		do_passwd,		9,0,	0},
  {"@paste",		do_paste,		9,1,	0},
  {"@pcreate",		do_pcreate,		9,0,	POW_PCREATE},
  {"@pe",		do_pemit,		2,1,	0},
  {"@pem",		do_pemit,		2,1,	0},
  {"@pemit",		do_pemit,		2,0,	0},
  {"@pennies",		do_pennies,		2,0,	POW_MONEY},
  {"@plane",		do_plane,		2,0,	POW_WIZATTR},
  {"@poll",		do_poll,		0,0,	0},
  {"@pow",		do_powers,		1,1,	0},
  {"@power",		do_powers,		1,1,	0},
  {"@powers",		do_powers,		1,0,	0},
  {"@pr",		do_print,		1,1,	0},
  {"@pri",		do_print,		1,1,	0},
  {"@print",		do_print,		1,0,	0},
  {"@ps",		do_queue,		1,0,	0},
  {"@pu",		do_push,		1,1,	0},
  {"@purge",		do_purge,		0,0,	POW_DB},
  {"@push",		do_push,		1,0,	0},
  {"@re",		do_remit,		2,1,	0},
  {"@reboot",		do_reboot,		1,0,	POW_SHUTDOWN},
  {"@redef",		do_redefattr,		2,1,	0},
  {"@redefattr",	do_redefattr,		2,0,	0},
  {"@redir",		do_redirect,		2,1,	0},
  {"@redirect",		do_redirect,		2,0,	0},
  {"@remit",		do_remit,		2,0,	0},
  {"@sea",		do_search,		2,1,	0},
  {"@sear",		do_search,		2,1,	0},
  {"@search",		do_search,		2,0,	0},
  {"@set",		do_set,			2,0,	0},
  {"@shutdown",		do_shutdown,		0,0,	POW_SHUTDOWN},
  {"@st",		do_stats,		1,1,	0},
  {"@stat",		do_stats,		1,1,	0},
  {"@stats",		do_stats,		1,0,	0},
  {"@swi",		do_switch,		3,1,	0},
  {"@swit",		do_switch,		3,1,	0},
  {"@switch",		do_switch,		3,0,	0},
  {"@tel",		do_teleport,		2,1,	0},
  {"@teleport",		do_teleport,		2,0,	0},
  {"@term",		do_terminate,		1,1,	POW_CONSOLE},
  {"@terminate",	do_terminate,		1,0,	POW_CONSOLE},
  {"@text",		do_text,		2,0,	0},
  {"@ti",		do_time,		0,1,	0},
  {"@time",		do_time,		0,0,	0},
  {"@tr",		do_trigger,		3,1,	0},
  {"@trig",		do_trigger,		3,1,	0},
  {"@trigger",		do_trigger,		3,0,	0},
  {"@uncapture",	do_uncapture,		1,0,	POW_CAPTURE},
  {"@undef",		do_undefattr,		1,1,	0},
  {"@undefattr",	do_undefattr,		1,0,	0},
  {"@undestroy",	do_undestroy,		1,0,	0},
  {"@unlink",		do_unlink,		1,0,	0},
  {"@unlock",		do_unlock,		1,0,	0},
  {"@up",		do_update,		1,1,	0},
  {"@update",		do_update,		1,0,	0},
  {"@upfront",		do_upfront,		1,0,	0},
  {"@wait",		do_wait,		2,0,	0},
  {"@wemit",		do_wemit,		0,0,	POW_BROADCAST},
  {"@wh",		do_while,		3,1,	0},
  {"@while",		do_while,		3,0,	0},
  {"@wipe",		do_wipeattr,		2,1,	0},
  {"@wipeattr",		do_wipeattr,		2,0,	0},
  {"@wipeout",		do_wipeout,		1,0,	POW_NUKE},
  {"@zemit",		do_zemit,		2,0,	0},
  {"@zone",		do_zone,		2,0,	0},
  {"ann",		do_announce,		0,1,	0},
  {"announce",		do_announce,		0,0,	0},
  {"clear",		do_cls,			0,0,	0},
  {"cls",		do_cls,			0,1,	0},
  {"dr",		do_drop,		1,1,	0},
  {"drop",		do_drop,		1,0,	0},
  {"en",		do_enter,		1,1,	0},
  {"ent",		do_enter,		1,1,	0},
  {"enter",		do_enter,		1,0,	0},
  {"ex",		do_examine,		2,1,	0, "all/brief/tech"},
  {"exam",		do_examine,		2,1,	0, "all/brief/tech"},
  {"examine",		do_examine,		2,0,	0, "all/brief/tech"},
  {"exit",		do_exits,		0,1,	0},
  {"exits",		do_exits,		0,0,	0},
  {"get",		do_take,		1,0,	0},
  {"give",		do_give,		2,0,	0},
  {"go",		do_move,		1,1,	0},
  {"goto",		do_move,		1,0,	0},
  {"gr",		do_grab,		2,1,	0},
  {"grab",		do_grab,		2,0,	0},
  {"h",			do_help,		0,1,	0},
  {"help",		do_help,		0,0,	0},
  {"i",			do_inventory,		0,1,	0},
  {"inv",		do_inventory,		0,1,	0},
  {"inventory",		do_inventory,		0,0,	0},
  {"j",			do_join,		1,1,	POW_JOIN},
  {"join",		do_join,		1,0,	POW_JOIN},
  {"l",			do_look,		1,1,	0},
  {"leave",		do_leave,		0,0,	0},
  {"lo",		do_look,		1,1,	0},
  {"look",		do_look,		1,0,	0},
  {"mo",		do_money,		1,1,	0},
  {"mon",		do_money,		1,1,	0},
  {"money",		do_money,		1,0,	0},
  {"move",		do_move,		1,0,	0},
  {"open",		do_area_open,		1,0,	0},
  {"p",			do_page,		18,1,	0},
  {"page",		do_page,		18,0,	0},
  {"pose",		do_pose_cmd,		0,0,	0},
  {"put",		do_put,			2,0,	0},
  {"run",		do_run,			0,0,	0},
  {"sa",		do_say,			0,1,	0},
  {"say",		do_say,			0,0,	0},
  {"sea",		do_area_search,		0,1,	0},
  {"sear",		do_area_search,		0,1,	0},
  {"search",		do_area_search,		0,0,	0},
  {"su",		do_summon,		1,1,	POW_SUMMON},
  {"summon",		do_summon,		1,0,	POW_SUMMON},
  {"t",			do_talk,		1,1,	0},
  {"ta",		do_take,		1,1,	0},
  {"take",		do_take,		1,0,	0},
  {"talk",		do_talk,		1,0,	0},
  {"to",		do_to,			0,0,	0},
  {"use",		do_use,			3,0,	0},
  {"w",			do_whisper,		1,1,	0},
  {"wh",		do_whisper,		1,1,	0},
  {"where",		do_whereis,		1,1,	0},
  {"whereis",		do_whereis,		1,0,	0},
  {"whi",		do_whisper,		18,1,	0},
  {"whisper",		do_whisper,		18,0,	0},
  {"who",		do_who,			2,0,	0},
  {0}};

/* list all commands available to a player */
void list_commands(dbref player)
{
  char buf[512];
  int a, col=get_cols(NULL, player);

  notify(player, "List of available hardcoded commands in alphabetical order:");
  notify(player, "");
  strcpy(buf, "");
  for(a=0;lcmd[a].name;a++) {
    if(lcmd[a].alias || (lcmd[a].perms && !power(player, lcmd[a].perms)))
      continue;
    if(*buf && strlen(buf)+strlen(lcmd[a].name) >= col-2) {
      notify(player, buf);
      strcpy(buf, "");
    }
    if(*buf)
      strcat(buf, ", ");
    strcat(buf, lcmd[a].name);
  }
  if(*buf)
    notify(player, buf);
  notify(player, "");
}

/* match a list of things for events */
static int list_check(thing, player, type, str)
dbref thing, player;
int type;
char *str;
{
  int match=0;

  DOLIST(thing, thing)
    if((thing == player || Typeof(thing) != TYPE_PLAYER) &&
       atr_match(thing, player, type, str))
      match=1;

  return match;
}

/* match a command word in a ^ event on specified object */
int cmd_match(player, cmd, name, argv)
dbref player;
char *cmd, *name, **argv;
{
  dbref thing;
  int a;
  
  if((thing=match_thing(player, name, MATCH_QUIET|(power(player, POW_REMOTE)?
     MATCH_EVERYTHING:MATCH_RESTRICT))) == NOTHING)
    return 0;

  /* set environment %0 through %9 */
  for(a=0;a<10 && argv[a+1];a++)
    strcpy(wptr[a], argv[a+1]);

  return atr_match(thing, player, '^', cmd);
}

/* routine to check for $ and ! events and execute them */
int atr_match(thing, player, type, str)
dbref thing, player;
char type; /* a symbol not modified by compress(). '$', '^' or '!' */
char *str; /* string to match */
{
  struct all_atr_list *ptr, *anext;
  char event[8192], *r, *s, *p;
  int match=0, locked;

  if(db[thing].flags & HAVEN)
    return 0;

  /* Does the player pass the @ulock? */
  if((locked=!could_doit(player, thing, A_ULOCK)) && type == '!')
    return 0;

  for(ptr=all_attributes(thing);ptr;free(ptr),ptr=anext) {
    anext=ptr->next;
    /* don't check unimportant attributes */
    if(!ptr->type || *ptr->value != type || (ptr->type->flags &
       (AF_LOCK|AF_FUNC|AF_DATE|AF_BITMAP|AF_DBREF|AF_HAVEN)))
      continue;

    /* search for first unescaped : */
    r=event;
    for(s=uncompress(ptr->value+1);*s && *s != ':';s++) {
      if(*s == '\\' && *(s+1) == ':')
        s++;
      *r++ = *s;
    }
    if(!*s++)
      continue;
    *r='\0';

    /* make sure event matches */
    if((type == '^' && strcasecmp(str, event)) ||
       (type != '^' && !wild_match(str, event, 0, 1)))
      continue;

    /* search for a locked attribute */
    if(*s == '/') {
      p=++s;
      while(*s && (*s != '/')) {
        if(*s == '[')
          while(*s && (*s != ']'))
            s++;
        s++;
      }
      if(!*s)
        continue;
      *s++=0;

      /* check lock */
      if(!eval_boolexp(player, thing, p))
        continue;
    }

    /* is object locked against the player? */
    if(locked) {
      /* Only show 'No such command' message if obj has no @ufail/@aufail */
      if(type != '!' &&
         (*atr_get(thing, A_UFAIL) || *atr_get(thing, A_AUFAIL))) {
          ansi_did_it(player,thing,A_UFAIL,NULL,A_OUFAIL,NULL,A_AUFAIL);
          return 1;
      }
      return 0;
    }

    parse_que(thing,s,player);
    match=1;
  }

  return match;
}

/* prints a random error message from the file "typo.txt" */
static void error_message(dbref player)
{
  static char **errmsgs;
  static int errsize=0;

  FILE *f;
  char buf[1024], *s, **ptr;
  int a, size=0;

  /* Objects are notified of an error in their command syntax. */
  if(Typeof(player) != TYPE_PLAYER) {
    notify(player, "Unrecognizable command.");
    return;
  }

  if(!errmsgs || !errsize) {
    errsize=0;
    if(!(f=fopen("msgs/typo.txt", "r")))
      perror("msgs/typo.txt");
    else {
      while(ftext(buf, 1024, f)) {
        if((s=strchr(buf, '#')))
          *s='\0';
        if(!*buf)
          continue;

        if(errsize >= size) {
          size += 25;
          if(!(ptr=realloc(errmsgs, size * sizeof(char *)))) {
            perror("error_message");
            for(a=0;a<errsize;a++)
              free(errmsgs[a]);
            free(errmsgs);

            errmsgs=NULL;
            errsize=0;
            break;
          }
          errmsgs=ptr;
        }

        SPTR(errmsgs[errsize++], buf);
      }
      fclose(f);
    }
  }

  if(!errsize)
    notify(player, "Unrecognizable command. Type 'help' for help.");
  else {
    sprintf(wptr[0], "%d", nstat[NS_NFCMDS]);
    notify(player, parse_function(GOD, player, errmsgs[rand_num(errsize)]));
  }
}

void process_command(player, command, cause, direct)
dbref player, cause;
char *command;
int direct;
{
  struct cmdlist *cp; /* command pointer */
  static hash *cmdhash=NULL;
  char cmdmain[8192], buff[8192], buff2[8192], aarg2[8192];
  char *option, *cmdarg, *r, *s, *pure, *arg1, *arg2;
  char *argv[100]={0};
  int a, match, argopt=0, *ptr, argc;

  /* secure permissions */
  if(Invalid(player) || Invalid(cause) || (player == GOD && cause != GOD))
    return;
  if(Invalid(db[player].location))
    return;

  /* write debugging log strings */
  strcpy(log_text, command);
  log_dbref=player;
  dberrcheck=0;

  /* Command Layout Structure:
   *
   * cmdarg pure                               No Arguments
   * cmdarg arg1                               1 Argument
   * cmdarg arg1=arg2                          2 Arguments
   * cmdarg arg1=argv1,argv2,..,argv100        Many Arguments
   * cmdarg/option ...                         Command Option
   */

  /* initialize command hashtable */
  if(!cmdhash)
    cmdhash=make_hashtab(lcmd, sizeof(struct cmdlist));

  /* erase leading whitespace */
  while(*command && isspace(*command))
    command++;
  if(!*command)
    return;

  /* erase following whitespace */
  s=strchr(command, '\0');
  while(isspace(*(--s)))
    *s=0;

  /* get pure command strings */
  if((pure=strchr(command,' ')))
    pure++;
  else
    pure="";

  /* reset memory allocation list */
  func_zerolev();
  speaker=player;

  /* calculate command counter */
  nstat[NS_NCMDS]++;
  if(direct)
    nstat[NS_NPCMDS]++;
  else
    nstat[NS_NQCMDS]++;
  
  /* attempt log of commands for all objects except #0 and #1 */
  if(player > 1) {
    static int counter=0;
    FILE *command_log;

    if((command_log=fopen("logs/commands","a"))) {
      fprintf(command_log, "%s(%d): %s\n",
              db[player].name, player, command);
      fclose(command_log);
    } else
      perror("logs/commands");

    /* max 150 commands in current log. should be average of last 10 *
     * seconds of activity on TinyMare, depending on # users online. */

    if(++counter >= 150) {
      counter=0;
      rename("logs/commands","logs/commands~");
    }

    /* is playerlog turned on? */
    if(direct && match_word(PLAYERLOG, db[player].name)) {
      if((command_log=fopen("logs/playerlog","a"))) {
        fprintf(command_log, "%s(%d): %s\n", db[player].name, player, command);
        fclose(command_log);
      } else
        perror("logs/playerlog");
    }

    /* log guest accounts */
    if(Guest(player)) {
      if((command_log=fopen("logs/guestlog","a"))) {
        fprintf(command_log, "%s(%d): %s\n", db[player].name, player, command);
        fclose(command_log);
      } else
        perror("logs/guestlog");
    }
  }

  if((db[player].flags & VERBOSE) && IS(db[player].owner, TYPE_PLAYER,
     PLAYER_CONNECT))
    raw_notify(db[player].owner, tprintf("%s>> %s", db[player].name, command));

  /* remark (does nothing) */
  if(command[0] == '@' && command[1] == '@')
    return;

  /* check for single-character commands */
  switch(*command) {
    case '"': do_say(player, NULL, NULL, NULL, command+1); return;
    case '\'': do_to(player, NULL, NULL, NULL, command+1); return;
    case ':': do_pose(player, command+1, 0); return;
    case ';': do_pose(player, command+1, 1); return;
    case '#': instant_force(player, atoi(command+1), pure, cause); return;
    case '=': do_com(player, "", command+1); return;
    case '\\': room_emote(player, NOTHING, NULL, command+1, 0); return;
  }

  /* does the command match an exit alias or direction of movement? */
  if((Typeof(player) == TYPE_PLAYER || Typeof(player) == TYPE_THING)) {
    /* check cardinal directions */
    if(match_word("home down d up u north n northeast ne east e southeast se south s southwest sw west w northwest nw", command)) {
      do_move(player, command);
      return;
    }

    /* check exit aliases */
    if(match_thing(player, command, MATCH_EXIT|MATCH_QUIET|MATCH_INVIS) !=
       NOTHING) {
      do_move(player, command);
      return;
    }
  }

  strcpy(cmdmain, command);
  s=cmdmain;

  /* cmdarg=first word */
  cmdarg=parse_up(&s, ' ');

  /* determine if any switches are used */
  if((option=strchr(cmdarg, '/')))
    *option++='\0';

  /* find command in hash table */
  cp=lookup_hash(cmdhash, cmdarg);

  /* argument 1 & 2 */
  if(cp && (cp->arg & 3) < 2) {  // Zero or one arguments.
    arg1=parse_perfect(&s, 0);
    arg2=NULL;
  } else {
    arg1=parse_perfect(&s, '=');
    if(*s) {
      arg2=s--;
      /* remove following space before = */
      while(isspace(*(--s)))
        *s=0;
      s=arg2;
    } else
      arg2=NULL;
  }

  if((cp = lookup_hash(cmdhash, cmdarg))) {
    /* check permissions/direct mode */
    if(cp->perms && !power(player, cp->perms))
      cp=NULL;
    else if((cp->arg & 8) && !direct) {
      notify(player, "Command available in direct mode only.");
      return;
    }

    /* get any command-line switches */
    if(option) {
      if(cp->opt) {
        strcpy(buff2, cp->opt);
        s=buff2;
        while((r=parse_up(&s,'/'))) {
          argopt++;
          if(string_prefix(r, option))
            break;
        }
      }
      if(!cp->opt || !r) {
        notify(player, tprintf("Unrecognized switch '%s' for command '%s'.",
               option, cmdarg));
        return;
      }
    }
  }

  /* process argument 1 */
  if(arg1)
    strcpy(buff, arg1);
  else
    *buff=0;
  if(arg1 && (!cp || (cp->arg&3) > 0)) {
    if(arg1)
      evaluate(&arg1, buff, player, cause, 0);
    arg1=arg1?buff:"";
  } else
    for(arg1=buff;*arg1 && *arg1 == ' ';arg1++);

  /* process argument 2 */
  if(arg2)
    strcpy(aarg2, arg2);
  else
    *aarg2=0;
  if(arg2 && (!cp || (!(cp->arg & 16) && (cp->arg&3) > 1))) {
    if(arg2)
      evaluate(&arg2, aarg2, player, cause, 0);
    arg2=arg2?aarg2:"";
  } else
    for(arg2=aarg2;*arg2 && *arg2 == ' ';arg2++);

  for(a=1;a<100;a++) {
    if(!(r=parse_perfect(&s,',')))
      break;
    if(!cp || (cp->arg&3) == 3) {
      evaluate(&r, buff2, player, cause, 0);
      SPTR(argv[a], buff2);
    } else
      SPTR(argv[a], r);
  }
  argc=a;

  if(!cp) {
    /* remove leading space after = */
    while(*arg2 && *arg2 == ' ')
      arg2++;
  }

  /* process command */
  if(cp) {
    cp->cmd(player,arg1,arg2,argv,pure,cause,argopt);
    for(a=1;a<argc;a++)
      free(argv[a]);
    return;
  }

  /* check for '@attr obj=value' setting */
  if(*cmdarg == '@' || *cmdarg == '&' || *cmdarg == '$')
    if(test_set(player, cmdarg, arg1, arg2)) {
      for(a=1;a<argc;a++)
        free(argv[a]);
      return;
    }

  /* leave aliases */
  if(Typeof(db[player].location) == TYPE_THING &&
     exit_alias(db[player].location, A_LALIAS, command)) {
    do_leave(player);
    for(a=1;a<argc;a++)
      free(argv[a]);
    return;
  }

  /* enter aliases */
  DOLIST(match, db[db[player].location].contents)
    if(Typeof(match) == TYPE_THING && exit_alias(match, A_EALIAS, command)) {
      do_enter(player, tprintf("#%d", a));
      for(a=1;a<argc;a++)
        free(argv[a]);
      return;
    }

  /* Match a command word event */
  if(*arg1 && cmd_match(player, cmdarg, arg1, argv)) {
    for(a=1;a<argc;a++)
      free(argv[a]);
    return;
  }

  for(a=1;a<argc;a++)
    free(argv[a]);

  /* Try matching user defined commands. First check the universal zones,
   * which take precedence over any player-defined command. Second, check all
   * room zones and nearby objects, executing double occurances if found. */

  for(match=0,ptr=uzo;ptr && *ptr != NOTHING;ptr++)
    match |= atr_match(*ptr, player, '$', command);
  if(match)
    return;

  /* check room zones */
  for(ptr=db[mainroom(player)].zone;ptr && *ptr != NOTHING;ptr++) {
    if(db[*ptr].flags & ZONE_UNIVERSAL)
      continue;
    match |= atr_match(*ptr,player,'$',command);
  }

  /* check all nearby objects */
  match |= atr_match(db[player].location,player,'$',command);
  match |= list_check(db[db[player].location].contents,player,'$',command);
  match |= list_check(db[db[player].location].exits,player,'$',command);
  match |= list_check(db[player].contents,player,'$',command);

  if(match)
    return;

  /* No match found. Print a random error message. */

  nstat[NS_NFCMDS]++;
  error_message(player);

  /* log failed command */
  if(LOG_FAILED_CMDS && Typeof(player) != TYPE_PLAYER) {
    FILE *command_log;

    if((command_log=fopen("logs/playerlog","a"))) {
      fprintf(command_log, "FAILED %s(%d): %s\n",
              db[player].name, player, command);
      fclose(command_log);
    } else
      perror("logs/playerlog");
    if(game_online)
      log_main(tprintf("FAILED %s(#%d): %s", db[player].name,
               player, command));
  }

  /* notify object's owner */
  if(Typeof(player) != TYPE_PLAYER && !(db[player].flags & PUPPET))
    notify(db[player].owner, tprintf("Failed Command on %s: %s",
           unparse_object(db[player].owner, player), command));
}
