/* game/queue.c */

#include "externs.h"

int num_waitcmds=0;

struct bque *qfirst=NULL, *qlast=NULL;
struct bque *qwait=NULL, *qwaitlast=NULL;


void immediate_que(player, cmd, thing)
dbref player, thing;
char *cmd;
{
  char buff[strlen(cmd)+1], *r, *s=buff;

  /* check that the player can do a queue command */
  if((db[player].flags & (GOING|HAVEN)) || Typeof(player) == TYPE_GARBAGE)
    return;

  strcpy(buff, cmd);
  while((r=ansi_parse_up(&s,';')))
    parse_immediate(player,r,thing);
}

void parse_immediate(player, cmd, thing)
dbref player, thing;
char *cmd;
{
  char buf[8192];

  func_zerolev();
  pronoun_substitute(buf, player, thing, cmd);
  process_command(player, buf, thing, 0);
}

void parse_que(player, text, cause)
dbref player, cause;
char *text;
{
  struct bque *new;
  char buff[strlen(text)+1], *cmd, *s=buff;
  int a=0;

  if((db[player].flags & (GOING|HAVEN)) || Typeof(player) == TYPE_GARBAGE)
    return;

  strcpy(buff, text);

  /* add commands to queue */
  while((cmd=ansi_parse_up(&s,';'))) {
    /* make sure player can afford to do it */
    if(!power(player, POW_QUEUE)) {
      if(db[player].queue >= QUEUE_QUOTA) {
        db[player].flags |= HAVEN;
        notify(db[player].owner, tprintf("Run away object: %s",
               unparse_object(db[player].owner, player)));
        return;
      } if(!db[db[player].owner].pennies &&
           !charge_money(player, rand_num(QUEUE_COST)?0:1, 0)) {
        notify(db[player].owner, tprintf(
               "Not enough money to queue command on %s.",
               unparse_object(db[player].owner, player)));
        return;
      }
    }

    /* create pointer reference */
    new=malloc(sizeof(struct bque));
    new->next=NULL;
    new->player=player;
    new->cause=cause;
    for(a=0;a<10;a++)
      SPTR(new->env[a], wptr[a]);
    SPTR(new->cmd, cmd);

    /* add to the queue list */
    if(qlast) {
      qlast->next=new;
      qlast=new; 
    } else
      qlast=qfirst=new;

    db[player].queue++;
  }
}

void do_wait(player, arg1, command, argv, pure, cause)
dbref player;
char *arg1, *command, **argv, *pure;
dbref cause;
{
  struct bque *new;
  int a, wait=atoi(arg1);

  if(db[player].flags & (GOING|HAVEN))
    return;

  /* execute command immediately if no time is specified */
  if(wait < 1) {
    parse_que(player, command, cause);
    return;
  } else if(wait > 31536000)  /* 1 year maximum */
    wait=31536000;

  /* make sure player can afford to do it */
  if(!power(player, POW_QUEUE)) {
    if(db[player].queue >= QUEUE_QUOTA) {
      db[player].flags |= HAVEN;
      notify(db[player].owner, tprintf("Run away object: %s",
             unparse_object(db[player].owner, player)));
      return;
    } if(!db[db[player].owner].pennies) {
      notify(db[player].owner,tprintf("Not enough money to queue @wait on %s.",
             unparse_object(db[player].owner, player)));
      return;
    }
  }

  if(num_waitcmds >= MAX_WAIT_COMMANDS) {
    notify(db[player].owner,tprintf("%s: Too many @wait commands in progress.",
           unparse_object(db[player].owner, player)));
    return;
  }

  /* create pointer reference */
  new=malloc(sizeof(struct bque));
  new->next=NULL;
  new->player=player;
  new->cause=cause;
  new->wait=now+(long)wait;
  for(a=0;a<10;a++)
    SPTR(new->env[a], wptr[a]);
  SPTR(new->cmd, command);

  /* add to the @wait queue */
  if(qwaitlast) {
    qwaitlast->next=new;
    qwaitlast=new; 
  } else
    qwaitlast=qwait=new;

  db[player].queue++;
  num_waitcmds++;
}

/* execute one round of queue commands */
void flush_queue()
{
  struct bque *ptr;
  int a, i, num=list_queue(bque, qfirst);
  char buff[8192];

  if(!num)
    return;

  nstat[NS_QUEUE]++;
  for(i=0;i<num && qfirst;i++) {
    ptr=qfirst;
    if(!(qfirst=ptr->next))
      qlast=NULL;
    if(ptr->player != NOTHING) {
      db[ptr->player].queue--;
      if(!(db[ptr->player].flags & (GOING|HAVEN))) {
        for(a=0;a<10;a++)
          strcpy(wptr[a], ptr->env[a]);
        func_zerolev();
        pronoun_substitute(buff, ptr->player, ptr->cause, ptr->cmd);
        process_command(ptr->player, buff, ptr->cause, 0);
      }
    }

    /* remove pointer */
    for(a=0;a<10;a++)
      free(ptr->env[a]);
    free(ptr->cmd);
    free(ptr);
  }

  for(a=0;a<10;a++)
    *wptr[a]='\0';
}

/* called every second to check for wait queue commands */           
void wait_second()
{
  struct bque *ptr, *next, *last=NULL;
  int a;

  for(ptr=qwait;ptr;ptr=next) {
    next=ptr->next;
    if(ptr->wait > now && !(db[ptr->player].flags & (GOING|HAVEN))) {
      last=ptr;
      continue;
    }

    /* process @wait command */
    db[ptr->player].queue--;
    num_waitcmds--;
    if(!(db[ptr->player].flags & (GOING|HAVEN))) {
      for(a=0;a<10;a++)
        strcpy(wptr[a], ptr->env[a]);
      parse_que(ptr->player, ptr->cmd, ptr->cause);
    }

    /* remove from queue */
    if(last)
      last->next=ptr->next;
    else
      qwait=ptr->next;
    for(a=0;a<10;a++)
      free(ptr->env[a]);
    free(ptr->cmd);
    free(ptr);
  }
  qwaitlast=last;
}

/* tell player what commands they have pending in the queue */
void do_queue(player,arg1,arg2)
dbref player;
char *arg1, *arg2;
{
  struct bque *tmp;
  dbref thing=NOTHING;
  int count, tcount=0;
  
  if(strcmp(arg2, ""))
    if((thing=match_thing(player, arg2, MATCH_EVERYTHING)) == NOTHING)
      return;

  if(string_prefix("immediate", arg1) || !strcmp(arg1, "") ||
     !strcasecmp(arg1, "all")) {
    count=0;
    for(tmp=qfirst;tmp;tmp=tmp->next)
    if(!Invalid(tmp->player) && (db[tmp->player].owner == db[player].owner ||
       controls(player, tmp->player, POW_QUEUE)) && (thing ==
       db[tmp->player].owner || thing == NOTHING || thing == tmp->player)) {
   if(!count)
     notify(player, "\e[1mImmediate commands:\e[m");
   notify(player,tprintf("%s:%s",unparse_object(player,tmp->player),tmp->cmd));
   count++; tcount++;
   }
   if(count > 0)
     notify(player, tprintf("-- Total: %d Immediate --", count));
  }
  if(string_prefix("waited", arg1) || !strcmp(arg1, "") ||
     !strcasecmp(arg1, "all")) {
    count=0;
    for(tmp=qwait;tmp;tmp=tmp->next)
    if(!Invalid(tmp->player) && (db[tmp->player].owner == db[player].owner ||
       controls(player, tmp->player, POW_QUEUE)) && (thing ==
       db[tmp->player].owner || thing == NOTHING || thing == tmp->player)) {
      if(!count)
        notify(player, "\e[1m@waited commands:\e[m");
      notify(player, tprintf("\e[1;36m[\e[37m%ld\e[36m]\e[m%s:%s", 
             (long)tmp->wait-now,unparse_object(player,tmp->player),tmp->cmd));
      count++; tcount++;
    }
    if(count > 0)
      notify(player, tprintf("-- Total: %d @waited --", count));
  }

  if(!tcount)
    notify(player, "Nothing in the queue.");
}

/* remove all queued commands from a certain player */
void do_halt(player, arg1, ncom, argv, pure, cause)
dbref player, cause;
char *arg1, *ncom, **argv, *pure;
{
  int thing=player;

  if(!strcmp(arg1, "ALL") && power(player, POW_SECURITY))
    thing=AMBIGUOUS;
  else if(*arg1) {
    if((thing=match_thing(player, arg1, MATCH_EVERYTHING)) == NOTHING)
      return;
    if(!controls(player, thing, POW_MODIFY)) {
      notify(player, "Permission denied.");
      return;
    }
  }

  if(!Quiet(player)) {
    if(thing == player)
      notify(player, "Everything halted.");
    else if(thing == AMBIGUOUS)
      notify(player, "Halted away every blessed thing.");
    else
      notify(player, tprintf("%s halted.", unparse_object(player, thing)));
  }

  /* remove the queue commands */
  halt_object(thing);

  /* execute extra command? */
  if(*ncom && !Invalid(thing))
    parse_que(thing,ncom,cause);
}

void halt_object(dbref thing)
{
  struct bque *ptr, *next, *last=NULL;
  int a;

  /* remove regular queue commands */
  for(ptr=qfirst;ptr;ptr=ptr->next)
    if(ptr->player != NOTHING && (thing == AMBIGUOUS || ptr->player == thing ||
       db[ptr->player].owner == thing)) {
      db[ptr->player].queue--;
      ptr->player=NOTHING;
    }

  /* remove wait queue commands */
  for(ptr=qwait;ptr;ptr=next) {
    next=ptr->next;
    if(ptr->player != thing && db[ptr->player].owner != thing &&
       thing != AMBIGUOUS) {
      last=ptr;
      continue;
    }

    db[ptr->player].queue--;
    num_waitcmds--;

    /* remove from queue */
    if(last)
      last->next=ptr->next;
    else
      qwait=ptr->next;
    for(a=0;a<10;a++)
      free(ptr->env[a]);
    free(ptr->cmd);
    free(ptr);
  }
  qwaitlast=last;
}


void save_semaphores(FILE *f, int rule)
{
}

int load_semaphores(FILE *f)
{
  while(getref(f) != NOTHING) {
    fgetc(f);
    getnum(f, 2);
    getnum(f, 4);
    getstring_noalloc(f);
  }
  return 0;
}
