/* game/unparse.c */

#include "externs.h"

int disp_exit=0;
int rainbow=0;
int infravision=0;


struct flagstr flaglist[]={
  {"Abode",       'A', TYPE_ROOM,   ROOM_ABODE},
  {"Builder",     'B', TYPE_PLAYER, PLAYER_BUILD},
  {"Chown_Ok",    'C', NOTYPE,      CHOWN_OK},
  {"Dark",        'D', NOTYPE,      DARK},
  {"Exit",        'E', TYPE_EXIT,   0},
  {"Floating",    'F', TYPE_ROOM,   ROOM_FLOATING},
  {"Going",       'G', NOTYPE,      GOING},
  {"Healing",     'H', TYPE_ROOM,   ROOM_HEALING},
  {"Illness",     'I', TYPE_ROOM,   ROOM_ILLNESS},
  {"Jump_Ok",     'J', TYPE_ROOM,   ROOM_JUMP_OK},
  {"Link_Ok",     'L', TYPE_ROOM,   ROOM_LINK_OK},
  {"Monster",     'M', TYPE_THING,  THING_MONSTER},
  {"Nest",        'N', TYPE_ROOM,   ROOM_NEST},
  {"Oceanic",     'O', TYPE_ROOM,   ROOM_OCEANIC},
  {"Player",      'P', TYPE_PLAYER, 0},
  {"Room",        'R', TYPE_ROOM,   0},
  {"Shaft",       'S', TYPE_ROOM,   ROOM_SHAFT},
  {"Transparent", 'T', TYPE_EXIT,   EXIT_TRANSPARENT},
  {"Universal",   'U', TYPE_ZONE,   ZONE_UNIVERSAL},
  {"Visible",     'V', NOTYPE,      VISIBLE},
  {"Wall",        'W', TYPE_ROOM,   ROOM_WALL},
  {"X-Zone",      'X', TYPE_ROOM,   ROOM_XZONE},
  {"Zone",        'Z', TYPE_ZONE,   0},
  {"Arena",       'a', TYPE_ROOM,   ROOM_ARENA},
  {"Bearing",     'b', NOTYPE,      BEARING},
  {"Connected",   'c', TYPE_PLAYER, PLAYER_CONNECT},
  {"Destroy_Ok",  'd', TYPE_THING,  THING_DESTROY_OK},
  {"Enter_Ok",    'e', TYPE_THING,  THING_ENTER_OK},
  {"Fishable",    'f', TYPE_ROOM,   ROOM_FISHABLE},
  {"Grab_Ok",     'g', TYPE_THING,  THING_GRAB_OK},
  {"Haven",       'h', NOTYPE,      HAVEN},
  {"Inactive_Ok", 'i', TYPE_PLAYER, PLAYER_INACTIVE_OK},
  {"Light",       'l', NOTYPE,      LIGHT},
  {"Mount",       'm', TYPE_THING,  THING_MOUNT},
  {"Opaque",      'o', NOTYPE,      OPAQUE},
  {"Puppet",      'p', NOTYPE,      PUPPET},
  {"Quiet",       'q', NOTYPE,      QUIET},
  {"Restricted",  'r', TYPE_ZONE,   ZONE_RESTRICTED},
  {"Slaved",      's', TYPE_PLAYER, PLAYER_SLAVE},
  {"Terse",       't', TYPE_PLAYER, PLAYER_TERSE},
  {"Unsaved",     'u', TYPE_THING,  THING_UNSAVED},
  {"Verbose",     'v', NOTYPE,      VERBOSE},
  {"Wander",      'w', TYPE_THING,  THING_WANDER},
  {0}};

void list_flags(dbref player)
{
  char buf[512];
  int a, b=0, count=0, rows, cols=(get_cols(NULL, player)+2)/20;

  if(cols < 2)
    cols=4;
  notify(player, "List of available flags in alphabetical order:");
  notify(player, "");

  /* print list in column-oriented display */
  for(;flaglist[count].name;count++);
  rows=(count+cols-1)/cols;
  for(a=0;a<rows;a++) {
    *buf='\0';
    for(b=0;b<cols;b++) {
      sprintf(buf+strlen(buf), "%s%-15.15s %c", *buf?"   ":"",
              flaglist[rows*b+a].name, flaglist[rows*b+a].flag);
      if(rows*(b+1)+a >= count) {
        notify(player, buf);
        break;
      }
    }
  }
  notify(player, "");
  notify(player, tprintf("Total %d flags.", count));
}

/* prints flag list; typ=0: Print flag letters only
                     typ=1: Print entire words of all non-objtype flags */
char *unparse_flags(player, thing, typ)
dbref player, thing;
int typ;
{
  static char buf[512];
  char *p=buf;
  int a;
  
  if(!typ) {
    switch(Typeof(thing)) {
      case TYPE_PLAYER: *p++ = 'P'; break;
      case TYPE_EXIT:   *p++ = 'E'; break;
      case TYPE_ROOM:   *p++ = 'R'; break;
      case TYPE_ZONE:   *p++ = 'Z'; break;
    }
  } else
    *buf='\0';

  for(a=0;flaglist[a].name;a++)
    if((flaglist[a].objtype == NOTYPE || flaglist[a].objtype == Typeof(thing))
       && flaglist[a].bits && (db[thing].flags & flaglist[a].bits)) {

      /* Don't show 'Connected' flag if player is hidden! */
      if(flaglist[a].objtype == TYPE_PLAYER && flaglist[a].bits ==
         PLAYER_CONNECT && !(could_doit(player, thing, A_HLOCK) ||
         controls(player, thing, POW_WHO)))
        continue;

      if(!typ)
        *p++ = flaglist[a].flag;
      else {
        if(*buf)
          strcat(buf, " ");
        strcat(buf, flaglist[a].name);
      }
    }

  if(!typ)
    *p='\0';
  return buf;
}

// Return a color code for printing an object on the screen.
//   type & 1=Darken intensity on text (mainly for object captions).
//   type & 2=Increase intensity on text (mainly for +com messages).
//   type & 4=Remove special color effects (blink/rainbow) for printing text.
//   type & 8=Suppress combat color adjustments.
//   type & 16=Object scanned under infravision.
int unparse_color(player, obj, type)
dbref player, obj;
int type;
{
  char buf[8192], *r, *s=buf;
  int color=7, field=0, bright=0;

  /* determine if player should be shown the bright or dark room color */
  bright=((type & 8) || !IS(db[player].location,TYPE_ROOM,DARK)) && TIMEOFDAY;

  /* determine @color */
  pronoun_substitute(buf, obj, player, atr_get(obj, A_COLOR));
  if(!*buf && Typeof(obj) == TYPE_EXIT && !Invalid(db[obj].link))
    pronoun_substitute(buf, db[obj].link, player, atr_get(db[obj].link,
                       A_COLOR));

  /* parse @color string */
  while((r=parse_up(&s, ','))) {
    ++field;

    /* match day & night */
    if(bright && (field == 2 || field > 6))
      continue;

    color=color_code(7, r);
  }

  /* Darken */
  if(type & 1) {
    if((color & 15) == 7)
      color++;
    else if((color & 15) > 8)
      color-=8;
  }

  /* Brighten */
  if(type & 2) {
    if(color & 0x0800)
      color&=~0x0800;
    else if((color & 15) < 8)
      color+=8;
    else if((color & 15) == 8)
      color--;
  }

  /* Channel/Description Text Modifier */
  if(type & 4)
    color&=~0x1ff0;

  return color;
}

/* Display an object */
char *unparse_object(dbref player, dbref loc)
{
  static char buf[1024];
  int num, idle, color;
  char *r;

  /* predefined names for non-existing objects */
  if(loc == NOTHING)
    return "*NOTHING*";
  else if(loc == AMBIGUOUS)
    return "*AMBIGUOUS*";
  else if(loc == HOME)
    return "*HOME*";
  else if(loc < -3 || loc >= db_top)
    return "*UNDEFINED*";

  /* determine color of object */
  color=unparse_color(player, loc, 0);
  sprintf(buf, "%s%s", (color & 1024)?"":textcolor(default_color, color),
          db[loc].name);

  /* display any exit aliases */
  if(disp_exit == 2 && Typeof(loc) == TYPE_EXIT && *(r=atr_get(loc, A_ALIAS)))
    sprintf(buf+strlen(buf), ";%s", r);

  /* rename object to rainbow colors if requested */
  if(color & 1024) {
    strcpy(buf, rainbowmsg(buf, color, 980));
    color=default_color;
  }

  /* check if we need to print #dbref or idle time */
  num = (!disp_exit && (!loc || controls(player,loc,POW_BACKSTAGE) ||
         IS(loc, TYPE_ROOM, ROOM_JUMP_OK|ROOM_ABODE|ROOM_LINK_OK) ||
         db[loc].flags & (CHOWN_OK|VISIBLE)));
  idle=(IS(loc, TYPE_PLAYER, PLAYER_CONNECT) && getidletime(loc) > 300 &&
        (could_doit(player, loc, A_HLOCK) || controls(player, loc, POW_WHO)));

  /* print #dbref and/or idle time and restore default color */

  if(num || idle) {
    strcat(buf, textcolor((color & 1024)?default_color:color,
           (color == 8)?7:(color|8)));
    if(idle)
      sprintf(buf+strlen(buf), "(%s)",time_format(TMS,(long)getidletime(loc)));
    if(num)
      sprintf(buf+strlen(buf), "(#%d%s)", loc, unparse_flags(player, loc, 0));
    strcat(buf, textcolor((color == 8)?7:(color|8), default_color));
  } else if(!(color & 1024))
    strcat(buf, textcolor(color, default_color));

  return buf;
}

char *unparse_exitname(player, loc, alias)
dbref player, loc;
int alias;
{
  static char buf[1024];

  disp_exit=alias?2:1;
  strcpy(buf, unparse_object(player, loc));
  disp_exit=0;

  return buf;
}

char *unparse_objheader(player, thing, hidden)
dbref player, thing;
int hidden;
{
  return (hidden == 2)?" \e[m(hidden)":"";
}

char *unparse_caption(player, thing, hidden)
dbref player, thing;
int hidden;
{
  static char buff[8192];
  char buf[8192], *title="";
  int color;

  /* display @caption, if available */
  strcpy(buff, atr_get(thing, A_CAPTION));
  pronoun_substitute(buf, thing, player, buff);
  if(!*buf && Typeof(thing) == TYPE_PLAYER && Guest(thing))
    strcpy(buf, "explores the world");
  if(!power(thing, POW_SPOOF))
    buf[80]='\0';
  strcat(buf, unparse_objheader(player, thing, hidden));

  /* determine caption color */
  color=unparse_color(player, thing, 5+(hidden == 3 ? 16:0));

  /* print result output */
  strcpy(buff, unparse_object(player, thing));
  strcat(buff, textcolor(default_color, color));
  sprintf(buff+strlen(buff), "%s%s%s%s%s", *title?" the ":"", *title?title:"",
          *buf?" ":"", buf, textcolor(color, default_color));
  return buff;
}
