/* hdrs/db.h */

extern dbref speaker;
extern int dberrcheck;
extern int boot_flags;
extern int mare_test_db;
extern int default_color;
extern dbref log_dbref;
extern long now;

extern FILE *db_read_file;
extern char *db_file;
extern char *wptr[10];
extern int dbref_len;

#define Typeof(x) (db[x].flags & 0x7)
#define IS(thing,type,flag) ((Typeof(thing)==type) && (db[thing].flags&(flag)))

struct flagstr {
  char *name;
  unsigned int flag:8;
  char objtype;
  int bits;
};

extern struct flagstr flaglist[];


/* Object type definitions */

#define TYPE_ROOM 	0
#define TYPE_THING 	1
#define TYPE_EXIT 	2
#define TYPE_PLAYER	3
#define TYPE_ZONE	4
#define TYPE_GARBAGE	7

#define NOTYPE		-1	// no particular type

/* Miscellaneous Flags (Any Object Type) */
#define CHOWN_OK	0x20
#define DARK		0x40
#define VERBOSE		0x100
#define	HAVEN		0x400
#define GOING		0x4000
#define PUPPET		0x20000
#define VISIBLE		0x100000
#define OPAQUE		0x800000
#define QUIET		0x1000000
#define LIGHT		0x4000000
#define BEARING		0x8000000

#define NOTYPE_FLAGMASK	0xd934560    // the above flags OR'd together

/* Object flags */
#define THING_MOUNT       0x8
#define THING_UNSAVED     0x80
#define THING_DESTROY_OK  0x200
#define THING_WANDER      0x800
#define THING_ENTER_OK    0x8000
#define THING_MONSTER     0x400000
#define THING_GRAB_OK     0x2000000

/* Exit flags */
#define EXIT_TRANSPARENT   0x8
#define EXIT_DOOR          0x10

/* Player flags */
#define PLAYER_BUILD       0x10
#define PLAYER_SLAVE       0x80
#define PLAYER_CONNECT     0x200
#define PLAYER_INACTIVE_OK 0x800
#define PLAYER_TERSE       0x400000

/* Room flags */
#define ROOM_XZONE	0x8
#define ROOM_OCEANIC	0x10
#define ROOM_WALL	0x80
#define ROOM_JUMP_OK	0x200
#define ROOM_HEALING	0x800
#define ROOM_FLOATING	0x1000
#define ROOM_ILLNESS	0x2000
#define ROOM_NEST	0x8000
#define ROOM_LINK_OK	0x40000
#define ROOM_ARENA	0x80000
#define ROOM_FISHABLE	0x200000
#define ROOM_SHAFT	0x400000
#define ROOM_ABODE	0x2000000

/* Zone flags */
#define ZONE_UNIVERSAL	0x10
#define ZONE_RESTRICTED	0x80

/* Internal (unsaved) flags */
#define ROOM_MARKED	0x80000000  // Check for disconnected rooms only


/* special dbref's */
#define GOD		 1	// ultimate wizard character
#define NOTHING		-1	// null dbref
#define AMBIGUOUS	-2	// multiple possibilities, for matchers
#define HOME		-3	// virtual room, represents mover's home

/* macros for gender substitutions */
enum gender_types { MALE, FEMALE, NEUTER, PLURAL };
enum pronoun_types { SUBJ=0, OBJN=4, POSS=8, APOSS=12 };

/* writes new string to pointer */
#define WPTR(ptr, str) ({ char *_str=(str); \
                          free(ptr); \
                          strcpy(ptr=malloc(strlen(_str)+1), _str); })

/* sets a new pointer regardless of previous content */
#define SPTR(ptr, str) ({ char *_str=(str); \
                          strcpy(ptr=malloc(strlen(_str)+1), _str); })

#define getstring(x,p) SPTR(p, getstring_noalloc(x))

#define putchr(f,num) fputc(num, f)

/* writes a number of <len> bytes to a file regardless of big/little endian */
#define putnum(f, num, len)     ({ int _x=num, _i=len-1; for(;_i >= 0;_i--) \
                                   fputc((_x >> (_i << 3)), f); })

/* reads a number of <len> bytes from file. NOTE: Does not work for 8-byte! */
#define getnum(f, len)		({ int _x=0, _i=len; for(;_i > 0;_i--) \
                                   _x=(_x << 8)+getc(f); _x; })

/* writes a 8-byte long integer to file */
#define putlong(f, num)   ({ quad _x=num; int _i=7; for(;_i >= 0;_i--) \
			     fputc((_x >> (_i << 3)), f); })

/* reads a 8-byte long integer from a file */
#define getlong(f)        ({ quad _x=0L; int _i=8; for(;_i > 0;_i--) \
			     _x=(_x << 8)+getc(f); _x; })


#define atr_get(thing, num) atr_get_obj(thing, NOTHING, num)
#define atr_add(thing, num, str) atr_add_obj(thing, NOTHING, num, str)

#define class(x) (db[x].data->class)
#define terminfo(x,y) (db[db[x].owner].data->term & (y))

#define iscombat(x) (db[x].bitmaplen)

#define power(x,pow) controls(x, NOTHING, pow)
#define could_doit(x,y,attr) eval_boolexp(x, y, atr_get(y, attr))
#define notify_except(loc,thing,msg) notify_room(loc, thing, NOTHING, msg)
#define atr_parse(x,y,attr) parse_function(x, y, atr_get(x, attr))
#define get_zone(player) ({ dbref *_zone=db[mainroom(player)].zone; \
                            _zone ? *_zone : NOTHING; })

#define Invalid(x) (x < 0 || x >= db_top)
#define Immortal(x) (db[db[x].owner].data->class >= CLASS_IMMORTAL)
#define Guest(x) (db[db[x].owner].data->class == CLASS_GUEST)
#define Builder(x) (IS(db[x].owner, TYPE_PLAYER, PLAYER_BUILD) || \
                    (power(db[x].owner, POW_BUILD)))
#define Quiet(x) ((db[x].flags & QUIET) || (Typeof(x) != TYPE_PLAYER && \
                  !(db[x].flags & PUPPET)))

#define ipaddr(x) ({ struct in_addr _addr; _addr.s_addr=x; inet_ntoa(_addr); })

#define ISTRUE(x) (*(x) && strncmp(x, "#-", 2) && (!isdigit(*(x)) || atoi(x)))

/* game startup flags */
#define BOOT_RESTART	0x1	// Hard reset the game during a reboot
#define BOOT_QUIET	0x2	// Suppress log_main from being displayed
#define BOOT_VALIDATE	0x4	// Do not save +party data and semaphores

/* database format versions */
#define F_MUSH		1	// TinyMush
#define F_MUSE		2	// TinyMuse
#define F_MUSE_ZONED	3	// TinyMuse Zoned
#define F_MARE		4	// TinyMare Compressed RAM
#define F_MUSH3		5	// TinyMush-3

/* tinymare db flags */
#define DB_LONGINT	0x0001	// Time values stored as 8 bytes instead of 4

/* tinymush db flags */
#define MUSH_ZONE	0x0001	// ZONE/DOMAIN field
#define MUSH_LINK	0x0002	// LINK field (exits from objs)
#define MUSH_GDBM	0x0004	// attrs are in a gdbm db, not here
#define MUSH_ATRNAME	0x0008	// NAME is an attr, not in the hdr
#define MUSH_ATRKEY	0x0010	// KEY is an attr, not in the hdr
#define MUSH_PARENT	0x0020	// db has the PARENT field
#define MUSH_ATRMONEY	0x0080	// Money is kept in an attribute
#define MUSH_XFLAGS	0x0100	// An extra word of flags
#define MUSH_POWERS	0x0200	// Powers?
#define MUSH_3FLAGS	0x0400	// Adding a 3rd flag word
#define MUSH_QUOTED	0x0800	// Quoted strings, ala PennMUSH

/* attribute structure */
typedef struct {
  char *name;		  // full name of attribute
  dbref obj;		  // object in which attr is defined
  unsigned char num;	  // attr number in cache (1-255)
  unsigned int flags:16;  // any flags the attribute holds (listed below)
} ATTR;

/* attribute flags */
#define AF_OSEE    0x0001 // players other than owner can see it
#define AF_DARK    0x0002 // wizards or owner of object defined on can see
#define AF_WIZ     0x0004 // only wizards can change it
#define AF_UNIMP   0x0008 // not important -- don't save it in the db
#define AF_HIDDEN  0x0010 // changed and viewed only by the high almighty
#define AF_DATE    0x0020 // date stored in universal longint form
#define INH        0x0040 // value inherited by children
#define AF_LOCK	   0x0080 // interpreted as a boolean expression
#define AF_FUNC    0x0100 // this is a user defined function
#define AF_HAVEN   0x0200 // can't execute commands on this attribute
#define AF_BITMAP  0x0400 // an attribute expressed in hexadecimal form
#define AF_DBREF   0x0800 // calls unparse_object on integer

extern char *attr_options[16];

/* external attribute list */
enum attributes {
#define DOATTR(num, var, name, flags) var=num,
#include "attrib.h"
#undef DOATTR
};


/* attribute list structure */
struct alist {
  struct alist *next;
  dbref obj;
  unsigned char num;
  char text[0];
};

struct atrdef {
  struct atrdef *next;
  ATTR *atr;
};

typedef struct atrdef ATRDEF;
typedef struct alist ALIST;


/* equipment definition structures */

struct item_entry {
  struct item_entry *next;
  unsigned int num:16;		// Spell or item number (1 to 9999)
  unsigned int level:8;		// Spell level, item quantity. (Equipment = 0)
  int value;			// Spell mastery %, special argument.
};

struct item_table {
  struct item_table *next;
  unsigned int type:8;		// Table type specifier.
  struct item_entry *list;	// Pointer to item list above.
};

typedef struct item_entry eqlist;
typedef struct item_table eqtech;

struct equip {
  struct equip *next;
  char *name;			// Equipment name
  char *desc;			// Full Description (compressed)

  unsigned int num:16;		// Inventory number (10000-65535)
  unsigned int type:8;		// Item type (wear/wield position)
  unsigned int density:8;	// Weapon/armor material composition

  unsigned int value;		// Shop value
  unsigned int str:16;		// Attack/defense power
  unsigned int maxstr:16;	// Equipment strength at 100%

  unsigned int weight:16;       // Weight of item in grams
  unsigned int color:8;		// Equipment color
  unsigned int flags:8;		// Special item options

  unsigned int element:16;	// Elemental enchantment
  unsigned int damage:8;	// Equipment destruction percentage

  unsigned int atkcond;         // Condition modification on target

  eqlist *item;			// Container contents (item list)
  unsigned int *eqfx;		// Equipment enchantments
};

typedef struct equip eqdef;

struct playerdata {
  unsigned int class:8;
  unsigned int rows:8;
  unsigned int cols:8;

  int tz:5;
  unsigned int tzdst:1;
  unsigned int gender:1;

  unsigned int term:16;
  int steps;
  int sessions;
  int age;

  long last;
  long lastoff;

  unsigned char pass[10];
  unsigned char powers[0];
};
  
/* database reference object data */
struct object {
  char *name;			// Name of object
  dbref location;		// Pointer to container
  dbref contents;		// Pointer to first item
  dbref exits;			// Pointer to first exit for rooms
  dbref link;			// Pointer to home/destination
  dbref next;			// Pointer to next in contents/exits chain
  dbref owner;			// Who controls this object

  unsigned int flags;		// Bitmap flag list
  int plane;			// Dimension plane setting
  int pennies;			// Money value

  unsigned int queue:16;	// Number of commands in queue on object
  unsigned int bitmaplen:16;	// Size of bitmap string below
  char *bitmap;			// Battle bitmap strings

  eqtech *tech;			// Inventory & Technique tables
  eqdef *equip;			// Equipment definitions

  ALIST *attrs;			// Attribute data list
  ATRDEF *atrdefs;		// User-defined attribute list

  dbref *parents;		// Parents list
  dbref *children;		// Children list
  dbref *relay;			// Relay list
  dbref *zone;			// Zone list

  struct playerdata *data;	// Additional data (see below)

  long mod_time;		// Date of modification
  long create_time;		// Date of creation
};

// The 'data' is a (long) value which represents different things based on
// the type of object stored in the database.
//
// Room:   Exploration database roomnumber.
// Player: Pointer to (struct playerdata) containing player info/password.

extern struct object *db;
extern dbref db_top;
extern dbref *uzo;

/* macros for object lists */

#define DOLIST(var, first) \
  for((var) = (first); (var) != NOTHING; (var) = db[(var)].next)
#define PUSH(thing, locative) \
  ((db[(thing)].next = (locative)), (locative) = (thing))

#define PUSH_F(list,value)	push_first(&list,value)
#define PUSH_L(list,value)	push_last(&list,value)
#define PULL(list,value)	pull_list(&list,value)
#define CLEAR(list)		clear_list(&list)

/* attribute list structure */
struct all_atr_list {
  struct all_atr_list *next;
  unsigned int numinherit:8;
  ATTR *type;
  char *value;
};

/* TinyMare-specific db configuration rules */
extern struct db_rule {
  int type;
  char *text;
  int (*load)();
  void (*save)();
} db_rule[];

/* command hashtable structures */
struct hashent {
  struct hashent *ptr;
};

typedef struct hashent hash;
