/* hdrs/queue.h */

/* Standard Command Queue */
struct bque
{
  struct bque *next;
  dbref player;		/* player who will do command */
  dbref cause;		/* player causing command (for %N) */
  long wait;		/* time of execution on a @wait */
  char *env[10];	/* environment, from wild match */
  char *cmd;		/* command text */
};

extern struct bque *qfirst,*qlast,*qwait;
extern struct bque *qlfirst,*qllast;


/* Combat System Player Entry */
struct fightque
{
  struct fightque *next;
  dbref player;           /* the player attacking */
  dbref enemy;            /* targeted enemy.  NOTHING=dead battle/magic only */

  unsigned int round:8;   /* consecutive round number */
  unsigned int timer:8;   /* 1/10ths seconds remaining before event triggers */

  unsigned int num:16;    /* technique number, or 0 if regular fight */
  unsigned int level:8;   /* additional background technique level */
  unsigned int ftimer:8;  /* background timer indicating vigor remaining */
  unsigned int stage:8;   /* fight sequence stage information */
  unsigned int cmdtyp:16; /* flags specifying type of attack being done */
  char *arg;              /* command-line argument (for spells, etc) */
};

extern struct fightque *qfight;


/* Semaphore Queue */
struct semaque
{
  struct semaque *next;
  dbref player;		/* target of semaphore */
  unsigned char type;	/* types listed below.. */
  unsigned int left:16;	/* seconds before semaphore triggers */
  int value;		/* value of semaphore corresponding to type */
  char *text;		/* text to be displayed to player */
};

extern struct semaque *qsema;

/* Semaphore Types */
#define SEMA_DISPLAY	0	// Display message only
#define SEMA_DISP_LOC	1	// Display message in location only
#define SEMA_OEMIT	2	// Display to room except <value>
#define SEMA_TELEPORT	3	// Teleport to a room number
#define SEMA_COND_ON	4	// Turn combat condition on
#define SEMA_COND_OFF	5	// Turn combat condition off
#define SEMA_SBAR	6	// Modify status bar output
#define SEMA_EVENT	7	// Trigger special routine event
#define SEMA_FLASH	8	// Screen flash with alert bell
#define SEMA_FLASH_OFF	9	// Restore inverse screen
#define SEMA_BOSS	10	// Boss explosion
#define SEMA_REVIVIFY	11	// Completely restore maximum HP (Divine)
#define SEMA_DEATHSEQ	12	// Animated death sequence

#define NUM_SEMAPHORES	13


/* Timer Structure */
struct timerint {
  char *name;
  int *interval;
  int ticks;
  void (*fun)();
};

extern struct timerint timerval[];


/* counts the number of entries in a queue */
#define list_queue(type,queue) ({ int _amt=0; struct type *_q; \
                                  for(_q=queue;_q;_q=_q->next,_amt++); \
                                  _amt; })
