/* hdrs/world.h */
/* World Geography Information */


/* Sector Terrain Types */
enum sector_terrain {
  SECTOR_NORMAL, SECTOR_TOWN, SECTOR_WILDERNESS, SECTOR_DUNGEON, SECTOR_CASTLE,
  SECTOR_UNDERWATER, SECTOR_MOUNTAIN, SECTOR_SWAMP, SECTOR_ARID, SECTOR_SHIP,
  SECTOR_SHRINE, SECTOR_CAVERNS, SECTOR_TOWER, SECTOR_ETHERAL, NUM_SECTORS };

/* Sector Flags */
#define SF_OUTSIDE      0x1     // Night/Day differences, Weather Messages
#define SF_DUNGEON      0x2     // Dungeon messages apparant in rooms
#define SF_TELEPORT     0x4     // Can teleport from this area
#define SF_ESCAPE       0x8     // Can use escape spell to get out
#define SF_DARK         0x10    // Rooms become dark during nightfall
#define SF_PLANT        0x20    // Plants can be found here
#define SF_MINERAL	0x40	// Minerals can be found here
#define SF_UNDERWATER	0x80	// Spells muted, Fire elemental ineffectual

/* Wilderness Sector Tables */
struct sectorlist {
  char *name;
  char *feeling;	// Adjective printed on 'score'
  int color;		// Status Bar icon color code
  int regen;		// Regeneration rate (0-12)
  int encounter;	// Enemy encounter frequency (0-7)
  int avgtemp;		// Average sector temperature
  int flags;		// Sector flags
};

extern struct sectorlist sector[NUM_SECTORS];


/* Village, Town, & Shrine List */
struct townlist {
  struct townlist *next;
  char *name;
  int bitpos;
  int zone;
};

extern struct townlist *towns;

/* Storyline Chapter List */
struct chapter {
  struct chapter *next;
  int num;
  char *name;
};

extern struct chapter *chapters;


/* Levelquest Flags */
#define LQ_REQUIRED	0x1	// Require completion to attain race levelup
#define LQ_NOEXP	0x2	// Does not grant any experience points
#define LQ_LOCKEXP	0x4	// EXP gained cannot go beyond 0 EXP for Next
#define LQ_GLOBAL	0x8	// Can be completed on any level whatsoever

/* Levelquest Structure */
struct lquest {
  struct lquest *next;

  unsigned int level:16;
  unsigned int bitpos:16;
  unsigned int flags:8;
  char *name;
  char *desc;
};

extern struct lquest *levelquests;


/**** Weather Module ****/

enum weather_requests {
  WEA_APPARENT, WEA_CURRENT, WEA_AVERAGE, WEA_LAKE, WEA_WINDSPEED, WEA_VISIB };

enum almanac { REVERSE=0x020, INSIDE=0x0080, NIGHT=0x0100,  DAY=0x0200,
               WINTER=0x1000, SPRING=0x2000, SUMMER=0x4000, AUTUMN=0x8000 };

extern int moonphase;
