/* io/console.c */

#include "externs.h"


extern char highfont[256];

/*
 *
 * Reserved for Linux Console Color Device (screen fading via spells/timeofday)

static unsigned char default_rgb[48]={
     0, 0, 0,     170, 0, 0,    0, 170, 0,    170, 85, 0,
     0, 0, 170,   170, 0, 170,  0, 170, 170,  170, 170, 170, 

     85, 85, 85,  255, 85, 85,  85, 255, 85,  255, 255, 85,
     85, 85, 255, 255, 85, 255, 85, 255, 255, 255, 255, 255};

static char tones[16]={0, 4, 6, 2, 3, 5, 7, 8, 1, 10, 12, 14, 9, 11, 13, 15};
 */

static char fadetab[40]={33, 27, 12, 9, 15, 19, 0, 31, 11, 16,
                         22, 34, 14, 4, 29, 17, 37, 20, 10, 38,
                         21, 25, 36, 3, 35, 30, 6, 13, 8, 39,
                         26, 18, 1, 23, 5, 32, 2, 28, 7, 24};

/* gets #cols from session <d>, or player <player> if not connected */
int get_cols(DESC *d, dbref player)
{
  player=db[player].owner;

  if(!d)
    d=match_desc(player);
  if(!d || d->cols < 4)
    return db[player].data->cols;
  return d->cols;
}

/* gets #rows from session <d>, or player <player> if not connected */
int get_rows(DESC *d, dbref player)
{
  player=db[player].owner;

  if(!d)
    d=match_desc(player);
  if(!d || d->rows < 4)
    return db[player].data->rows;
  return d->rows;
}

/* console reset */
void clear_screen(DESC *d, int sbar)
{
  char buf[512];
  int a, b;

  if(Invalid(d->player)) {
    output2(d, "\ec\e(K\e[1;5]\e[2J\e[H\n");
    return;
  }

  /* clear screen without ansi */
  if(!Invalid(d->player) && !terminfo(d->player, TERM_ANSI)) {
    b=get_rows(d, d->player);
    for(a=1;a<b;a++)
      buf[a-1]='\n';
    buf[a-1]=0;
    output2(d, buf);
    return;
  }

  /* perform ansi reset escape sequences */
  output2(d, tprintf("\e[m\e[r%s\e[1;5]\e[2J\e[%dH\n", (terminfo(d->player,
          TERM_HIGHFONT) || terminfo(d->player, TERM_GRAPHICS))?"\e(K":"",
          sbar?4:1));
}

/* fade away screen with 2x1 character blocks */
void fade_screen(DESC *d)
{
  char buf[256];
  int i, x, y, count=0, cols, rows;

  if(!terminfo(d->player, TERM_ANSI))
    return;
  cols=get_cols(d, d->player);
  rows=get_rows(d, d->player);

  *buf=0;
  for(i=0;i<40;i++)
    for(y=1;y<rows;y++)
      for(x=(fadetab[(fadetab[i]+y)%40]<<1)+1; x<cols; x+=80) {
        sprintf(buf+strlen(buf), "\e[%d;%dH  ", y, x);
        if(++count > 10) {
          output2(d, tprintf("\e7%s\e8\e[A\n", buf));
          *buf=0; count=0;
        }
      }

  if(*buf)
    output2(d, tprintf("\e7%s\e8\e[A\n", buf));
}


/* Terminal Configuration Section */

struct term_config term_config[]={
  {"Rows",	0},
  {"Columns",	0},
  {"Ansi",	TERM_ANSI},
  {"Highfont",	TERM_HIGHFONT},
  {"Graphics",	TERM_GRAPHICS},
  {"Beeps",	TERM_BEEPS},
  {"Sound",	TERM_SOUND},
  {"Wordwrap",	TERM_WORDWRAP},
  {"Reginfo",	TERM_REGINFO},
  {"Protocol",	0},
  {0}};

/* terminal configuration */
void do_term(player, arg1, arg2)
dbref player;
char *arg1, *arg2;
{
  DESC *d;
  int a=0, b;
  dbref thing=db[player].owner;

  /* Match configuration variable */
  if(*arg1) {
    for(a=0;term_config[a].name;a++)
      if(string_prefix(term_config[a].name, arg1))
        break;
    if(!term_config[a].name) {
      if((thing=lookup_player(arg1)) <= NOTHING) {
        notify(player, "Unknown terminal setting.");
        return;
      } else if(!controls(player, thing, POW_EXAMINE)) {
        notify(player, "Permission denied.");
        return;
      }
    }
  }

  d=match_desc(thing);

  /* Display terminal configuration */
  if(!*arg1 || !term_config[a].name) {
    char buf[12];

    notify(player, tprintf("\e[1;32m%s%s Terminal Configuration:\e[m",
           db[thing].name, possname(db[thing].name)));
    for(a=0;term_config[a].name;a++) {
      if(a < 2)
        sprintf(buf, "%d", a?get_cols(d, thing):get_rows(d, thing));
      else {
        if(!term_config[a].flag)
          break;
        strcpy(buf, (db[thing].data->term & term_config[a].flag)?"On":"Off");
      }
      notify(player, tprintf("\e[1;36m%12.12s: \e[37m%s\e[m",
             term_config[a].name, buf));
    }

    if(d) {
      notify(player, "");
      notify(player, "\e[1;32mInferred Terminal Specifications:\e[m");
      notify(player, tprintf("\e[1;36m    Protocol: \e[37m%s\e[m",
             (d->flags & C_TINYFUGUE)?"TinyFugue":"Telnet"));
    }
    return;
  }

  /* Determine correct argument */
  if(a < 2) {
    b=atoi(arg2);
    if(b < 4 || b > 250) {
      notify(player, tprintf("%s may only be between 4 and 250.",
             term_config[a].name));
      return;
    }
  } else if(a != 9) {
    if(!strcasecmp(arg2, "Off"))
      b=0;
    else if(!strcasecmp(arg2, "On"))
      b=1;
    else {
      notify(player, tprintf("%s may either be On or Off.",
             term_config[a].name));
      return;
    }
  } else {
    if(!strcasecmp(arg2, "Telnet"))
      b=0;
    else if(!strcasecmp(arg2, "TinyFugue"))
      b=1;
    else {
      notify(player, tprintf("%s must be one of: Telnet TinyFugue",
             term_config[a].name));
      return;
    }
  }

  /* do additional things if removing ANSI */
  if(a == 2 && !b)
    db[thing].data->term &= ~TERM_SOUND;

  switch(a) {
  case 0 ... 1:
    if(a) {
      db[thing].data->cols=b;
      if(d)
        d->cols=b;
    } else {
      db[thing].data->rows=b;
      if(d)
        d->rows=b;
    }
    notify(player, tprintf("%s set to %d.", term_config[a].name, b));
    break;
  case 2 ... 8:
    if(b)
      db[thing].data->term |= term_config[a].flag;
    else
      db[thing].data->term &= ~term_config[a].flag;
    notify(player, tprintf("%s turned %s.",term_config[a].name, b?"On":"Off"));
    break;
  case 9:
    if(!d) {
      notify(player, "That player is not connected.");
      return;
    }
    if(b)
      d->flags |= C_TINYFUGUE;
    else
      d->flags &= ~C_TINYFUGUE;
    notify(player, tprintf("%s set to %s.", term_config[a].name, b?"TinyFugue":
           "Telnet"));
    break;
  }

  if(a == 1 && d && (b == 80 || b == 132) && terminfo(thing, TERM_ANSI))
    output2(d, tprintf("\e[?3%c\e[A\n", b<132?'l':'h'));
}

/* refixes player's screen */
void do_cls(dbref player)
{
  DESC *d=match_desc(player);

  if(d)
    clear_screen(d, 0);
}


/* player console settings */
void do_console(player, name, options)
dbref player;
char *name, *options;
{
  static struct {
    char *name;
    int bitmask;	// Connection flag bitmask
    int inverse;	// Set flag when the option is disabled
  } cflags[]={
    {"Input",	C_INPUTOFF,	1},
    {"Output",	C_OUTPUTOFF,	1},
    {"Sbar",	C_SBAROFF,	1},
    {0}};

  DESC *d;
  char *r, *s=options;
  int a, inv, thing;

  /* match login descriptor to player */
  if((thing=match_thing(player, name, MATCH_EVERYTHING)) == NOTHING)
    return;
  if(Typeof(thing) != TYPE_PLAYER) {
    notify(player, "Console settings can only be changed for players.");
    return;
  } if(!controls(player, thing, POW_CONSOLE)) {
    notify(player, "Permission denied.");
    return;
  } if(thing == GOD) {
    notify(player, tprintf("Cannot change console options for %s.",
           unparse_object(player, thing)));
    return;
  } if(!(d=match_desc(thing)) || !(d->flags & C_CONNECTED)) {
    notify(player, "That player is not connected.");
    return;
  }

  while((r=parse_up(&s,' '))) {
    if(*r == '!') {
      inv=1;
      r++;
    } else
      inv=0;
    if(!*r)
      continue;

    for(a=0;cflags[a].name;a++)
      if(string_prefix(cflags[a].name, r))
        break;
    if(!cflags[a].name) {
      notify(player, tprintf("No such console option '%s'.", r));
      continue;
    }
    
    if((!(inv ^ cflags[a].inverse) && (d->flags & cflags[a].bitmask)) ||
         ((inv ^ cflags[a].inverse) && !(d->flags & cflags[a].bitmask))) {
      if(!Quiet(player))
        notify(player, tprintf("Option already %sset.", inv?"un":""));
      return;
    }

    if(inv ^ cflags[a].inverse)
      d->flags &= ~cflags[a].bitmask;
    else
      d->flags |= cflags[a].bitmask;

    if(!Quiet(player))
      notify(player, tprintf("Option %sset.", inv?"un":""));
  }
}


/* Text Emulation & Translation */

char *term_emulation(DESC *d, char *msg)
{
  static unsigned char buf[8192];
  int a, b, nobeep=(!Invalid(d->player) && !terminfo(d->player, TERM_BEEPS));

  /* apply wordwrap to output string only in non-graphics mode */
  if(!Invalid(d->player) && terminfo(d->player, TERM_WORDWRAP) &&
     (msg[0] != '\e' || msg[1] == '['))
    msg=wordwrap(d->player, msg);

  /* format message for output */
  *buf=0;
  if(d->flags & C_REMOTE)
    sprintf(buf, "%d ", d->concid);
  a=strlen(buf);
  for(b=0;a<8000 && msg[b];b++) {
    if(msg[b] == '\a' && nobeep)
      continue;
    if(msg[b] == '\n')
      buf[a++]='\r';
    if(msg[b] != '\n' || !(d->flags & C_REMOTE))
      buf[a++]=msg[b];
  }
  if((a == 8000 && msg[b]) || (d->flags & C_REMOTE))
    buf[a++]='\n';
  buf[a]=0;

  /* change 8-bit font to 7-bit for players */
  if(Invalid(d->player) || !terminfo(d->player, TERM_HIGHFONT)) {
    for(a=0;buf[a];a++) {
      /* Telnet IAC */
      b=buf[a];
      if((b == 255 && buf[a+1]) || (b == 251 && buf[a+1] == 1)) {
        a++;
        continue;
      }

      buf[a]=highfont[b];
    }
  }

  if(Invalid(d->player) || terminfo(d->player, TERM_ANSI))
    return buf;

  return ansi_strip(buf);
}

// Silently strips ansi escape sequences from string.
//
char *ansi_strip(char *str)
{
  static char buff[8192];
  char *s, *t=buff;

  for(s=str;*s;s++) {
    if(*s != '\e') {  /* copy everything that isn't an escape character */
      *t++=*s;
      continue;
    }

    /* we found an escape character */
    if(!*(++s))
      break;

    /* 1-digit escape sequence */
    if(*s != '[' && *s > 47)
      continue;
    s++;

    /* 2-digit sequence */
    if(*s < 48) {
      if(!*(++s))
        break;
      continue;
    }

    /* variable length escape code; terminates with a character > 63 */
    while(*s && *s < 64)
      s++;
    if(!*s)
      break;
  }

  *t='\0';
  return buff;
}

// Enable wordwrap around ansi characters.
//
char *wordwrap(dbref player, char *str)
{
  static char buf[8192];
  char *r=buf, *p=NULL, *s=str;
  int wlen=0, llen=0, width=get_cols(NULL, player)-1;

  while(*s && r-buf < 8000) {
    if(*s == '\n') {	// Newline
      *r++=*s++;
      llen=wlen=0;
      p=NULL;
      continue;
    }

    if(*s == '\t') {    // Pseudo-tab
      if(llen < width)
        llen+=(llen%8)?8-(llen%8):8;
      if(llen >= width) {
        *r++='\n';
        s++;
        llen=wlen=0;
        p=NULL;
      } else
        *r++=*s++;
      continue;
    }

    if(*s == ' ') {	// Space
      if(llen == width) {
        *r++='\n';
        if(!wlen)
          *r++=*s;
	s++;
        llen=wlen=0;
        p=NULL;
      } else {
        p=r;
        *r++=*s++;
        llen++;
        wlen=0;
      }
      continue;
    }

    if(*s == '\e') {	// Escape character; copy ansi codes
      if(!*(s+1))
        break;
      *r++=*s++;
      if(*s != '[' && *s > 47) {
        *r++=*s++;
        continue;
      }
      *r++=*s++;
      if(!*s)
        break;
      if(*s < 48) {
        *r++=*s++;
        continue;
      }
      while(*s && r-buf < 8000 && *s < 64)
        *r++=*s++;
      if(!*s)
        break;
      *r++=*s++;
      continue;
    }

    if(llen == width) {
      if(p) {
        *p='\n';
        llen=wlen;
        p=NULL;
      } else {
        *r++='\n';
        p=NULL;
        llen=wlen=0;
      }
    }

    *r++=*s++;
    llen++;
    wlen++;
  }
  *r='\0';
  return buf;
}


/* Color Manipulation */

int default_color=7;

/* filter list of flags and return accompanying color code */
int color_code(int prev, char *s)
{
  static char flags[34]="h+f#u-i!*/InxrgybmvpcwNXRGYBMVPCW";
  static int attr[34]={8, 8, 128, 128, 256, 256, 512, 512, 1024, 2048, 4096,
                       -1, 0, 1,  2,  3,  4,  5,  5,  5,  6,  7,
                       -1, 0, 16, 32, 48, 64, 80, 80, 80, 96, 112};
  int a, color=prev;

  /* treat numbers as a base starting color */
  if(isdigit(*s)) {
    if((a=atoi(s)))
      color=a;
    while(isdigit(*++s));  // Skip pointer over digits to start of flags
  }

  /* parse flags string */
  for(;*s;s++)
    for(a=0;a<33;a++)
      if(flags[a] == *s) {
        switch(a) {
          case 11:          color = prev; break;     // Default @color.
	  case 22:          color = 7; break;        // All attributes off.
          case 12 ... 21:   color = (color & ~7) + attr[a]; break;
          case 23 ... 32:   color = (color & ~112) + attr[a]; break;
          default:          color |= attr[a];
        }
        break;
      }

  return color;
}

char *textcolor(int prev, int color)
{
  static char buf[24];

  if((prev & ~1024) == (color & ~1024))
    return "";
  if((color & ~1024) == 7)
    return "\e[m";

  strcpy(buf, "\e[");
  if(((prev&128) && !(color&128)) || ((prev&8) && !(color&8)) ||
     ((prev&256) && !(color&256))) {
    strcat(buf, "0;");
    prev=7;
  }
  if(!(prev&4096) && (color&4096))  // Invisible Text
    strcat(buf, "8;");
  if(!(prev&2048) && (color&2048))  // Half-bright Text
    strcat(buf, "2;");
  if(!(prev&512) && (color&512))    // Inverse Text
    strcat(buf, "7;");
  if(!(prev&256) && (color&256))    // Underline/Italics
    strcat(buf, "4;");
  if(!(prev&128) && (color&128))    // Blinking Text
    strcat(buf, "5;");
  if(!(prev&8) && (color&8))        // Hilight Text
    strcat(buf, "1;");
  if((prev&7) != (color&7))
    sprintf(buf+strlen(buf), "3%d;", color&7);       // Foreground Color
  if((prev&112) != (color&112))
    sprintf(buf+strlen(buf), "4%d;", (color>>4)&7);  // Background Color
  buf[strlen(buf)-1]='m';
  return buf;
}

char *rainbowmsg(char *s, int color, int size)
{
  static int seq[16]={4, 3, 6, 11, 5, 13, 14, 13, 6, 1, 2, 10, 4, 9, 12, 11};
  static char temp[8192];
  int prev=default_color;
  char *t=temp;

  /* If the first color is Normal, set it to Light Violet */
  if((color & 15) == 7)
    color=(color & ~15)+13;

  while(*s && t-temp < size) {
    strcpy(t, textcolor(prev, color));
    t=strchr(t, 0);
    *t++=*s++;
    if(*s != ' ') {
      prev=color;
      color=(color & ~15)+seq[color & 15];
    }
  }
  strcpy(t, textcolor(prev, default_color));
  return temp;
}
