/* io/file.c */

#include <dirent.h>
#include <errno.h>
#include "externs.h"

/* logfile structurization */

struct log
  main_log = {"logs/main", "_log_main", 1},
  error_log = {"logs/error", "_log_err", 1},
  io_log = {"logs/io", "_log_io", 0},
  resolv_log = {"logs/resolv", "_log_resolv", 0};

/* record a message in the local logfile */
void do_log(player, arg1, arg2)
dbref player;
char *arg1, *arg2;
{
  if(!strcasecmp(arg1, "main"))
    mud_log(&main_log, arg2);
  else if(!strcasecmp(arg1, "error"))
    mud_log(&error_log, arg2);
  else if(!strcasecmp(arg1, "io"))
    mud_log(&io_log, arg2);
  else if(!strcasecmp(arg1, "resolv"))
    mud_log(&resolv_log, arg2);
  else
    notify(player, tprintf("No such logfile '%s'.", arg1));
}

void mud_log(l, str)
struct log *l;
char *str;
{
  FILE *f;
  struct tm *tv;
  char buf[300], temp[256];
  long tt=time(NULL);

  strncpy(temp, str, 250);
  strcpy(temp+250, "...");

  if(mare_test_db) {
    printf("%s\n", temp);  /* stdout */
    return;
  } else if((boot_flags & BOOT_QUIET) || (!dberrcheck && game_online))
    com_send(NOTHING, l->com_channel, temp);
  else if(l->output) {
    DESC *d;

    sprintf(buf, "%s\n", temp);
    for(d=descriptor_list;d;d=d->next)
      if(!(d->flags & (C_OUTBOUND|C_SOCKET|C_DNS|C_PARENT|C_SERVICE))) {
        output2(d, buf);
        process_output(d);
      }
  }

  if((f=fopen(l->filename, "a")) == NULL) {
    mkdir("logs", 0755);
    if((f=fopen(l->filename, "a")) == NULL)
      return;
  }

  tv=localtime((time_t *)&tt);
  fprintf(f, "%02d/%02d-%02d:%02d:%02d| %s\n", tv->tm_mon+1, tv->tm_mday,
          tv->tm_hour, tv->tm_min, tv->tm_sec, temp);
  fclose(f);
}

void lperror(file, line, header)
char *file, *header;
int line;
{
  log_error(tprintf("%s:%d: %s: %s", file, line, header, strerror(errno)));
}


int get_filetype(char *name)
{
  struct stat statbuf;

  if((lstat(name, &statbuf)) == -1)
    return errno;

  if((statbuf.st_mode & S_IFMT) == S_IFDIR) /* directory */
    errno=EISDIR;
  else if((statbuf.st_mode & S_IFMT) != S_IFREG) /* not a regular file. eck! */
    errno=ENOENT;
  else if(statbuf.st_size >= 0x40000) /* file too large */
    errno=EFBIG;
  else
    errno=0;
  return errno;
}

/* get line from file removing leading \r\n */
char *ftext(char *buf, int len, FILE *f)
{
  char *s;

  if(!fgets(buf, len, f))
    return NULL;

  s=buf+strlen(buf)-1;
  if(s >= buf && (*s == '\n' || *s == '\r'))
    *s--='\0';
  if(s >= buf && *s == '\r')
    *s='\0';
  return buf;
}

void do_text(player, arg1, arg2, argv)
dbref player;
char *arg1, *arg2, *argv[];
{
  int rslt, top=1, bot=0;
  char filename[270];
  dbref thing;

  /* if no filename present, view local directory */
  if(!*arg1)
    arg1=".";

  /* find object to display file to */
  if(!argv[3] || !strcmp(argv[3], ""))
    thing = player;
  else {
    thing = match_thing(player, argv[3], MATCH_EVERYTHING);
    if(thing == NOTHING)
      return;
  }
  if(Typeof(thing) != TYPE_PLAYER && !power(thing, POW_ANIMATION))
    thing=db[thing].owner;

  /* check permissions */
  if(!controls(player, thing, POW_REMOTE)) {
    notify(player, "Permission denied.");
    return;
  }

  /* make sure we are viewing a valid secure filename, please */
  if(!ok_filename(arg1)) {
    notify(player, "Bad file name.");
    return;
  }

  sprintf(filename, "textfiles/%s", arg1);
  if(get_filetype(filename)) {
    if(errno == EISDIR) {
      list_directory(thing, filename);
      if(player != thing && !Quiet(player))
        notify(player, tprintf("Player %s just saw directory '%s'.",
               unparse_object(player, thing), filename));
      return;
    }
    notify(player, tprintf("%s.", strerror(errno)));
    return;
  }

  /* select file length */
  if(argv[1])
    if((top=atoi(argv[1])) < 1)
      top=1;
  if(argv[2])
    if((bot=atoi(argv[2])) < 1)
      bot=0;

  rslt=display_file(thing, filename, top, bot, argv[4] ?: "");

  /* display end message */
  if(!rslt)
    notify(player, "No lines to view.");
  else if(player != thing && !Quiet(player))
    notify(player, tprintf("Player %s just saw file '%s'.",
           unparse_object(player, thing), arg1));
}

int display_file(player, name, top, bot, wild)
dbref player;
char *name, *wild;
int top, bot;
{
  FILE *f;
  unsigned char buf[1024];
  int a, lin, count=0;

  /* find and open the requested file */
  if((f=fopen(name, "r")) == NULL) {
    perror(name);
    return -1;
  }

  /* Test top 5 lines for Binary File */
  for(lin=0;lin < 5 && fgets(buf, 1024, f);lin++)
    for(a=0;buf[a];a++)
      if(buf[a] < 32 && buf[a] != 13 && buf[a] != 27 &&
         (buf[a] < 7 || buf[a] > 10)) {
        notify(player, "That is a binary file, which cannot be displayed.");
        fclose(f);
        return -1;
      }

  lin=1;
  rewind(f);

  /* if we're showing data not from the top, read lines until desired point */
  while(lin < top) {
    if(!fgets(buf, 1024, f))
      break;
    lin++;
  }

  /* read and display remainder of file to object */
  while(bot < 1 || lin <= bot) {
    if(!ftext(buf, 1024, f))
      break;
    if(!*wild || wild_match(buf, wild, 0, 0))
      notify(player, buf);
    lin++; count++;
  }

  fclose(f);
  return count;
}

void do_motd(player)
dbref player;
{
  display_file(db[player].owner, "msgs/motd.txt", 0, 0, "*");
}

void list_directory(player, name)
dbref player;
char *name;
{
  DIR *f;
  struct dirent *filp;
  struct io_queue *dirlist=NULL, *dirp, *newp, *last;
  int a, count=0, colsize=4;
  char buf[256];

  /* open directory for reading */
  if((f=opendir(name)) == NULL) {
    perror(name);
    notify(player, "Directory does not exist.");
    return;
  }
  notify(player, tprintf("%s:", name));
  notify(player, "");

  /* read the directory and allocate all its names */
  while((filp=readdir(f))) {
    if(*filp->d_name == '.') /* skip files that begin with . */
      continue;
    sprintf(buf, "%s/%s", name, filp->d_name);

    /* check to see if it's a viewable file or directory */
    if((a=get_filetype(buf))) {
      if(a != EISDIR)
        continue;
    }

    /* scan thru directory list to sort */
    last=NULL;
    for(dirp=dirlist;dirp;dirp=dirp->next) {
      if(strcmp(filp->d_name, dirp->text) < 0)
        break;
      last=dirp;
    }
    newp=malloc(sizeof(struct io_queue));
    newp->text=malloc(strlen(filp->d_name)+(a?2:1));
    strcpy(newp->text, filp->d_name);
    if(a)
      strcat(newp->text, "/");
    if(!last)
      dirlist=newp;
    else
      last->next=newp;
    newp->next=dirp;

    if(strlen(newp->text)+4 > colsize)
      colsize=strlen(newp->text)+4;
    count++;
  }
  closedir(f);

  if(!count) {
    notify(player, "No files in directory.");
    return;
  }

  showlist(player, dirlist, colsize);
}

void showlist(player, dirlist, colsize)
dbref player;
struct io_queue *dirlist;
int colsize;
{
  struct io_queue *ptr, *next;
  char temp[32], buf[256];
  int a, cols=get_cols(NULL, player)-1;

  sprintf(temp, "%%-%d.%ds", colsize, colsize);

  a=0; *buf=0;
  for(ptr=dirlist;ptr;ptr=next) {
    next=ptr->next;
    if(++a >= cols/colsize || !ptr->next) {
      strcat(buf, ptr->text);
      notify(player, buf);
      *buf=0; a=0;
    } else
      strcat(buf, tprintf(temp, ptr->text));
    free(ptr->text);
    free(ptr);
  }
}

/* reads a particular line of a file */
char *read_file_line(name, line)
char *name;
int line;
{
  static char buf[1024];
  FILE *f;
  int count=0;

  /* try to open filename */
  if((f=fopen(name, "r")) == NULL) {
    perror(name);
    return "";
  }

  /* scan for line# */
  while(ftext(buf, 1024, f)) {
    ++count;
    if(count == line) {
      fclose(f);
      return buf;
    }
  }

  fclose(f);
  return "";
}

/* display local unix file to connection */
void display_textfile(d, filename)
DESC *d;
char *filename;
{
  FILE *fd;
  char buf[8192];

  if(!(fd=fopen(filename, "r"))) {
    char *r, *s=filename;

    if((r=strchr(s, '/')))
      s=r+1;
    sprintf(buf, "[No %s file present]\n", s);
    output2(d, buf);
    return;
  }

  while(fgets(buf, 8192, fd))
    output2(d, buf);
  fclose(fd);
  output2(d, "\n");
}

/* display file index block entry, to descriptor or player */
/* set <pattern> to NULL to display file header (before first &) */
void display_textblock(player, d, filename, pattern)
dbref player;
DESC *d;
char *filename, *pattern;
{
  FILE *fd;
  char buf[8192], *r, *s;

  if(!(fd=fopen(filename, "r"))) {
    s=filename;
    if((r=strchr(s, '/')))
      s=r+1;
    if(d)
      output2(d, tprintf("[No %s file present]\n", s));
    else
      notify(player, tprintf("[No %s file present]", s));
    return;
  }

  /* scan for beginning of text block */
  while(pattern) {
    if(!ftext(buf, 8192, fd)) {
      if(d)
        output2(d, tprintf("[No data available for %s]\n", pattern));
      else
        notify(player, tprintf("[No data available for %s]", pattern));
      fclose(fd);
      return;
    }

    if(*buf == '&') {
      for(s=buf+1;*s && isspace(*s);s++);
      if(!*pattern || string_prefix(s, pattern))
        break;
    }
  }

  /* print data until next text block is found */
  while(ftext(buf, 8192, fd)) {
    if(*buf == '&')
      break;
    if(d)
      output2(d, tprintf("%s\n", buf));
    else
      notify(player, buf);
  }

  fclose(fd);
}
