/* mare/timer.c */

#include "externs.h"
#include <sys/resource.h>

int object_atime=NOTHING;
int cmdav[2];
double cmdfp[2]={0, 0};

extern int halt_dump;
extern int getrusage();
static void update_cpu_percent();

/* these values should not be changed under any "normal" circumstances */
int CMDAV_INTERVAL=5;
int STATUS_INTERVAL=60;


static void db_autosave()
{
  if(!halt_dump)
    fork_and_dump(DB_FORK);
}

void reset_dump(secs)
int secs;
{
  timerval[0].ticks=DUMP_INTERVAL-secs;
}

static void update_cmdav()
{
  cmdfp[0]+=(((double)(nstat[NS_NPCMDS]-cmdav[0])/5)-cmdfp[0])/3;
  cmdfp[1]+=(((double)(nstat[NS_NQCMDS]-cmdav[1])/5)-cmdfp[1])/3;
  cmdav[0]=nstat[NS_NPCMDS];
  cmdav[1]=nstat[NS_NQCMDS];
}

static void next_atime()
{
  static int regen_loop=0, wander_loop=0;
  char *s;
  int a, i;

  if(ATIME_INTERVAL < 1)
    return;

  /* make sure @atime parameters are within range */
  if(RESURRECT_LOOP < 1)
    RESURRECT_LOOP=1;
  if(WANDER_LOOP < 1)
    WANDER_LOOP=1;

  for(i=db_top/ATIME_INTERVAL;i >= 0;i--) {
    /* when @atime reaches the last object in the database,
       loop around to the beginning. */
    if(++object_atime >= db_top) {
      object_atime = NOTHING;
      if(++regen_loop >= RESURRECT_LOOP)
        regen_loop=0;
      if(++wander_loop >= WANDER_LOOP)
        wander_loop=0;
      return;
    }

    /* skip over destroyed objects or disconnected players */
    if((db[object_atime].flags & (GOING|HAVEN)) || (Typeof(object_atime) ==
       TYPE_PLAYER && !(db[object_atime].flags & PLAYER_CONNECT)))
      continue;

    /* set %# for messages */
    speaker=object_atime;

    /* execute @atime */
    if(*(s=atr_get(object_atime, A_ATIME)))
      immediate_que(object_atime, s, object_atime);

    /* Objects set WANDER occasionally move around */
    if((object_atime % WANDER_LOOP) == (wander_loop % WANDER_LOOP) &&
       IS(object_atime, TYPE_THING, THING_WANDER)) {
      a=random_exit(object_atime, NOTHING);
      if(!Invalid(a))
        do_move(object_atime, db[a].name);
    }

    for(a=0;a<10;a++)
      *wptr[a]='\0';
  }
}

/* triggers once every second */
static void do_second()
{
  DESC *d;

  /* update players' online status */
  for(d=descriptor_list;d;d=d->next)
    if(d->flags & C_CONNECTED && !IS(db[d->player].location,
       TYPE_ROOM, ROOM_XZONE))
      db[d->player].data->age++;

  /* other stuff to do every second */

#ifndef __CYGWIN__
  dns_events();			// keep track of expired dns cache records
#endif

  update_io();			// updates login prompt and input quotas
  wait_second();		// process queue of @waited commands
  next_atime();			// exec the next set of @atimes
}

/* triggers each minute on the minute */
static void do_minute()
{
  static int last=3;
  int weekday;
  DESC *d;

  for(d=descriptor_list;d;d=d->next) {
    /* ping concentrators every minute */
    if(d->flags & C_PARENT)
      output2(d, "*\n");

    /* ping remotelink services */
    else if((d->flags & C_SERVICE) && d->state == 33) {
      DESC *k;
      int a;

      for(a=0,k=descriptor_list;k;k=k->next)
        if((k->flags & C_CONNECTED) && could_doit(0, k->player, A_HLOCK))
          a++;
      output2(d, tprintf("$20 %d\n", a));
    }

    /* process session inactivity timeout */
    else if(!(d->flags & (C_OUTBOUND|C_SOCKET|C_DNS|C_PARENT|C_SERVICE)))
      check_timeout(d);
  }

  /* other stuff to do every minute */

  fix_free_list();		// process objects scheduled for destruction
  update_cpu_percent();		// calculate new cpu% usage for past 60sec

  /* stuff to do on the verge of a new day */

  weekday=(int)((now+(TIMEZONE*3600L))/86400L) % 7;
  if(INACTIVENUKE > 0 && last != weekday) {
    last=weekday;

    // Remove old players from the game every Sunday at midnight.
    if(weekday == 3) {
      log_main("Invoking inactivity deletion...");
      inactive_nuke();
    }
  }
}


struct timerint timerval[]={
  {"Dump", &DUMP_INTERVAL, 0, db_autosave},
  {"Cmdav", &CMDAV_INTERVAL, 0, update_cmdav},
  {"Status", &STATUS_INTERVAL, 0, do_minute},
  {0}};


/* main timer loop */
void dispatch()
{
  static long recent=0L;
  int a, secs;

  /* initialize timer */
  if(!recent) {
    recent=now;
    timerval[2].ticks=(int)now%60;
    return;
  }

  /* only execute stuff when the second changes */
  if(!(secs=(int)(now-recent)))
    return;
  recent=now;

  /* execute game routines every second */
  do_second();

  /* increment timer function list */
  for(a=0;timerval[a].interval;a++)
    if(*(timerval[a].interval) > 0 && (timerval[a].ticks += secs) >=
       *(timerval[a].interval)) {
      timerval[a].fun();
      if(a == 2)
        timerval[a].ticks=(int)now%60;
      else
        timerval[a].ticks=0;
    }
}


/* functions for grabbing and calculating resource/cpu time used by mare */

double internal_lag=1;
struct timeval sys_diff={0}, sys_old={0}, sys_new={0},
	       time_diff={0}, time_old={0}, time_new={0},
	       idle_diff={0}, idle_old={0}, idle_new={0};

struct timeval timeval_sub(now,then)
struct timeval now, then;
{
  now.tv_sec -= then.tv_sec;
  now.tv_usec -= then.tv_usec;
  while(now.tv_usec < 0) {
    now.tv_usec += 1000000;
    now.tv_sec--;
  }
  if(now.tv_sec < 0)
    now.tv_sec = 0;
  return now;
}

static void update_cpu_percent()
{
  struct rusage usage;

  getrusage(RUSAGE_SELF, &usage);

  /* move previous new times to old */
  if(!(time_new.tv_sec))
    time_old.tv_sec = (int) mud_up_time;
  else {
    time_old.tv_sec = time_new.tv_sec;
    time_old.tv_usec = time_new.tv_usec;
    sys_old.tv_sec = sys_new.tv_sec;
    sys_old.tv_usec = sys_new.tv_usec;
  }

  /* set new times */
  gettimeofday(&time_new, NULL);
  sys_new.tv_sec=usage.ru_utime.tv_sec+usage.ru_stime.tv_sec;
  sys_new.tv_usec=usage.ru_utime.tv_usec+usage.ru_stime.tv_usec;

  time_diff=timeval_sub(time_new,time_old);
  sys_diff=timeval_sub(sys_new,sys_old);
}

double get_cpu_percent()
{
  double timediff, sysdiff;

  timediff=((double)time_diff.tv_sec)+(((double)time_diff.tv_usec)/1000000);
  sysdiff=(((double)sys_diff.tv_sec)+(((double)sys_diff.tv_usec)/1000000))*100;

  if(!timediff)
    return 0;
  return (sysdiff/timediff);
}

double get_cputime()
{
  struct rusage usage;

  getrusage(RUSAGE_SELF, &usage);
  return (double)usage.ru_utime.tv_sec+(double)usage.ru_stime.tv_sec+
       ((double)usage.ru_utime.tv_usec+(double)usage.ru_stime.tv_usec)/1000000;
}

void update_clock()
{
  double curdiff;

  /* move previous new times to old */
  if(!(idle_new.tv_sec)) {
    gettimeofday(&idle_old, NULL);
    idle_old.tv_sec--;
  } else {
    idle_old.tv_sec = idle_new.tv_sec;
    idle_old.tv_usec = idle_new.tv_usec;
  }

  gettimeofday(&idle_new, NULL);
  idle_diff=timeval_sub(idle_new,idle_old);

  /* calculate average internal lag */
  curdiff=((double)idle_diff.tv_sec)+(((double)idle_diff.tv_usec)/1000000);
  internal_lag+=(curdiff-internal_lag)/3;
}
