/* mare/utils.c */

#include "externs.h"
#include <math.h>

int make_quiet=0;
int match_count=0;
int match_accept=0;

char *wptr[10];
int wpos=0;


/* Matching routines for object finding in location, inventory, or database. */

int match_thing(player, name, type)
dbref player;
char *name;
int type;
{
  char shortname[512], buf[512], *r, *s;
  dbref match_abs=NOTHING, exact_match=NOTHING;
  int full, count=0, *ptr, alt_location=db[player].location;

  if(Invalid(player))
    return NOTHING;
  if(!*name && !make_quiet && !(type & MATCH_QUIET)) {
    notify(player, "No object specified.");
    return NOTHING;
  }

  strncpy(buf, name, 480);
  buf[480]='\0';
  strcpy(shortname, buf);

  /* check possessive */
  if((type & MATCH_POSSESSIVE) && (s=strchr(buf, '\'')) && s > buf &&
     *(s+1) == 's' && *(s+2) == ' ') {
    r=s+3;
    *s='\0';

    if((alt_location=match_thing(player, buf, type & ~MATCH_POSSESSIVE)) ==
       NOTHING)
      return NOTHING;

    /* set things up to match contents of the possessive */
    type &= MATCH_QUIET|MATCH_INVIS|MATCH_FULLNAME;
    type |= MATCH_CON;
    strcpy(shortname, r);
  }

  /* set static variables. must we check for a player's full name? */
  match_count=0;
  full=type & MATCH_FULLNAME;

  /* find number to match */
  if((s=strrchr(shortname, ' ')) && isdigit(*(s+1))) {
    match_accept=atoi(s+1);
    *s='\0';
  } else
    match_accept=1;

  /* find absolute global database reference number */
  if(*name == '#' && isdigit(*(name+1)) && !Invalid(atoi(name+1)))
    match_abs = atoi(name+1);

  /* match dbref */
  if(type&MATCH_ABS)
    exact_match = match_abs;

  /* match an object in the zone of <player> */
  if(type&MATCH_INZONE && exact_match == NOTHING && match_abs != NOTHING &&
     Typeof(player)==TYPE_ZONE && ((db[player].flags & ZONE_UNIVERSAL) ||
     inlist(db[mainroom(match_abs)].zone, player)))
    exact_match = match_abs;

  /* match player name */
  if(type&MATCH_PLAYER && exact_match == NOTHING && *name == '*' &&
     (count=lookup_player(name+1)) > NOTHING)
    exact_match = count;

  /* match reflexive */
  if(type&MATCH_ME && exact_match == NOTHING)
    if(!strcasecmp(name, "me"))
      exact_match = player;

  /* match home */
  if(type&MATCH_HOME && exact_match == NOTHING)
    if(!strcasecmp(name, "home"))
      exact_match = HOME;

  /* match linked zone#s */
  if(type&MATCH_ZONE && exact_match == NOTHING)
    for(ptr=db[mainroom(player)].zone;ptr && *ptr != NOTHING;ptr++)
      if(match_abs == *ptr)
        exact_match = *ptr;

  /* match container */
  if(type&MATCH_HERE && exact_match == NOTHING)
    if(!strcasecmp(name, "here") && !Invalid(db[player].location))
      exact_match = db[player].location;

  /* match exits in room */
  if(type&MATCH_EXIT && exact_match == NOTHING && !Invalid(db[player].location))
    exact_match = match_exit(player, name, shortname, match_abs, type);

  /* match inventory objects */
  if(type&MATCH_INVEN && exact_match == NOTHING)
    exact_match = match_list(player, player, name, shortname, match_abs,
                             type, &full);

  /* match contents in room */
  if(type&MATCH_CON && exact_match == NOTHING && !Invalid(db[player].location)
     && db[player].location != player) {
    count = match_list(player, alt_location, name, shortname, match_abs,
                       type, &full);
    if(!Invalid(count))
      exact_match=count;
  }

  /* destroyed objects do not exist */
  if(!Invalid(exact_match) && !(type&MATCH_DESTROYED) && Typeof(exact_match)
     == TYPE_GARBAGE)
    exact_match=NOTHING;

  /* return results */
  if(exact_match == NOTHING && !(type&MATCH_QUIET) && !make_quiet) {
    sprintf(buf, "I don't see '%s' here.", shortname);
    notify(player, buf);
  } else if(full == 2 && strcasecmp(db[exact_match].name, name)) {
    /* full name of player not specified */
    notify(player, "You must type out the player's name in full.");
    return NOTHING;
  }

  return exact_match;
}

/* utility for matching objects or exits in room */
static int can_match(player, thing, invis)
dbref player, thing;
int invis;
{
  /* check multi-dimensional plane */
  if(db[thing].plane != NOTHING && db[thing].plane != db[player].plane)
    return 0;

  /* matching self is always allowed, POW_DARK matches everything else */
  if(player == thing || controls(player, thing, POW_DARK))
    return 1;

  if((Typeof(thing) == TYPE_PLAYER && !(db[thing].flags & PLAYER_CONNECT)) ||
     !could_doit(player, thing, A_VLOCK) || (!invis &&
     Typeof(thing) == TYPE_EXIT && !can_see(player, thing)))
    return 0;

  return 1;
}

int match_exit(player, match_name, match_prefix, absolute, type)
dbref player, absolute;
char *match_name, *match_prefix;
int type;
{
  dbref exit, cmatch=NOTHING;

  /* search room's exitlist */
  DOLIST(exit, db[db[player].location].exits) {
    /* if exit# matches absolute, finish up */
    if(exit == absolute)
      return exit;
    if(!can_match(player, exit, (type & MATCH_INVIS)))
      continue;

    /* if exit name matches full name */
    if(!strcasecmp(db[exit].name, match_name))
      return exit;
    if(!strcasecmp(db[exit].name, match_prefix)) {
      if(++match_count == match_accept)
        cmatch=exit;
      continue;
    }

    /* check exit aliases */
    if(exit_alias(exit, A_ALIAS, match_name))
      return exit;
    if(exit_alias(exit, A_ALIAS, match_prefix)) {
      if(++match_count == match_accept)
        cmatch=exit;
      break;
    }
  }

  return cmatch;
}

int match_list(player, loc, match_name, match_prefix, absolute, type, full)
dbref player, loc, absolute;
char *match_name, *match_prefix;
int type, *full;
{
  dbref first, cmatch=NOTHING;

  if(IS(db[loc].location, TYPE_ROOM, ROOM_XZONE))
    return NOTHING;
  DOLIST(first, db[loc].contents) {
    if(first == absolute)
      return first;
    if(!can_match(player, first, (type & MATCH_INVIS)))
      continue;
    if(!strcasecmp(db[first].name, match_name))
      return first;
    if(string_match(db[first].name, match_prefix))
      if(++match_count == match_accept) {
        cmatch=first;
        if(*full && Typeof(first) == TYPE_PLAYER)
          *full=2;
      }
  }
  return cmatch;
}


/* Time Display Functions */

/* Timezone List */
static struct {
  char *zone;
  int have_dst;
  int utc_offset;
  char *name;
} tzones[]={
  {"GMT",  0,     0, "Greenwich Mean"},
  {"UT",   0,     0, "Universal Coordinated"},
  {"UTC",  0,     0, "Universal Coordinated"},
  {"WET",  0,     0, "Western European"},
  {"BST",  1,     0, "British Summer"},
  {"CET",  0,  +100, "Central European"},
  {"MET",  0,  +100, "Middle European"},
  {"MEWT", 0,  +100, "Middle European Winter"},
  {"MEST", 1,  +100, "Middle European Summer"},
  {"MESZ", 1,  +100, "Middle European Summer"},
  {"SWT",  0,  +100, "Swedish Winter"},
  {"SST",  1,  +100, "Swedish Summer"},
  {"FWT",  0,  +100, "French Winter"},
  {"FST",  1,  +100, "French Summer"},
  {"EET",  0,  +200, "Eastern Europe, USSR Zone 1"},
  {"BT",   0,  +300, "Baghdad, USSR Zone 2"},
  {"IT",   0,  +330, "Iran"},
  {"ZP4",  0,  +400, "USSR Zone 3"},
  {"ZP5",  0,  +500, "USSR Zone 4"},
  {"IST",  0,  +530, "Indian Standard"},
  {"ZP6",  0,  +600, "USSR Zone 5"},
  {"NST*", 0,  +630, "North Sumatra"},
  {"WAST", 0,  +700, "West Australia Standard"},
  {"WADT", 1,  +700, "West Australia Daylight"},
  {"SST*", 0,  +700, "South Sumatra, USSR Zone 6"},
  {"JT",   0,  +730, "Java"},
  {"WST",  0,  +800, "Western Australia Standard"},
  {"WDT",  1,  +800, "Western Australia Daylight"},
  {"CCT",  0,  +800, "China Coast, USSR Zone 7"},
  {"MT",   0,  +830, "Moluccas"},
  {"JST",  0,  +900, "Japan Standard, USSR Zone 8"},
  {"KST",  0,  +900, "Korea Standard"},
  {"CAST", 0,  +930, "Central Australian Standard"},
  {"CADT", 1,  +930, "Central Australian Daylight"},
  {"SAT",  0,  +930, "Southern Australian Standard"},
  {"SADT", 1,  +930, "Southern Australian Daylight"},
  {"EAST", 0, +1000, "Eastern Australian Standard"},
  {"EADT", 1, +1000, "Eastern Australian Daylight"},
  {"GST",  0, +1000, "Guam Standard, USSR Zone 9"},
  {"NZST", 0, +1200, "New Zealand Standard"},
  {"NZDT", 1, +1200, "New Zealand Daylight"},
  {"NZT",  0, +1200, "New Zealand"},
  {"IDLE", 0, +1200, "International Date Line East"},

  {"WAT",  0,  -100, "West Africa"},
  {"AT",   0,  -200, "Azores"},
  {"BST*", 0,  -300, "Brazil Standard"},
  {"GST*", 0,  -300, "Greenland Standard"},
  {"NFT",  0,  -330, "Newfoundland Standard"},
  {"NST",  0,  -330, "Newfoundland Standard"},
  {"NDT",  1,  -330, "Newfoundland Daylight"},
  {"AST",  0,  -400, "Atlantic Standard"},
  {"ADT",  1,  -400, "Atlantic Daylight"},
  {"EST",  0,  -500, "Eastern Standard"},
  {"EDT",  1,  -500, "Eastern Daylight"},
  {"CST",  0,  -600, "Central Standard"},
  {"CDT",  1,  -600, "Central Daylight"},
  {"MST",  0,  -700, "Mountain Standard"},
  {"MDT",  1,  -700, "Mountain Daylight"},
  {"PST",  0,  -800, "Pacific Standard"},
  {"PDT",  1,  -800, "Pacific Daylight"},
  {"YST",  0,  -900, "Yukon Standard"},
  {"YDT",  1,  -900, "Yukon Daylight"},
  {"HST",  0, -1000, "Hawaii Standard"},
  {"HDT",  1, -1000, "Hawaii Daylight"},
  {"AHST", 0, -1000, "Alaska-Hawaii Standard"},
  {"AHDT", 1, -1000, "Alaska-Hawaii Daylight"},
  {"CAT",  0, -1000, "Central Alaska"},
  {"NT",   0, -1100, "Nome"},
  {"IDLW", 0, -1200, "International Date Line West"},
  {0}};

char *months[12]={
 "January", "February", "March", "April", "May", "June",
 "July", "August", "September", "October", "November", "December"};

char *days[7]={
 "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"};

char *gamedays[7]={
 "Luna", "Jinn", "Undine", "Dryad", "Sylphid", "Gaia", "Solaris"};

/* retrieve game "weather"-time */
char *wtime(unsigned long td, int typ)
{
  static char buf[40];
  char temp[10];
  int sec, day, month, year;
  quad tt;

  if(HOURS_PER_DAY < 1)
    HOURS_PER_DAY=24;

  if(td == -1L) {
    struct timeval current;
    gettimeofday(&current, NULL);
    tt=((quad)current.tv_sec-(quad)BEGINTIME)*24LL/(quad)HOURS_PER_DAY;
    tt+=((quad)current.tv_usec*24LL/(quad)HOURS_PER_DAY)/1000000LL;
  } else
    tt=(quad)td*24LL/(quad)HOURS_PER_DAY;

  year=tt/31449600LL;
  sec=tt%31449600LL;
  day=sec/86400;
  sec%=86400;
  month=(day<180)?day/30:(day==180 || day==181)?5:(day>181 && day<362)?
        (day-2)/30:11;

  if(typ)
    sprintf(buf, "%s, ", gamedays[day%7]);
  else
    sprintf(buf, "%-3.3s ", gamedays[day%7]);
  sprintf(temp, "%02d:%02d:%02d", sec/3600, (sec%3600)/60, sec%60);
  if(typ)
    strcat(buf, tprintf("%s %d, %s%d", months[month], ((day>181?day-2:day)-
           (month*30))+1, (year+FIRST_YEAR) < 10?"Year ":"", year+FIRST_YEAR));
  else
    strcat(buf, tprintf("%-3.3s %2d %s %04d", months[month],
           ((day>181?day-2:day)-(month*30))+1, temp, year+FIRST_YEAR));
  return buf;
}

/* convert time() output into full date format */
char *mkdate(char *str)
{
  static char buf[32];
  int a;

  *buf=0;
  for(a=0;a<7;a++)
    if(!strncmp(days[a], str, 3)) {
      sprintf(buf, "%s, ", days[a]);
      break;
    }

  for(a=0;a<12;a++)
    if(!strncmp(months[a], str+4, 3)) {
      sprintf(buf+strlen(buf), "%s %d, %04d", months[a], atoi(str+8),
              atoi(str+20));
      break;
    }

  return buf;
}

/* convert time() output into "mm/dd/yyyy" format */
char *filter_date(char *string)
{
  static char buf[12];
  int a;

  for(a=0;a<12;a++)
    if(!strncmp(string+4, months[a], 3))
      break;
  sprintf(buf, "%02d/%02d/%04d", a+1, atoi(string+8), atoi(string+20));
  return buf;
}

/* converts # of seconds into readable time format */
char *strtime(long tt)
{
  char *s=ctime((time_t *)&tt);

  s[strlen(s)-1]=0;
  return s;
}

/* set player's individual timezone for use with mktm() */
void do_tzone(dbref player, char *arg1)
{
  int tz=atoi(arg1), dst=HAVE_DST;
  char *s;

  if(Typeof(player) != TYPE_PLAYER) {
    notify(player, "Only players may change their timezones.");
    return;
  } if(!*arg1) {
    notify(player, "Usage: +tzone <gmt offset>:<dst? y|n>");
    notify(player, "");
    notify(player, tprintf("You are currently set at: %d:%c",
           db[db[player].owner].data->tz,
           db[db[player].owner].data->tzdst ? 'y':'n'));
    return;
  }

  if(tz < -12 || tz > 13 || (!isdigit(*arg1) && *arg1 != '-')) {
    notify(player,
           "Offset from Greenwich Mean Time can only be from -12 to 13.");
    return;
  }
  if((s=strchr(arg1, ':')))
    dst=(*++s == 'y' || *s == 'Y')?1:0;

  db[db[player].owner].data->tz=tz;
  db[db[player].owner].data->tzdst=dst;
  notify(player, tprintf("Local time set to: %s", mktm(now, player)));
}

/* converts # of seconds into readable time string w/timezone setting */
char *mktm(long cl, dbref thing)
{
  long utcdiff;

  /* calculate difference from UTC, if present */
  utcdiff = (((long)db[db[thing].owner].data->tz)-((long)TIMEZONE))*3600L;

  /* calculate daylight savings time differences */
  if(localtime((time_t *)&cl)->tm_isdst)
    utcdiff += ((long)db[db[thing].owner].data->tzdst-(HAVE_DST?1L:0L))*3600L;

  cl += utcdiff;

  /* generate ascii string */
  return strtime(cl);
}

/* convert time representation in numbers & words into #seconds */
long get_date(char *str, dbref thing, long epoch)
{
  struct tm tm;
  char ch, buf[strlen(str)+1], *r, *s, *t;
  int a, x, tz, tm_isset[6]={0}, tm_rel[6]={0};
  long tt;

  strcpy(buf, str);

  /* null string = current time */
  if(!*buf)
    return now;

  /* string contains # seconds since epoch */
  for(s=buf;*s && (isdigit(*s) || (s == buf && *s == '-'));s++);
  if(!*s)
    return atol(buf)+epoch;

  /* set default time value based on player's active timezone */
  memcpy(&tm, localtime((time_t *)&now), sizeof(struct tm));

  s=buf;
  while((r=strspc(&s))) {
    /* no-effect keywords */
    if(!*r || match_word("from to of on in and now today", r))
      continue;

    /* special keywords */
    if(!strcasecmp(r, "ago")) {
      for(a=0;a<6;a++)
        tm_rel[a]=-tm_rel[a];
      continue;
    } if(!strcasecmp(r, "tomorrow")) {
      tm_rel[2]++;
      continue;
    } if(!strcasecmp(r, "yesterday")) {
      tm_rel[2]--;
      continue;
    } if(match_word("a.m. am", r)) {
      if(tm_isset[5]++ || !tm_isset[3])
        return -1L;
      if(tm.tm_hour < 1 || tm.tm_hour > 12)
        return -1L;
      if(tm.tm_hour == 12)
        tm.tm_hour=0;
      continue;
    } if(match_word("p.m. pm", r)) {
      if(tm_isset[5]++ || !tm_isset[3])
        return -1L;
      if(tm.tm_hour < 1 || tm.tm_hour > 12)
        return -1L;
      if(tm.tm_hour != 12)
        tm.tm_hour+=12;
      continue;
    }

    /* strip punctuation symbols */
    if(r[strlen(r)-1] == ',')
      r[strlen(r)-1] = '\0';
    if(r[strlen(r)-1] == '.')
      r[strlen(r)-1] = '\0';

    /* parse date mm/dd/yyyy format */
    if(strchr(r, ch='/') || strchr(r, ch='-') || strchr(r, ch='.')) {
      /* month */
      if(tm_isset[1]++)
        return -1L;
      t=strchr(r, ch);
      *t++='\0';
      tm.tm_mon=atoi(r)-1;
      if(tm.tm_mon < 0)
        for(a=0;a<12;a++)
          if(string_prefix(months[a], r)) {
            tm.tm_mon=a;
            break;
          }
      if(tm.tm_mon < 0 || tm.tm_mon > 11)
        return -1L;

      /* day */
      if(tm_isset[2]++)
        return -1L;
      if((tm.tm_mday=atoi(t)) < 1)
        return -1L;

      /* year */
      if((r=strchr(t, ch))) {
        if(tm_isset[0]++)
          return -1L;
        r++;
        tm.tm_year=atoi(r);
        if(tm.tm_year < 0)
          return -1L;
        if(tm.tm_year < 40)
          tm.tm_year+=100;
        else if(tm.tm_year >= 100 && tm.tm_year < 1900)
          return -1L;
        else if(tm.tm_year >= 1900)
          tm.tm_year-=1900;
      }

      /* blank out hh:mm:ss if not yet specified */
      if(!tm_isset[3])
        tm.tm_hour=tm.tm_min=tm.tm_sec=0;

      continue;
    }

    /* parse time hh:mm:ss format */
    if(strchr(r, ':')) {
      if(tm_isset[3]++)
        return -1L;

      /* hour */
      tm.tm_hour=atoi(r);
      if(tm.tm_hour < 0 || tm.tm_hour > 23)
        return -1L;

      /* minute */
      r=strchr(r, ':')+1;
      tm.tm_min=atoi(r);
      if(tm.tm_min < 0 || tm.tm_min > 59)
        return -1L;

      /* second */
      if(strchr(r, ':')) {
        r=strchr(r, ':')+1;
        tm.tm_sec=atoi(r);
        if(tm.tm_sec < 0 || tm.tm_sec > 59)
          return -1L;
      } else
        tm.tm_sec=0;

      /* Set A.M. or P.M. */
      if(strchr(r, 'a') || strchr(r, 'A')) {
        tm_isset[5]=1;
        if(tm.tm_hour < 1 || tm.tm_hour > 12)
          return -1L;
        if(tm.tm_hour == 12)
          tm.tm_hour=0;
      } else if(strchr(r, 'p') || strchr(r, 'P')) {
        tm_isset[5]=1;
        if(tm.tm_hour < 1 || tm.tm_hour > 12)
          return -1L;
        if(tm.tm_hour != 12)
          tm.tm_hour+=12;
      }
      continue;
    }

    /* match relative time displacement */

    /* ordinal values */

    /* set day of month or year */
    if(isdigit(*r)) {
      if((x=atoi(r)) < 1)
        return -1L;
      if(x < 1900) {  /* try setting the date */
        if(tm_isset[2]++)
          return -1L;
        tm.tm_mday=x;
        continue;
      }
      if(tm_isset[0]++)  /* must be a year */
        return -1L;
      tm.tm_year=x-1900;
      continue;
    }

    /* scan weekdays */
    for(a=0;a<7;a++)
      if(string_prefix(days[a], r))
        break;
    if(a != 7)
      continue;

    /* scan monthnames */
    if(!tm_isset[1]) {
      for(a=0;a<12;a++)
        if(string_prefix(months[a], r))
          break;
      if(a != 12) {
        tm.tm_mon=a;
        tm_isset[1]=1;
        continue;
      }

      /* blank out hh:mm:ss if not yet specified */
      if(!tm_isset[3])
        tm.tm_hour=tm.tm_min=tm.tm_sec=0;
    }

    /* scan timezones for a match */
    if(!tm_isset[4]++) {
      for(a=0;tzones[a].name;a++)
        if(!strcasecmp(tzones[a].zone, r))
          break;
      if(tzones[a].name) {
        tz=a;
        continue;
      }
    }

    return -1L;  /* no keyword match */
  }

  tm.tm_sec   += tm_rel[5];
  tm.tm_min   += tm_rel[4];
  tm.tm_hour  += tm_rel[3];
  tm.tm_mday  += tm_rel[2];
  tm.tm_mon   += tm_rel[1];
  tm.tm_year  += tm_rel[0];
  tm.tm_isdst = -1;

  return tt=mktime(&tm);
}

// Prints # of seconds in one of three formats. Format =
//  0:  17d, 16h, 45m	All time values, TMA
//  1:  17d 16:45	Online column in who list, TML
//  2:  2w		Idle column in who list, TMS
//  3:  2 weeks		Full format, TMF
//
char *time_format(int format, long dt)
{
  const char *ys[6]={"year", "week", "day", "hour", "minute", "second"};
  const long ts[6]={31536000L, 604800L, 86400L, 3600L, 60L, 1L};

  static char buf[64];
  long amt;
  int i;
  
  if(dt < 0L)
    dt=0L;

  switch(format) {
  case 0:
    for(*buf='\0',i=0;i<6;i+=i?1:2)  /* skip identifier for weeks */
      if((amt=dt/ts[i]) || (i == 5 && !*buf)) {
        sprintf(buf+strlen(buf), "%s%ld%c", *buf?", ":"", amt, *ys[i]);
        dt-=amt*ts[i];
      }
    return buf;
  case 1:
    if(dt/86400L > 99L)
      sprintf(buf, "%-4ld%02ld:%02ld", dt/86400L, dt/3600L%24L, dt/60L%60L);
    else if(dt/86400L)
      sprintf(buf, "%ldd %02ld:%02ld", dt/86400L, dt/3600L%24L, dt/60L%60L);
    else
      sprintf(buf, "%02ld:%02ld", dt/3600L, dt/60L%60L);
    return buf;
  }

  for(i=0;i<6;i++)
    if((amt=dt/ts[i]) || i == 5)
      break;

  if(format == 2)
    sprintf(buf, "%ld%c", amt, *ys[i]);
  else if(amt == 1L)
    sprintf(buf, "%s %s", (i == 3)?"an":"a", ys[i]);
  else
    sprintf(buf, "%ld %ss", amt, ys[i]);

  return buf;
}

char *grab_time(dbref player)
{
  static char buf[16];
  char *s=mktm(now, player);

  sprintf(buf, "[%d%-3.3s %c.m.]", (atoi(s+11)+11)%12+1, s+13,
          (atoi(s+11) > 11)?'p':'a');
  return buf;
}


/* built-in functions not present on some platforms: */

/* return length of file in bytes */
long fsize(char *file)
{
  struct stat statbuf;

  if(stat(file, &statbuf))
    return -1L;
  return statbuf.st_size;
}

/* return modification date of file */
long fdate(char *file)
{
  struct stat statbuf;

  if(stat(file, &statbuf))
    return -1L;
  return statbuf.st_mtime;
}

/* displays a random number from 0 to num-1 */
int rand_num(int num)
{
  if(num < 2)
    return 0;
  return (((int)random() & 0x7fffffff) % num);  /* return positive integers */
}

/* randomized division for floating point effect */
int rdiv(int x, int y)
{
  if(!y)
    return 0;
  return (x/y)+(rand_num(y) < x%y);
}

/* convert <x,y> coordinate to direction from <0,0> in degrees */
double degrees(double x, double y)
{
  double var;

  if(!y)
    return (x < 0)?270:(x > 0)?90:0;
  var=180 * atan(x/y) / 3.14159265358979323846;  /* PI */
  if(y < 0)
    return var+180;
  if(var < 0)
    var+=360;
  return var;
}

/* truncates trailing zeroes on a floating point number */
char *trunc_float(char *buf, double val)
{
  char *s;

  sprintf(buf, "%.9f", val);
  s=strchr(buf, 0)-1;
  while(*s == '0')
    *s--='\0';
  if(*s == '.')
    *s='\0';

  return buf;
}

/* returns 1 if <prefix> is a prefix of <string> */
int string_prefix(char *string, char *prefix)
{
  if(!*prefix)
    return 0;
  while(*string && *prefix && tolower(*string) == tolower(*prefix))
    string++, prefix++;
  return (*prefix == '\0');
}

/* accepts only nonempty matches starting at the beginning of a word */
char *string_match(char *src, char *sub)
{
  if(!*sub)
    return 0;

  while(*src) {
    if(string_prefix(src, sub))
      return src;

    /* else scan to beginning of next word */
    while(*src && *src != ' ' && *src != '-')
      src++;
    while(*src && (*src == ' ' || *src == '-'))
      src++;
  }
  return 0;
}

/* check if a string matches a list of semicolon-separated aliases */
int exit_alias(dbref thing, unsigned char attr, char *str)
{
  char buf[512], *p, *r, *s=buf;

  strncpy(buf, atr_get(thing, attr), 511);
  buf[511]='\0';
  while((r=parse_up(&s, ';'))) {
    /* remove spaces around aliases */
    while(*r && *r == ' ')
      r++;
    p=r+strlen(r);
    while(p > r && *(p-1) == ' ')
      *--p='\0';
    if(*r && !strcasecmp(r, str))
      return 1;
  }
  return 0;
}

/* matches a word from a space-separated list, returning word position */
int match_word(char *str, char *word)
{
  char list[16384], *r, *s;
  int count=0;

  strcpy(list, str);
  s=list;
  while((r=parse_up(&s,' '))) {
    ++count;
    if(!strcasecmp(r, word))
      return count;
  }
  return 0;
}

/* function to split up a list given a seperator */
/* note str will get hacked up */
char *parse_up(char **str, int delimit)
{
  int deep=0;
  char *s=*str, *os=*str;

  if(!*s)
    return 0;
  while(*s && (*s != delimit))
    if(*s++ == '{') {
      deep=1;
      while(deep && *s)
        switch(*s++) {
        case '{':
          deep++;
          break;
        case '}':
          deep--;
          break;
        }
    }

  if(*s)
    *s++=0;
  *str=s;
  return os;
}

/* same as above, but removes surrounding {}'s */
char *parse_perfect(char **str, int del)
{
  char *old=parse_up(str, del);
  int a;

  if(!old)
    return NULL;
  a=strlen(old);
  if(*old == '{' && *(old+a-1) == '}') {
    *(old+a-1) = '\0';
    old++;
  }
  return old;
}

/* like parse_up, but skips over symbols inside ansi escape codes */
char *ansi_parse_up(char **str, int delimit)
{
  int deep=0;
  char *s=*str, *os=*str;

  if(!*s)
    return 0;
  while(*s && (*s != delimit)) {
    if(*s == '\e') {
      /* 1-digit escape sequence */
      if(*++s != '[' && *s > 47)
        continue;
      s++;

      /* 2-digit sequence */
      if(*s < 48) {
        if(!*(++s))
          break;
        continue;
      }

      /* variable length escape code; terminates with a character > 63 */
      while(*s && *s < 64)
        s++;
      if(!*s)
        break;
    } else if(*s++ == '{') {
      deep=1;
      while(deep && *s)
        switch(*s++) {
        case '{':
          deep++;
          break;
        case '}':
          deep--;
          break;
        }
    }
  }

  if(*s)
    *s++=0;
  *str=s;
  return os;
}

/* separates string by whitespace */
char *strspc(char **s)
{
  char *str=*s;

  if(!**s)
    return 0;
  while(**s && !isspace(**s))
    (*s)++;
  while(**s && isspace(**s))
    *(*s)++='\0';
  return str;
}

/* center string in field of 'len' spaces, padding with 'fill' characters */
char *center(char *str, int len, char fill)
{
  static char buf[8192];
  int i=strlen(str), j=0;

  if(len > i) {
    while(j < (len-i)/2)
      buf[j++]=fill;
    strcpy(buf+j, str);
    j+=i;
    i=((len-i+1)/2)+j;
    while(j < i)
      buf[j++]=fill;
    buf[j]='\0';
  } else {
    strncpy(buf, str, len);
    buf[len]='\0';
  }
  return buf;
}

/* justify sentence/description in width of characters */
char *eval_justify(char **str, int width)
{
  static char buf[256];
  char *p=*str, *s=*str;
  int a=0, b;

  if(!*s)
    return NULL;

  while(*s) {
    if(*s == '\\' || *s == '\n') {
      *str=s+1;
      buf[a]='\0';
      return buf;
    }

    if(*s == '\t') {	// Pseudo-tab
      b=a+((a%8)?8-(a%8):8);
      if(b >= width) {
        while(*s && (*s == ' ' || *s == '\t'))
          s++;
        *str=s;
        while(a > 0 && buf[a-1] == ' ')
          a--;
        buf[a]='\0';
        return buf;
      }
      while(a < b)
        buf[a++]=' ';
      p=s+1;
      s++;
      continue;
    }

    if(isspace(*s)) {
      if(a == width) {
        while(*s && isspace(*s))
          s++;
        *str=s;
        while(a > 0 && buf[a-1] == ' ')
          a--;
        buf[a]='\0';
        return buf;
      }
      buf[a++]=' ';
      p=s+1;
      s++;
      continue;
    }

    if(a == width) {
      if(s-p >= width) {
        buf[a]='\0';
        *str=s;
        return buf;
      }
      while(a > 0 && buf[a-1] != ' ')
        a--;
      while(a > 0 && buf[a-1] == ' ')
        a--;
      buf[a]='\0';
      *str=p;
      return buf;
    }

    buf[a++]=*s++;
  }

  buf[a]='\0';
  *str=s;
  return buf;
}

/* string editing utility */
int edit_string(dest, string, old, new)
char *dest, *string, *old, *new;
{
  int len, newlen, count=0, i=0;

  /* nothing to do? */
  if(!*old) {
    strcpy(dest, string);
    return 0;
  }

  /* prepend string */
  if(!strcmp(old, "^")) {
    len=strlen(new);
    strcpy(dest, new);
    strncpy(dest+len, string, 8000-len);
    dest[8000]=0;
    return 1;
  }

  /* append string */
  if(!strcmp(old, "$")) {
    len=strlen(string);
    strcpy(dest, string);
    strncpy(dest+len, new, 8000-len);
    dest[8000]=0;
    return 1;
  }

  len=strlen(old);
  newlen=strlen(new);

  /* match and substitute */
  while(i < 8000 && *string)
    if(i+newlen <= 8000 && !strncmp(string,old,len)) {
      strcpy(dest+i, new);
      i += newlen;
      string += len;
      count++;
    } else
      dest[i++]=*string++;
  dest[i]=0;

  return count;
}


/* Wildcard Matching Procedures */

// d=string list; command typed in
// s=pattern to check; contains the wildcard symbols
//  flags-
// boolean: check mathematical boolean arguments ! > < on *s
// record:  record match data in %0 through %9 (beginning at %record-1).

int wild_match(d, s, bool, rec)
char *d, *s;
int bool, rec;
{
  static char buff[8192];
  int a, e=0, inv=0, rslt=0;

  /* set initialization parameters first time through */
  if(bool != -1) {
    wpos=(rec < 1 || rec > 10) ? 0:(rec-1);
    d=ansi_strip(d);  /* remove any stray ansi escape codes */
  }

  /* check for boolean arguments ! > < */
  if(bool == 1) {
    if(*s == '!') {
      s++;
      inv=1;
    }
    if(*s == '>') {
      s++;
      if((isdigit(*s) || (*s=='-')) || (isdigit(*d) || (*d=='-')))
        rslt=(atoi(s) < atoi(d));
      else
        rslt=(strcasecmp(s,d) < 0);
      e=1;
    } else if(*s == '<') {
      s++;
      if((isdigit(*s) || (*s=='-')) || (isdigit(*d) || (*d=='-')))
        rslt=(atoi(s) > atoi(d));
      else
        rslt=(strcasecmp(s,d) > 0);
      e=1;
    }
    if(e) {
      if(inv)
        return !rslt;
      return rslt;
    }
  }

  while(*s != '*') {
    switch(*s) {
    case '?':
      /* single character match; return false if at end of data */
      if(!*d)
        return inv;
      if(rec && wpos < 10) {
        sprintf(buff, "%c", (*d == 27)?0:*d);
        strcpy(wptr[wpos++], buff);
      }
      break;
    case '\\':
      /* escape character; force literal match of next character */
      s++;
    default:
      /* literal character match; if matching end of data, return true */
      if(tolower(*s) != tolower(*d))
        return inv;
      if(!*d)
        return 1-inv;
    }
    s++; d++;
  }

  /* skip over excess wildcards */
  while((*s == '?') || (*s == '*'))
    if(*s++ == '?')
      if(!*d++)
        return inv;  /* end of data reached */

  /* return true on trailing '*' */
  if(*s == '\\')
    s++;
  if(!*s) {
    if(rec && wpos < 10) {
      while(*d && isspace(*d))
        d++;
      a=0;
      while(*d)
        buff[a++]=*d++;
      buff[a]=0;
      strcpy(wptr[wpos++], buff);
    }
    return 1-inv;
  }

  /* scan for possible matches */
  a=0;
  while(*d) {
    if(!*d)
      break;
    if(*s == '\\')
      s++;
    if(tolower(*d) == tolower(*s) && wild_match(d+1,s+1,-1,0)) {
      if(rec && wpos < 10) {
        buff[a]=0;
        while(a && isspace(buff[a-1]))  /* remove following spaces */
          buff[--a]=0;
        strcpy(wptr[wpos++], buff);  /* record this * element */
        wild_match(d+1,s+1,-1,rec);  /* now record the other elements */
      }
      return 1-inv;
    } else if(rec && !(isspace(*d) && !a))  /* remove leading spaces */
      buff[a++]=*d;
    d++;
  }
  return inv;
}
