/* prog/eval.c - Function Parser */

#include "externs.h"
#include <math.h>

int func_recur=0;	// The depth level that we're on.
int func_total=0;	// Number of total functions processed in command.

#define FUNCTION(x) static void fun_##x(char *buff, char *args[10], \
                                        dbref privs, dbref doer, int nargs)

#define INIT_MATCH(x,y)    dbref (x)=match_thing(privs,y,MATCH_EVERYTHING)
#define INIT_CMATCH(x,y,z) dbref (x)=match_controlled(privs,y,z)

#define ADDWORD(string, s, stmt)  \
 ({  if(s != string) { \
       if(s-string > 7980) { strcpy(s, " ..."); return; } \
       *s++=' '; \
     } stmt; s=strchr(s, 0);  })

#define DELIMCH(x,y,ch)		char (x)=((nargs > y && *args[y] > 32 && \
                                           *args[y] < 127) ? *args[y] : ch)
#define DELIM(x,y) DELIMCH(x,y,' ')

#define ERR_PERM	"#-1 Permission denied."
#define ERR_PLAYER	"#-1 No such player."
#define ERR_RANGE	"#-1 Out of range."
#define ERR_LOC		"#-1 Too far away to see."
#define ERR_CONN	"#-1 Not connected."
#define ERR_DATE	"#-1 Invalid date."
#define ERR_ENV		"#-1 Invalid global register."


/* Set 1: Basic Math *********************************************************/

FUNCTION(add) {
  int a, result=0;
  char *r, *s;

  for(a=0;a<nargs;a++) {
    s=args[a];
    while((r=parse_up(&s,' ')))
      result+=atoi(r);
  }
  sprintf(buff, "%d", result);
}

FUNCTION(fadd) {
  int a;
  char *r, *s;
  double result=0;

  for(a=0;a<nargs;a++) {
    s=args[a];
    while((r=parse_up(&s,' ')))
      result+=atof(r);
  }
  trunc_float(buff, result);
}

FUNCTION(mul) {
  int a, result=1;
  char *r, *s;

  for(a=0;a<nargs;a++) {
    s=args[a];
    while((r=parse_up(&s,' ')))
      result*=atoi(r);
  }
  sprintf(buff, "%d", result);
}

FUNCTION(fmul) {
  int a;
  char *r, *s;
  double result=1;

  for(a=0;a<nargs;a++) {
    s=args[a];
    while((r=parse_up(&s,' ')))
      result*=atof(r);
  }
  trunc_float(buff, result);
}

FUNCTION(sub) { sprintf(buff, "%d", atoi(args[0])-atoi(args[1])); }
FUNCTION(fsub) { trunc_float(buff, atof(args[0])-atof(args[1])); }
FUNCTION(div) { if(!atoi(args[1])) strcpy(buff, "inf");
                  else sprintf(buff, "%d", atoi(args[0])/atoi(args[1])); }
FUNCTION(fdiv) { if(!atof(args[1])) strcpy(buff, "inf");
                   else trunc_float(buff, atof(args[0])/atof(args[1])); }
FUNCTION(mod) { if(!atoi(args[1])) strcpy(buff, "0");
                  else sprintf(buff, "%d", atoi(args[0])%atoi(args[1])); }

/* Set 2: Predicate Calculus / Logic *****************************************/

FUNCTION(min) {
  int a, count=0, result=0;
  char *r, *s;

  for(a=0;a<nargs;a++) {
    s=args[a];
    while((r=parse_up(&s,' ')))
      if(!count++ || atoi(r) < result)
        result=atoi(r);
  }
  sprintf(buff, "%d", result);
}

FUNCTION(max) {
  int a, count=0, result=0;
  char *r, *s;

  for(a=0;a<nargs;a++) {
    s=args[a];
    while((r=parse_up(&s,' ')))
      if(!count++ || atoi(r) > result)
        result=atoi(r);
  }
  sprintf(buff, "%d", result);
}

FUNCTION(gt)   { sprintf(buff, "%d", atof(args[0]) >  atof(args[1])); }
FUNCTION(gteq) { sprintf(buff, "%d", atof(args[0]) >= atof(args[1])); }
FUNCTION(lt)   { sprintf(buff, "%d", atof(args[0]) <  atof(args[1])); }
FUNCTION(lteq) { sprintf(buff, "%d", atof(args[0]) <= atof(args[1])); }
FUNCTION(eq)   { sprintf(buff, "%d", atof(args[0]) == atof(args[1])); }
FUNCTION(neq)  { sprintf(buff, "%d", atof(args[0]) != atof(args[1])); }

FUNCTION(truth) { sprintf(buff, "%d", ISTRUE(args[0])); }
FUNCTION(not) { sprintf(buff, "%d", !ISTRUE(args[0])); }
FUNCTION(if) { strcpy(buff, ISTRUE(args[0])?args[1]:""); }
FUNCTION(ifelse) { strcpy(buff, ISTRUE(args[0])?args[1]:args[2]); }
FUNCTION(default) { strcpy(buff, ISTRUE(args[0])?args[0]:args[1]); }
FUNCTION(between) { sprintf(buff, "%d", (atof(args[0])>=atof(args[1]) &&
                                         atof(args[0])<=atof(args[2]))); }

FUNCTION(band) { sprintf(buff, "%d", atoi(args[0]) & atoi(args[1])); }
FUNCTION(bnand) { sprintf(buff, "%d", atoi(args[0]) & ~atoi(args[1])); }
FUNCTION(bor) { sprintf(buff, "%d", atoi(args[0]) | atoi(args[1])); }
FUNCTION(bxor) { sprintf(buff, "%d", atoi(args[0]) ^ atoi(args[1])); }
FUNCTION(land) { sprintf(buff, "%d", ISTRUE(args[0]) && ISTRUE(args[1])); }
FUNCTION(lor) { sprintf(buff, "%d", ISTRUE(args[0]) || ISTRUE(args[1])); }
FUNCTION(lxor) { sprintf(buff, "%d", ISTRUE(args[0]) ^ ISTRUE(args[1])); }

FUNCTION(shl) { sprintf(buff, "%d", atoi(args[0]) << atoi(args[1])); }
FUNCTION(shr) { sprintf(buff, "%d", atoi(args[0]) >> atoi(args[1])); }

/* Set 3: Math Library *******************************************************/

FUNCTION(e) { strcpy(buff, "2.718281828"); }
FUNCTION(pi) { strcpy(buff, "3.141592654"); }
FUNCTION(sin) { trunc_float(buff, sin(atof(args[0]))); }
FUNCTION(cos) { trunc_float(buff, cos(atof(args[0]))); }
FUNCTION(tan) { trunc_float(buff, tan(atof(args[0]))); }
FUNCTION(asin) { if(abs(atof(args[0]) > 1)) strcpy(buff, "inf");
                   else trunc_float(buff, asin(atof(args[0]))); }
FUNCTION(acos) { if(abs(atof(args[0]) > 1)) strcpy(buff, "inf");
                   else trunc_float(buff, acos(atof(args[0]))); }
FUNCTION(atan) { trunc_float(buff, atan(atof(args[0]))); }

FUNCTION(exp) { trunc_float(buff, exp(atof(args[0]))); }
FUNCTION(pow) { trunc_float(buff, pow(atof(args[0]),atof(args[1]))); }
FUNCTION(sqrt) { if(atof(args[0]) < 0) strcpy(buff, "inf");
                   else trunc_float(buff, sqrt(atof(args[0]))); }
FUNCTION(ln) { if(atof(args[0]) <= 0) strcpy(buff, "inf");
                 else trunc_float(buff, log(atof(args[0]))); }

FUNCTION(neg) { trunc_float(buff, -atof(args[0])); }
FUNCTION(abs) { trunc_float(buff, abs(atof(args[0]))); }
FUNCTION(sign) { sprintf(buff, "%d", atof(args[0])>0?1:atof(args[0])<0?-1:0); }

FUNCTION(int) { sprintf(buff, "%d", atoi(args[0])); }
FUNCTION(ceil) { sprintf(buff, "%.0f", ceil(atof(args[0]))); }
FUNCTION(floor) { sprintf(buff, "%.0f", floor(atof(args[0]))); }

FUNCTION(degrees) { trunc_float(buff, degrees(atof(args[0]),atof(args[1]))); }
FUNCTION(dist2d) { trunc_float(buff, sqrt(pow(atof(args[0])-atof(args[2]),2)+
                        pow(atof(args[1])-atof(args[3]),2))); }
FUNCTION(dist3d) { trunc_float(buff, sqrt(pow(atof(args[0])-atof(args[3]),2)+
                        pow(atof(args[1])-atof(args[4]),2)+
                        pow(atof(args[2])-atof(args[5]),2))); }

FUNCTION(round) {
  int range=(nargs==2)?atoi(args[1]):0;
  sprintf(buff, tprintf("%%.%df", range<0?0:range>9?9:range), atof(args[0]));
}

FUNCTION(log) {
  double base=(nargs==2)?atof(args[1]):10.0;

  if(atof(args[0]) <= 0 || base <= 0)
    strcpy(buff, "inf");
  else
    trunc_float(buff, log(atof(args[0]))/log(base));
}

FUNCTION(factor) {
  int count=0, d=3, q, n=atoi(args[0]);
  char *s=buff;

  if(!n) {
    strcpy(buff, "0");
    return;
  } if(n < 0) {
    strcpy(buff, "-");
    n=-n;
  } else
    *buff='\0';

  while(!(n & 1) && n) {
    sprintf(s=strchr(s, 0), "%s2", count++?" ":"");
    n >>= 1;
  }

  do {
    while(n == (q=n/d) * d) {
      sprintf(s=strchr(s, 0), "%s%d", count++?" ":"", d);
      n = q;
    }
  } while ((d += 2) <= q);

  if(!count || n > 1)
    sprintf(s=strchr(s, 0), "%s%d", count++?" ":"", n);
}

/* Set 4: Game Runtime Configuration *****************************************/

FUNCTION(mudname) { strcpy(buff, MUD_NAME); }
FUNCTION(uptime) { sprintf(buff, "%ld", now-mud_up_time); }
FUNCTION(version) { strcpy(buff, mud_version); }
FUNCTION(restarts) { sprintf(buff, "%d", nstat[NS_EXEC]); }
FUNCTION(dbtop) { sprintf(buff, "#%d", db_top-1); }
FUNCTION(upfront) { sprintf(buff, "#%d", get_upfront()); }
FUNCTION(config) { get_config(privs, buff, args[0]); }
FUNCTION(cpu) { sprintf(buff, "%.3f", get_cpu_percent()); }
FUNCTION(cputime) { sprintf(buff, "%.2f", get_cputime()); }
FUNCTION(lag) { sprintf(buff, "%.3f", internal_lag); }

FUNCTION(netstat) {
  const char *ns[NET_STATS]={
    "FILES", "OUTPUT", "INPUT", "EXEC", "CRASHES", "DBSAVE", "LOGINS",
    "CONNECTS", "FAILED", "NEWCR", "GUESTS", "NCMDS", "NPCMDS", "NQCMDS",
    "NFCMDS", "QUEUE", "MAIL", "EMAIL", "ERECV", "RESOLV", "QUERY", "ATKS",
    "GOLD", "TCHEST", "EXP", "LEVEL", "TECHPTS", "DMONSTER", "DPLAYER",
    "ALMANAC", "CASTROD", "FISH", "PLANTS"};

  int a;

  for(a=0;a<NET_STATS && ns[a];a++)
    if(!strcasecmp(ns[a], args[0]))
      break;
  if(a == NET_STATS)
    strcpy(buff, "#-1 No such variable.");
  else
    sprintf(buff, "%d", nstat[a]);
}

/* Set 5: Global Weather Almanac *********************************************/

FUNCTION(wtime) { strcpy(buff, wtime(nargs?atol(args[0]):-1L,0)); }
FUNCTION(wdate) { strcpy(buff, wtime(nargs?atol(args[0]):-1L,1)); }
FUNCTION(timeofday) { sprintf(buff, "%d", TIMEOFDAY); }

/* Set 6: Time and Session Management ****************************************/

FUNCTION(time) {
  long tt=nargs?get_date(args[0], privs, 0L):now;
  strcpy(buff, (tt == -1L)?ERR_DATE:mktm(tt, privs));
}

FUNCTION(date) {
  long tt=nargs?get_date(args[0], privs, 0L):now;
  strcpy(buff, (tt == -1L)?ERR_DATE:mkdate(mktm(tt, privs)));
}

FUNCTION(rtime) {
  long tt=nargs?get_date(args[0], privs, 0L):now;
  if(tt == -1L)
    strcpy(buff, ERR_DATE);
  else {
    char *s=mktm(tt, privs);
    sprintf(buff, "%02d%-6.6s %cM", (atoi(s+11)+11)%12+1, s+13,
            (atoi(s+11) > 11)?'P':'A');
  }
}

FUNCTION(xtime) {
  if(!nargs) {
    struct timeval current;
    gettimeofday(&current, NULL);
    sprintf(buff, "%f", (double)current.tv_sec+
            ((double)current.tv_usec/1000000));
  } else {
    long tt=get_date(args[0], privs, 0L);
    if(tt == -1L)
      strcpy(buff, ERR_DATE);
    else
      sprintf(buff, "%ld", tt);
  }
}

FUNCTION(last) {
  dbref thing=lookup_player(args[0]);

  if(thing == AMBIGUOUS)
    thing=privs;
  if(thing == NOTHING || Typeof(thing) != TYPE_PLAYER)
    strcpy(buff, ERR_PLAYER);
  else
    sprintf(buff, "%ld", db[thing].data->last);
}

FUNCTION(lastoff) {
  dbref thing=lookup_player(args[0]);

  if(thing == AMBIGUOUS)
    thing=privs;
  if(thing == NOTHING || Typeof(thing) != TYPE_PLAYER)
    strcpy(buff, ERR_PLAYER);
  else if(!controls(privs, thing, POW_WHO))
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%ld", db[thing].data->lastoff);
}

FUNCTION(tml) { strcpy(buff, time_format(TML, atol(args[0]))); }
FUNCTION(tms) { strcpy(buff, time_format(TMS, atol(args[0]))); }
FUNCTION(tmf) { strcpy(buff, time_format(TMF, atol(args[0]))); }
FUNCTION(tma) { strcpy(buff, time_format(TMA, atol(args[0]))); }

static void io_list(buff, privs, name, typ)
char *buff, *name;
dbref privs;
int typ;
{
  DESC *d;
  dbref who;

  if((who=lookup_player(name)) == AMBIGUOUS)
    who=privs;
  if(who == NOTHING || Typeof(who) != TYPE_PLAYER)
    strcpy(buff, ERR_PLAYER);
  else if(!controls(privs, who, POW_WHO) && !could_doit(privs, who, A_HLOCK))
    strcpy(buff, ERR_PERM);
  else if(!(d=match_desc(who)))
    strcpy(buff, ERR_CONN);
  else switch(typ) {
    case 1: sprintf(buff, "%ld", now - d->last_time); return;
    case 2: sprintf(buff, "%ld", now - d->login_time); return;
    case 3: strcpy(buff, d->addr); return;
    case 4: sprintf(buff, "%d", d->concid); return;
    case 5: sprintf(buff, "%d", d->port); return;
    case 6: sprintf(buff, "%d", d->cmds); return;
    case 7: sprintf(buff, "%d", d->input); return;
    case 8: sprintf(buff, "%d", d->output); return;
    default: *buff='\0';
  }
}

FUNCTION(idle)   { io_list(buff, privs, args[0], 1); }
FUNCTION(onfor)  { io_list(buff, privs, args[0], 2); }
FUNCTION(host)   { io_list(buff, privs, args[0], 3); }
FUNCTION(concid) { io_list(buff, privs, args[0], 4); }
FUNCTION(port)   { io_list(buff, privs, args[0], 5); }
FUNCTION(cmds)   { io_list(buff, privs, args[0], 6); }
FUNCTION(input)  { io_list(buff, privs, args[0], 7); }
FUNCTION(output) { io_list(buff, privs, args[0], 8); }

FUNCTION(lwho) {
  DESC *d;
  char *s=buff;
 
  *buff=0;
  for(d=descriptor_list;d;d=d->next)
    if((d->flags & C_CONNECTED) && (controls(privs, d->player, POW_WHO) ||
       could_doit(privs, d->player, A_HLOCK)))
      ADDWORD(buff, s, sprintf(s, "#%d", d->player));
}

/* Set 7: Database Values ****************************************************/

FUNCTION(num) {
  sprintf(buff, "#%d", match_thing(privs,args[0],MATCH_EVERYTHING));
}

FUNCTION(pnum) {
  dbref thing=lookup_player(args[0]);
  sprintf(buff, "#%d", (thing == AMBIGUOUS) ? NOTHING : thing);
}

FUNCTION(name) {
  INIT_MATCH(it, args[0]);
  strcpy(buff, it==NOTHING?"":db[it].name);
}

FUNCTION(fullname) {
  INIT_MATCH(it, args[0]);
  char *s;

  strcpy(buff, it==NOTHING?"":db[it].name);
  if(Typeof(it) == TYPE_EXIT && *(s=atr_get(it, A_ALIAS)) &&
     strlen(s)+strlen(buff) <= 8000)
    sprintf(buff+strlen(buff), ";%s", s);
}

FUNCTION(flags) {
  INIT_MATCH(thing, args[0]);
  
  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else
    strcpy(buff, unparse_flags(privs, thing, 0));
}

FUNCTION(unparse) {
  INIT_MATCH(thing, args[0]);

  if(thing == NOTHING && *args[0] == '#')
    thing=atoi(args[0]+1);

  strcpy(buff, unparse_object(doer, thing));
}

FUNCTION(owner) {
  ATTR *attr;
  dbref thing;
  char *s;

  /* return owner of object */
  if(nargs == 1 && !(s=strchr(args[0], '/'))) {
    if((thing=match_thing(privs, args[0], MATCH_EVERYTHING)) != NOTHING)
      thing=db[thing].owner;
    sprintf(buff, "#%d", thing);
    return;
  }

  /* Find object to which an attribute is defined upon.
   * A #-1 means it is a builtin attribute. */

  if(nargs == 1)
    *s++='\0';
  else
    s=args[1];

  if((thing=match_thing(privs, args[0], MATCH_EVERYTHING)) == NOTHING ||
     !(attr=atr_str(thing, s)))
    strcpy(buff, "#-1");
  else if(can_see_atr(privs, thing, attr))
    sprintf(buff, "#%d", attr->obj);  /* note: not owner of object! */
  else
    strcpy(buff, ERR_PERM);
}

FUNCTION(sessions) {
  INIT_CMATCH(thing, args[0], POW_FUNCTIONS);

  if(thing == NOTHING)
    strcpy(buff, ERR_PERM);
  else if(Typeof(thing) != TYPE_PLAYER)
    strcpy(buff, ERR_PLAYER);
  else
    sprintf(buff, "%d", db[thing].data->sessions);
}

FUNCTION(steps) {
  INIT_CMATCH(thing, args[0], POW_FUNCTIONS);

  if(thing == NOTHING)
    strcpy(buff, ERR_PERM);
  else if(Typeof(thing) != TYPE_PLAYER)
    strcpy(buff, ERR_PLAYER);
  else
    sprintf(buff, "%d", db[thing].data->steps);
}

FUNCTION(age) {
  INIT_CMATCH(thing, args[0], POW_FUNCTIONS);

  if(thing == NOTHING)
    strcpy(buff, ERR_PERM);
  else if(Typeof(thing) != TYPE_PLAYER)
    strcpy(buff, ERR_PLAYER);
  else
    sprintf(buff, "%d", db[thing].data->age);
}

FUNCTION(cols) {
  INIT_MATCH(player, args[0]);
  
  if(player == NOTHING)
    strcpy(buff, "#-1");
  else
    sprintf(buff, "%d", get_cols(NULL, player));
}

FUNCTION(rows) {
  INIT_MATCH(player, args[0]);
  
  if(player == NOTHING)
    strcpy(buff, "#-1");
  else
    sprintf(buff, "%d", get_rows(NULL, player));
}

FUNCTION(term) {
  INIT_CMATCH(player, args[0], POW_FUNCTIONS);
  DESC *d;
  int i;

  if(player == NOTHING) {
    strcpy(buff, ERR_PERM);
    return;
  }
  player=db[player].owner;
  d=match_desc(player);

  for(i=0;term_config[i].name;i++)
    if(string_prefix(term_config[i].name, args[1]))
      break;

  if(!term_config[i].name)
    strcpy(buff, "#-1 No such term setting.");
  else
    switch(i) {
    case 0: sprintf(buff, "%d", get_rows(d, player)); break;
    case 1: sprintf(buff, "%d", get_cols(d, player)); break;
    case 9:
      sprintf(buff, "%d", d?((d->flags & C_TINYFUGUE)?1:0):0);
      break;
    default:
      sprintf(buff, "%d", (db[player].data->term & term_config[i].flag)?1:0);
    }
}

FUNCTION(type) {
  INIT_MATCH(thing, args[0]);
  strcpy(buff, (thing == NOTHING)?"#-1":typenames[Typeof(thing)]);
}

FUNCTION(class) {
  INIT_MATCH(thing, args[0]);
  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else if(Typeof(thing) == TYPE_PLAYER)
    strcpy(buff, classnames[class(thing)]);
  else
    strcpy(buff, typenames[Typeof(thing)]);
}

FUNCTION(rank) {
  INIT_MATCH(thing, args[0]);
  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else
    sprintf(buff, "%d", class(db[thing].owner));
}

FUNCTION(immortal) {
  INIT_MATCH(thing, args[0]);
  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else
    sprintf(buff, "%d", Immortal(thing));
}

FUNCTION(gender) {
  INIT_MATCH(it, args[0]);

  char *types[4]={"Male", "Female", "Neuter", "Plural"};

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) && !nearby(privs, it))
    strcpy(buff, ERR_LOC);
  else
    strcpy(buff, types[get_gender(it)]);
}

static void gender_substitute(buff, privs, name, type)
char *buff, *name;
dbref privs;
int type;
{
  INIT_MATCH(it, name);

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) && !nearby(privs, it))
    strcpy(buff, ERR_LOC);
  else
    strcpy(buff, gender_subst(it, type));
}

FUNCTION(subj)  { gender_substitute(buff, privs, args[0], SUBJ); }
FUNCTION(objn)  { gender_substitute(buff, privs, args[0], OBJN); }
FUNCTION(poss)  { gender_substitute(buff, privs, args[0], POSS); }
FUNCTION(aposs) { gender_substitute(buff, privs, args[0], APOSS); }

FUNCTION(rmatch) {
  INIT_CMATCH(who, args[0], POW_FUNCTIONS);
  if(who == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "#%d", match_thing(who, args[1], MATCH_EVERYTHING));
}

FUNCTION(nearby) {
  INIT_MATCH(obj1, args[0]);
  INIT_MATCH(obj2, args[1]);
  sprintf(buff, "%d", (locatable(privs, obj1) && locatable(privs, obj2) &&
          nearby(obj1, obj2)));
}

FUNCTION(sees) {
  INIT_MATCH(obj1, args[0]);
  INIT_MATCH(obj2, args[1]);

  if(obj1 == NOTHING || obj2 == NOTHING)
    strcpy(buff, "#-1");
  else if((!controls(privs, obj1, POW_FUNCTIONS) && !nearby(privs, obj1)) ||
          (!controls(privs, obj2, POW_FUNCTIONS) && !nearby(privs, obj2)))
    strcpy(buff, ERR_LOC);
  else
    sprintf(buff, "%d", can_see(obj1, obj2));
}

FUNCTION(loc) {
  INIT_MATCH(it, args[0]);
  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!locatable(privs, it))
    strcpy(buff, ERR_LOC);
  else
    sprintf(buff, "#%d", db[it].location);
}

FUNCTION(room) {
  INIT_MATCH(it, args[0]);
  int dep, var=it;

  if(it == NOTHING) {
    strcpy(buff, "#-1");
    return;
  } if(!locatable(privs, it)) {
    strcpy(buff, ERR_LOC);
    return;
  }

  for(dep=0;dep < 100;dep++) {
    var=db[var].location;
    if(Invalid(var) || Typeof(var) == TYPE_ROOM)
      break;
    if(var == db[var].location) {
      var=NOTHING;
      break;
    }
  }

  sprintf(buff, "#%d", var);
}

FUNCTION(rloc) {
  INIT_MATCH(it, args[0]);
  int dep, var=it, limit=atoi(args[1]);

  if(it == NOTHING) {
    strcpy(buff, "#-1");
    return;
  } if(!locatable(privs, it)) {
    strcpy(buff, ERR_LOC);
    return;
  }

  /* set a reasonable limit to the recursion */
  if(limit > 100)
    limit=100;

  for(dep=0;dep < limit;dep++) {
    var=db[var].location;
    if(Invalid(var) || Typeof(var) == TYPE_ROOM)
      break;
    if(var == db[var].location) {
      var=NOTHING;
      break;
    }
  }

  sprintf(buff, "#%d", var);
}

FUNCTION(con) {
  INIT_MATCH(it, args[0]);

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) && !nearby(privs, it))
    strcpy(buff, ERR_LOC);
  else {
    DOLIST(it, db[it].contents)
      if((controls(privs, it, POW_FUNCTIONS) || can_see(privs, it)) &&
         (!IS(it, TYPE_THING, THING_MONSTER) || power(privs, POW_COMBAT)))
        break;
    sprintf(buff, "#%d", it);
  }
}

FUNCTION(lcon) {
  INIT_MATCH(it, args[0]);
  char *s=buff;

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) && !nearby(privs, it))
    strcpy(buff, ERR_LOC);
  else {
    *buff=0;
    DOLIST(it, db[it].contents)
      if((controls(privs, it, POW_FUNCTIONS) || can_see(privs, it)) &&
         (!IS(it, TYPE_THING, THING_MONSTER) || power(privs, POW_COMBAT)))
        ADDWORD(buff, s, sprintf(s, "#%d", it));
  }
}

FUNCTION(plcon) {
  INIT_MATCH(it, args[0]);
  char *s=buff;

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) && !nearby(privs, it))
    strcpy(buff, ERR_LOC);
  else {
    *buff=0;
    DOLIST(it, db[it].contents)
      if(Typeof(it) == TYPE_PLAYER && (controls(privs, it, POW_FUNCTIONS) ||
         can_see(privs, it)))
        ADDWORD(buff, s, sprintf(s, "#%d", it));
  }
}

FUNCTION(exit) {
  INIT_MATCH(it, args[0]);

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) && !nearby(privs, it))
    strcpy(buff, ERR_LOC);
  else {
    DOLIST(it, db[it].exits)
      if(controls(privs, it, POW_FUNCTIONS) || can_see(privs, it))
        break;
    sprintf(buff, "#%d", it);
  }
}

FUNCTION(lexits) {
  INIT_MATCH(it, args[0]);
  char *s=buff;

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) && !nearby(privs, it))
    strcpy(buff, ERR_LOC);
  else {
    *buff=0;
    DOLIST(it, db[it].exits)
      if(controls(privs, it, POW_EXAMINE) || can_see(privs, it))
        ADDWORD(buff, s, sprintf(s, "#%d", it));
  }
}

FUNCTION(randexit) {
  INIT_CMATCH(thing, args[0], POW_FUNCTIONS);
  if(thing == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "#%d", random_exit(thing, NOTHING));
}

FUNCTION(next) {
  INIT_MATCH(it, args[0]);

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if((!controls(privs, it, POW_FUNCTIONS) &&
     !controls(privs, db[it].location, POW_FUNCTIONS) &&
     !nearby(privs, db[it].location)))
    strcpy(buff, ERR_LOC);
  else {
    DOLIST(it, db[it].next)
      if((controls(privs, it, POW_FUNCTIONS) || can_see(privs, it)) &&
         (!IS(it, TYPE_THING, THING_MONSTER) || power(privs, POW_COMBAT)))
        break;
    sprintf(buff, "#%d", it);
  }
}

FUNCTION(objlist) {
  INIT_MATCH(it, args[0]);
  char *s=buff;

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) &&
     !controls(privs, db[it].location, POW_FUNCTIONS) &&
     !nearby(privs, db[it].location))
    strcpy(buff, ERR_LOC);
  else {
    *buff=0;
    DOLIST(it, it)
      if((controls(privs, it, POW_FUNCTIONS) || can_see(privs, it)) &&
         (!IS(it, TYPE_THING, THING_MONSTER) || power(privs, POW_COMBAT)))
        ADDWORD(buff, s, sprintf(s, "#%d", it));
  }
}

FUNCTION(link) {
  INIT_CMATCH(it, args[0], POW_FUNCTIONS);
  if(it == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "#%d", db[it].link);
}

FUNCTION(linkup) {
  INIT_MATCH(thing, args[0]);
  char *s=buff;
  dbref i;

  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else {
    *buff=0;
    for(i=0;i<db_top;i++)
      if(db[i].link == thing && (controls(privs, i, POW_FUNCTIONS) ||
         controls(privs, thing, POW_FUNCTIONS)))
        ADDWORD(buff, s, sprintf(s, "#%d", i));
  }
}

FUNCTION(entrances) {
  INIT_MATCH(thing, args[0]);
  char *s=buff;
  dbref i;

  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else {
    *buff=0;
    for(i=0;i<db_top;i++)
      if(db[i].link == thing && Typeof(i) == TYPE_EXIT && (controls(privs, i,
         POW_FUNCTIONS) || controls(privs, thing, POW_FUNCTIONS)))
        ADDWORD(buff, s, sprintf(s, "#%d", i));
  }
}

FUNCTION(oexit) {
  INIT_MATCH(thing, args[0]);
  char *s=buff;
  dbref it;

  if(thing == NOTHING || Typeof(thing) != TYPE_EXIT)
    strcpy(buff, "#-1");
  else if(!controls(privs, thing, POW_FUNCTIONS))
    strcpy(buff, ERR_PERM);
  else if(db[thing].link == NOTHING)
    *buff=0;
  else {
    it=(db[thing].link == HOME) ? db[privs].link : db[thing].link;

    *buff=0;
    DOLIST(it, db[it].exits)
      if(db[it].link == db[thing].location && (controls(privs, it,
         POW_FUNCTIONS) || can_see(privs, it)))
        ADDWORD(buff, s, sprintf(s, "#%d", it));
  }
}

FUNCTION(powers) {
  INIT_MATCH(player, args[0]);

  if(player == NOTHING)
    strcpy(buff, "#-1");
  else if(RESTRICT_HASPOW && !controls(privs, player, POW_EXAMINE))
    strcpy(buff, ERR_PERM);
  else
    disp_powers(buff, db[player].owner);
}

FUNCTION(controls) {
  INIT_MATCH(player, args[0]);
  INIT_MATCH(object, args[1]);
  int k;

  /* Determine power to check */
  if(nargs == 3) {
    for(k=0;k<NUM_POWS;k++)
      if(!strcasecmp(powers[k].name, args[2]))
        break;
  } else
    k=POW_MODIFY;

  if(player == NOTHING || object == NOTHING)
    strcpy(buff, "#-1");
  else if(k == NUM_POWS)
    strcpy(buff, "#-1 No such power.");
  else if(nargs == 3 && RESTRICT_HASPOW && (!controls(privs, player,
          POW_EXAMINE) || !controls(privs, object, POW_EXAMINE)))
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", controls(player, object, k));
}

FUNCTION(haspow) {
  INIT_MATCH(player, args[0]);
  int k;

  for(k=0;k<NUM_POWS;k++)
    if(!strcasecmp(powers[k].name, args[1]))
      break;

  if(player == NOTHING)
    strcpy(buff, "#-1");
  else if(k == NUM_POWS)
    strcpy(buff, "#-1 No such power.");
  else if(RESTRICT_HASPOW && !controls(privs, player, POW_EXAMINE))
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", power(player, k));
}

FUNCTION(parents) {
  INIT_MATCH(thing, args[0]);
  char *s=buff;
  dbref *ptr;

  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else {
    *buff=0;
    for(ptr=db[thing].parents;ptr && *ptr != NOTHING;ptr++)
      if(controls(privs, thing, POW_FUNCTIONS) || controls(privs, *ptr,
         POW_FUNCTIONS))
        ADDWORD(buff, s, sprintf(s, "#%d", *ptr));
  }
}

FUNCTION(children) {
  INIT_MATCH(thing, args[0]);
  char *s=buff;
  dbref *ptr;

  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else {
    *buff=0;
    for(ptr=db[thing].children;ptr && *ptr != NOTHING;ptr++)
      if(controls(privs, thing, POW_FUNCTIONS) || controls(privs, *ptr,
         POW_FUNCTIONS))
        ADDWORD(buff, s, sprintf(s, "#%d", *ptr));
  }
}

FUNCTION(is_a) {
  INIT_MATCH(thing, args[0]);
  INIT_MATCH(parent, args[1]);

  if(thing == NOTHING || parent == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, thing, POW_FUNCTIONS) && !controls(privs, parent,
          POW_FUNCTIONS))
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", is_a(thing, parent));
}

FUNCTION(has_a) {
  INIT_MATCH(room, args[0]);
  INIT_MATCH(parent, args[1]);
  dbref it;
  char *s=buff;

  if(room == NOTHING || parent == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, room, POW_FUNCTIONS) && !nearby(privs, room))
    strcpy(buff, ERR_LOC);
  else {
    *buff=0;
    DOLIST(it, db[room].contents)
      if(is_a(it, parent) && (controls(privs, it, POW_FUNCTIONS) ||
         controls(privs, parent, POW_FUNCTIONS)))
        ADDWORD(buff, s, sprintf(s, "#%d", it));
  }
}

FUNCTION(relay) {
  INIT_CMATCH(thing, args[0], POW_FUNCTIONS);
  char *s=buff;
  dbref *ptr;

  if(thing == NOTHING)
    strcpy(buff, ERR_PERM);
  else {
    *buff=0;
    for(ptr=db[thing].relay;ptr && *ptr != NOTHING;ptr++)
      ADDWORD(buff, s, sprintf(s, "#%d", thing));
  }
}

FUNCTION(zone) {
  INIT_MATCH(thing, args[0]);
  char *s=buff;
  dbref *ptr;

  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else {
    *buff=0;
    for(ptr=db[mainroom(thing)].zone;ptr && *ptr != NOTHING;ptr++)
      ADDWORD(buff, s, sprintf(s, "#%d", *ptr));
  }
}

FUNCTION(inzone) {
  INIT_CMATCH(it, args[0], POW_FUNCTIONS);
  char *s=buff;
  dbref i;

  if(it == NOTHING)
    strcpy(buff, ERR_PERM);
  else {
    *buff=0;
    for(i=0;i<db_top;i++)
      if(Typeof(i) == TYPE_ROOM && inlist(db[i].zone, it))
        ADDWORD(buff, s, sprintf(s, "#%d", i));
  }
}

FUNCTION(zwho) {
  INIT_CMATCH(it, args[0], POW_FUNCTIONS);
  char *s=buff;
  dbref i;

  if(it == NOTHING)
    strcpy(buff, ERR_PERM);
  else {
    *buff=0;
    for(i=0;i<db_top;i++)
      if(Typeof(i) == TYPE_PLAYER && inlist(db[mainroom(i)].zone, it))
        ADDWORD(buff, s, sprintf(s, "#%d", i));
  }
}

FUNCTION(plane) {
  INIT_CMATCH(thing, args[0], POW_EXAMINE);

  if(thing == NOTHING || !power(privs, POW_DARK))
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", db[thing].plane);
}


/* Set 8: Database Attributes ************************************************/

FUNCTION(v) {
  if(strlen(args[0]) > 1) {   /* not a null or 1-character string. */
    ATTR *attr;
    
    if(!(attr=atr_str(privs, args[0])))
      *buff='\0';
    else if(can_see_atr(privs, privs, attr))
      strcpy(buff, atr_get_obj(privs, attr->obj, attr->num));
    else
      *buff='\0';
    return;
  }

  switch(tolower(*args[0])) {
    case '0' ... '9': strcpy(buff, wptr[*args[0]-'0']); return;
    case 'n': strcpy(buff, db[doer].name); return;
    case '#': sprintf(buff, "#%d", doer); return;
    case '!': sprintf(buff, "#%d", privs); return;
    case 'u': strcpy(buff, db[privs].name); return;
    case 'l': sprintf(buff, "#%d", locatable(privs, doer) ?
                      db[doer].location:NOTHING); return;
    default: *buff='\0';
  }
}

FUNCTION(get) {
  ATTR *attr;
  dbref thing;
  char *s;

  /* backward compatibility.. get(obj/attr) */
  if(nargs == 1) {
    if(!(s=strchr(args[0], '/'))) {
      strcpy(buff, "#-1 No attribute name specified.");
      return;
    }
    *s++='\0';
  } else
    s=args[1];

  if((thing=match_thing(privs, args[0], MATCH_EVERYTHING)) == NOTHING)
    strcpy(buff, "#-1");
  else if(!(attr=atr_str(thing, s)))
    *buff='\0';
  else if(can_see_atr(privs, thing, attr))
    strcpy(buff, atr_get_obj(thing, attr->obj, attr->num));
  else
    strcpy(buff, ERR_PERM);
}

FUNCTION(zoneattr) {
  ATTR *attr;
  dbref thing, *zone;
  char *s;

  /* also allow zoneattr(obj/attr) */
  if(nargs == 1) {
    if(!(s=strchr(args[0], '/'))) {
      strcpy(buff, "#-1 No attribute name specified.");
      return;
    }
    *s++='\0';
  } else
    s=args[1];

  if((thing=match_thing(privs, args[0], MATCH_EVERYTHING)) == NOTHING) {
    strcpy(buff, "#-1");
    return;
  }

  /* we only check zones that we can see attributes on */
  for(zone=db[mainroom(thing)].zone;zone && *zone != NOTHING;zone++)
    if((attr=atr_str(*zone, s)) && can_see_atr(privs, *zone, attr)) {
      strcpy(buff, atr_get_obj(*zone, attr->obj, attr->num));
      return;
    }

  *buff='\0';
  return;
}

FUNCTION(isattr) {
  INIT_MATCH(thing, args[0]);
  ATTR *attr;

  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else
    sprintf(buff, "%d", (attr=atr_str(thing, args[1])) &&
                         can_see_atr(privs, thing, attr));
}

FUNCTION(hasattr) {
  INIT_MATCH(thing, args[0]);
  ATTR *attr;

  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else
    sprintf(buff, "%d", (attr=atr_str(thing, args[1])) &&
            can_see_atr(privs, thing, attr) &&
            *(atr_get_obj(thing, attr->obj, attr->num)));
}

FUNCTION(attropts) {
  ATTR *attr;
  dbref thing;
  char *s;

  if(nargs == 1) {
    if(!(s=strchr(args[0], '/'))) {
      strcpy(buff, "#-1 No attribute name specified.");
      return;
    }
    *s++='\0';
  } else
    s=args[1];

  if((thing=match_thing(privs, args[0], MATCH_EVERYTHING)) == NOTHING)
    strcpy(buff, "#-1");
  else if(!(attr=atr_str(thing, s)))
    *buff='\0';
  else if(can_see_atr(privs, thing, attr))
    strcpy(buff, unparse_attrflags(attr->flags));
  else
    strcpy(buff, ERR_PERM);
}

FUNCTION(hasflag) {
  INIT_MATCH(thing, args[0]);
  ATTR *attr;
  char *s=strchr(args[0], '/');
  int a;

  if((nargs == 2 && s) || nargs == 3) {
    if(nargs == 2)
      *s++='\0';
    else
      s=args[1];
  } else
    s=NULL;

  if((thing=match_thing(privs, args[0], MATCH_EVERYTHING)) == NOTHING) {
    strcpy(buff, "#-1");
    return;
  }

  /* search for a flag on a particular attribute */
  if(s) {
    for(a=0;a<16;a++)
      if(string_prefix(attr_options[a], args[nargs-1]))
        break;
    if(a == 16)
      strcpy(buff, "#-1 No such attribute option.");
    else if(!(attr=atr_str(thing, s)) || !can_see_atr(privs, thing, attr))
      strcpy(buff, "0");
    else
      sprintf(buff, "%d", (attr->flags & (1 << a)) ? 1:0);
    return;
  }

  /* search for a flag on an object */

  for(a=0;flaglist[a].name;a++)
    if((strlen(args[1]) == 1 && flaglist[a].flag == *args[1]) ||
       (strlen(args[1]) > 1 && string_prefix(flaglist[a].name, args[1])))
      break;
  if(!flaglist[a].name)
    strcpy(buff, "#-1 No such flag.");
  else
    /* Don't show 'Connected' flag if player is hidden! */
    if(flaglist[a].objtype == TYPE_PLAYER && flaglist[a].bits ==
       PLAYER_CONNECT && !(could_doit(privs, thing, A_HLOCK) ||
       controls(privs, thing, POW_WHO)))
      strcpy(buff, "0");
  else
    sprintf(buff, "%d",
      (!flaglist[a].bits && Typeof(thing) == flaglist[a].objtype) ||
      (flaglist[a].bits && (flaglist[a].objtype == NOTYPE ||
       flaglist[a].objtype == Typeof(thing)) &&
       (db[thing].flags & flaglist[a].bits)));
}

FUNCTION(lattr) {
  INIT_MATCH(obj, args[0]);
  char *s=buff;
  ALIST *a;
  ATTR *atr;

  *buff=0;
  for(a=db[obj].attrs;a;a=a->next)
    if((atr=atr_num(a->obj, a->num)) && can_see_atr(privs, obj, atr)) {
      if(s-buff > 7930) {
        strcpy(s, " ...");
        return;
      }
      if(s != buff)
        *s++=' ';
      strcpy(s, atr->name);
      s=strchr(s, 0);
    }
}

FUNCTION(lattrdef) {
  INIT_MATCH(it, args[0]);
  char *s=buff;
  ATRDEF *k;

  *buff=0;
  if((db[it].flags & VISIBLE) || controls(privs, it, POW_EXAMINE))
    for(k=db[it].atrdefs;k;k=k->next) {
      if(s-buff > 7930) {
        strcpy(s, " ...");
        return;
      }
      if(s != buff)
        *s++=' ';
      strcpy(s, k->atr->name);
      s=strchr(s, 0);
    }
}

FUNCTION(setq) {
  int a=atoi(args[0]);

  if(a < 0 || a > 9)
    strcpy(buff, ERR_ENV);
  else {
    strcpy(buff, "");
    strcpy(wptr[a], args[1]);
  }
}

FUNCTION(setr) {
  int a=atoi(args[0]);

  if(a < 0 || a > 9)
    strcpy(buff, ERR_ENV);
  else {
    strcpy(buff, args[1]);
    strcpy(wptr[a], args[1]);
  }
}

FUNCTION(s) { pronoun_substitute(buff, privs, doer, args[0]); }

FUNCTION(s_as) {
  INIT_CMATCH(newdoer, args[1], POW_MODIFY);
  INIT_CMATCH(newprivs, args[2], POW_MODIFY);

  if(newdoer == NOTHING || newprivs == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    pronoun_substitute(buff, newprivs, newdoer, args[0]);
}

FUNCTION(s_as_with) {
  INIT_CMATCH(newdoer, args[1], POW_MODIFY);
  INIT_CMATCH(newprivs, args[2], POW_MODIFY);

  char *sptr[10];
  int a;

  if(newdoer == NOTHING || newprivs == NOTHING) {
    strcpy(buff, ERR_PERM);
    return;
  }

  /* preserve local registers */
  for(a=0;a<10;a++) {
    SPTR(sptr[a], wptr[a]);
    strcpy(wptr[a], (a < nargs-3)?args[a+3]:"");
  }

  pronoun_substitute(buff, newprivs, newdoer, args[0]);

  /* restore local registers */
  for(a=0;a<10;a++) {
    strcpy(wptr[a], sptr[a]);
    free(sptr[a]);
  }
}

FUNCTION(parse_lock) {
  INIT_MATCH(player, args[0]);
  INIT_CMATCH(thing, args[1], POW_EXAMINE);

  if(player == NOTHING || thing == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", eval_boolexp(player, thing, args[2]));
}

FUNCTION(create) {
  INIT_CMATCH(it, args[0], POW_FUNCTIONS);
  if(it == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%ld", db[it].create_time);
}

FUNCTION(modify) {
  INIT_CMATCH(it, args[0], POW_FUNCTIONS);
  if(it == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%ld", db[it].mod_time);
}

FUNCTION(pennies) {
  INIT_CMATCH(who, args[0], POW_MONEY);
  if(who == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", db[who].pennies);
}

FUNCTION(stats) {
  dbref owner=AMBIGUOUS;
  int i, total, obj[8], pla[NUM_CLASSES];
  
  if(nargs) {
    if((owner=lookup_player(args[0])) == NOTHING) {
      strcpy(buff, ERR_PLAYER);
      return;
    }

    if(owner == AMBIGUOUS)
      owner=db[privs].owner;
    else if(!controls(privs, owner, POW_STATS)) {
      strcpy(buff, ERR_PERM);
      return;
    }
  }

  calc_stats(owner, &total, obj, pla);
  sprintf(buff, "%d %d", db_top, total);
  for(i=0;i<8;i++)
    sprintf(buff+strlen(buff), " %d", obj[i]);
}

FUNCTION(objmem) {
  INIT_CMATCH(thing, args[0], POW_STATS);
  if(thing == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", mem_usage(thing));
}

FUNCTION(playmem) {
  INIT_CMATCH(thing, args[0], POW_STATS);
  if(thing == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", playmem_usage(db[thing].owner));
}


/* Set 9: String Utilities **************************************************/

FUNCTION(strlen) { sprintf(buff, "%d", (int)strlen(ansi_strip(args[0]))); }
FUNCTION(edit) { edit_string(buff, ansi_strip(args[0]), args[1], args[2]); }
FUNCTION(okitem) { sprintf(buff, "%d", ok_itemname(args[0])); }
FUNCTION(secure) { strcpy(buff, secure_string(ansi_strip(args[0]))); }
FUNCTION(ascii) { sprintf(buff, "%d", *args[0]); }
FUNCTION(match) { sprintf(buff, "%d", wild_match(args[0], args[1], 1, 0)); }

FUNCTION(comp) {
  int i=strcmp(args[0], args[1]);
  sprintf(buff, "%d", (i>0)?1 : (i<0)?-1 : 0);
}

FUNCTION(strcat) {
  int i;

  *buff='\0';
  for(i=0;i<nargs;i++) {
    strncpy(buff+strlen(buff), args[i], 8001-strlen(buff));
    buff[8000]='\0';
  }
}

FUNCTION(wmatch) {
  DELIM(delim, 2);
  char *r, *t=args[0];
  int wcount=1;

  while((r=parse_up(&t, delim))) {
    if(!*r && delim == ' ')
      continue;
    if(wild_match(r, args[1], 0, 0)) {
      sprintf(buff, "%d", wcount);
      return;
    }
    wcount++;
  }
  strcpy(buff, "0");
}

FUNCTION(matchall) {
  DELIM(delim, 2);
  char *r, *s=buff, *t=args[0];
  int wcount=1;

  *buff=0;
  while((r=parse_up(&t, delim))) {
    if(!*r && delim == ' ')
      continue;
    if(wild_match(r, args[1], 0, 0))
      ADDWORD(buff, s, sprintf(s, "%d", wcount));
    wcount++;
  }
}
  
FUNCTION(regmatch) {
  char *sptr[10];
  int a=(nargs==3)?atoi(args[2]):0, b, result;

  if(a < 0 || a > 9) {
    strcpy(buff, ERR_ENV);
    return;
  }

  /* preserve copy of old environment */
  for(b=a;b<10;b++)
    SPTR(sptr[b], wptr[b]);

  if(!(result=wild_match(args[0], args[1], 0, a+1)))
    for(b=a;b<10;b++)
      strcpy(wptr[b], sptr[b]);  /* restore old environment if case failed */

  for(b=a;b<10;b++)
    free(sptr[b]);
  sprintf(buff, "%d", result);
}

FUNCTION(capstr) {
  char *s=ansi_strip(args[0]);

  *buff=toupper(*s);
  if(*s)
    strcpy(buff+1, s+1);
}

FUNCTION(ucstr) {
  char *s=buff, *t=ansi_strip(args[0]);

  while(*t)
    *s++=toupper(*t++);
  *s='\0';
}

FUNCTION(lcstr) {
  char *s=buff, *t=ansi_strip(args[0]);

  while(*t)
    *s++=tolower(*t++);
  *s='\0';
}

FUNCTION(pos) {
  char *s=strstr(args[1], ansi_strip(args[0]));
  sprintf(buff, "%d", s?(int)(s-args[1]+1):0);
}

FUNCTION(lpos) {
  char c=(*args[1] > 32 && *args[1] < 127) ? *args[1] : ' ';
  char *r=ansi_strip(args[0])-1, *s=buff, *t=r;

  *buff=0;
  while((r=strchr(r+1, c)))
    ADDWORD(buff, s, sprintf(s, "%d", (int)(r-t)));
}

FUNCTION(mid) {
  int l=atoi(args[1]), len=atoi(args[2]);
  char *s=ansi_strip(args[0]);

  if(l < 0 || len < 0)
    strcpy(buff, ERR_RANGE);
  else if(l >= strlen(s))
    *buff='\0';
  else {
    if(len > strlen(s+l))
      len=strlen(s+l);
    strncpy(buff, s+l, len);
    buff[len]='\0';
  }
}

FUNCTION(first) {
  DELIM(delim, 1);
  char *s;

  if((s=strchr(args[0], delim)))
    *s='\0';
  strcpy(buff, args[0]);
}

FUNCTION(rest) {
  DELIM(delim, 1);
  char *s;

  if((s=strchr(args[0], delim))) {
    if(delim == ' ')
      while(*(s+1) == ' ')
        s++;
    strcpy(buff, s+1);
  } else
    *buff='\0';
}

FUNCTION(extract) {
  DELIM(delim, 3);
  char *s=args[0], *t=args[0];
  int first=atoi(args[1]), len=first+atoi(args[2])-1, wcount=1;

  if(len < 1 || len < first) {
    *buff=0;
    return;
  }

  while((t=strchr(t, delim))) {
    if(wcount >= len)
      break;
    t++;
    if(delim == ' ')
      while(*t == delim)
        t++;
    if(!*t || ++wcount == first)
      s=t;
  }

  if(wcount < first)
    *buff=0;
  else {
    if(t)
      *t='\0';
    strcpy(buff, s);
  }
}

FUNCTION(delete) {
  int len=atoi(args[1]), l=atoi(args[2]);

  if(len < 0 || len > 7999)
    strcpy(buff, ERR_RANGE);
  else if(l < 1)
    strcpy(buff, args[0]);
  else {
    strncpy(buff, args[0], len);
    if(len+l < strlen(args[0]))
      strcpy(buff+len, args[0]+len+l);
    else
      buff[len]='\0';
  }
}

FUNCTION(remove) {
  DELIM(delim, 3);
  char *r=args[0], *s=args[0];
  int first=atoi(args[1]), len=first+atoi(args[2]), wcount=1;

  *buff=0;
  if(len < first || len <= 1) {
    strcpy(buff, args[0]);
    return;
  }

  while((s=strchr(s, delim))) {
    if(++wcount == first) {
      *s++='\0';
      strcpy(buff, r);
    } else
      s++;
    if(delim == ' ')
      while(*s == ' ')
        s++;
    if(wcount == len) {
      if(*buff)
        sprintf(buff+strlen(buff), "%c", delim);
      strcat(buff, s);
    }
  }
}

FUNCTION(insert) {
  DELIM(delim, 3);
  char *r=args[0], *s=args[0];
  int first=atoi(args[1]), wcount=1;

  if(first < 2)
    strcpy(buff, args[2]);  // First word = arg2
  else {
    /* find nth word division */
    while((s=strchr(s, delim))) {
      if(++wcount == first) {
        *s++='\0';
        break;
      } else
        s++;
      if(delim == ' ')
        while(*s == ' ')
          s++;
    }

    /* Copy first half of arg0 to buff */
    sprintf(buff, "%s%c", r, delim);
    strncpy(buff+strlen(buff), args[2], 8001-strlen(buff));
    buff[8000]='\0';
  }

  /* copy second half of arg0 to string */
  if(s && *s) {
    r=strchr(buff, 0);
    *r++=delim;
    strncpy(r, s, 8001-(r-buff));
    buff[8000]='\0';
  }
}

FUNCTION(wcount) {
  DELIM(delim, 1);
  char *s=args[0];
  int wcount=1;
  
  if(!*args[0]) {
    strcpy(buff, "0");
    return;
  }

  while((s=strchr(s, delim))) {
    wcount++;
    s++;
    if(delim == ' ')
      while(*s == ' ')
        s++;
  }

  sprintf(buff, "%d", wcount);
}

FUNCTION(flip) {
  char *s=ansi_strip(args[0]), *p=strchr(s, 0), *r=buff;
  while(p > s)
    *r++=*--p;
  *r='\0';
}

FUNCTION(wordflip) {
  DELIM(delim, 1);
  char *s=args[0], *p=strchr(s, 0), *r=buff;

  *buff=0;
  while(p >= s) {
    while(p >= s && *p != delim)
      p--;
    if(*buff)
      *r++=delim;
    strcpy(r, p+1);
    r=strchr(r, 0);

    if(delim == ' ')
      while(p >= s && *p == ' ')
        *p--='\0';
    else if(p >= s)
      *p--='\0';
  }
}

FUNCTION(scramble) {
  char c, *s=ansi_strip(args[0]);
  int i, j, n=strlen(s);

  strcpy(buff, s);

  for(i=0;i<n;i++) {
    j=rand_num(n);
    c=buff[i];
    buff[i]=buff[j];
    buff[j]=c;
  }
}

FUNCTION(rand) { sprintf(buff, "%d", rand_num(atoi(args[0]))); }

FUNCTION(dice) {
  int total, count, n=atoi(args[0]), die=atoi(args[1]);

  if(n > 100)
    strcpy(buff, ERR_RANGE);
  else {
    for(total=0,count=0;count < n;count++)
      total += rand_num(die);
    sprintf(buff, "%d", total+n);
  }
}

FUNCTION(randword) {
  DELIM(delim, 1);
  char *s=args[0], *t=args[0];
  int wcount=1, num;
  
  if(!*args[0]) {
    *buff='\0';
    return;
  }

  while((s=strchr(s, delim))) {
    wcount++;
    s++;
    if(delim == ' ')
      while(*s == ' ')
        s++;
  }

  num=rand_num(wcount);
  wcount=0;
  s=args[0];

  while((t=strchr(t, delim))) {
    if(wcount >= num)
      break;
    t++;
    if(delim == ' ')
      while(*t == delim)
        t++;
    if(!*t || ++wcount == num)
      s=t;
  }

  if(t)
    *t='\0';
  strcpy(buff, s);
}

FUNCTION(base) {
  unsigned int decimal=0, oldbase=atoi(args[1]), newbase=atoi(args[2]);
  int i, digit, neg=0;
  char temp[40];

  if(oldbase < 2 || oldbase > 36 || newbase < 2 || newbase > 36) {
    strcpy(buff, "#-1 Bases must be between 2 and 36.");
    return;
  }
  
  if(*args[0] == '-') {
    neg=1;
    *args[0]='0';
  }

  for(i=0;args[0][i];i++) {
    decimal *= oldbase;
    if(isdigit(args[0][i]))
      digit=args[0][i]-'0';
    else if(args[0][i] >= 'a' && args[0][i] <= 'z')
      digit=args[0][i]-'a'+10;
    else if(args[0][i] >= 'A' && args[0][i] <= 'Z')
      digit=args[0][i]-'A'+10;
    else {
      strcpy(buff, "#-1 Illegal digit.");
      return;
    }

    if(digit >= oldbase) {
      strcpy(buff, "#-1 Digit out of range.");
      return;
    }
    decimal += digit;
  }

  for(i=0;decimal > 0;i++) {
    digit=(decimal % newbase);

    if(digit < 10)
      temp[i]=digit+'0';
    else
      temp[i]=digit+'a'-10;
    decimal /= newbase;
  }

  if(neg)
    *buff='-';

  for(i--;i >= 0;i--)
    buff[neg++]=temp[i];
  buff[neg]='\0';
}

FUNCTION(inc) {
  char *s, *str=ansi_strip(args[0]);

  if(!*str) {
    strcpy(buff, "1");
    return;
  }
  s=strchr(str, 0)-1;
  if(!isdigit(*s)) {
    strcpy(buff, "#-1 Argument must end in an integer.");
    return;
  }

  while(s > str && (isdigit(*(s-1)) || *(s-1) == '-'))
    s--;
  strcpy(buff, str);
  sprintf(buff+(s-str), tprintf("%d", atoi(s)+1));
}

FUNCTION(dec) {
  char *s, *str=ansi_strip(args[0]);

  if(!*str) {
    strcpy(buff, "-1");
    return;
  }
  s=strchr(str, 0)-1;
  if(!isdigit(*s)) {
    strcpy(buff, "#-1 Argument must end in an integer.");
    return;
  }

  while(s > str && (isdigit(*(s-1)) || *(s-1) == '-'))
    s--;
  strcpy(buff, str);
  sprintf(buff+(s-str), tprintf("%d", atoi(s)-1));
}

FUNCTION(spc) {
  int a, s=atoi(args[0]);

  if(s > 8000)
    s=8000;
  for(a=0;a<s;a++)
    buff[a]=' ';
  buff[a]=0;
}

FUNCTION(lnum) {
  int a, i=(nargs >= 2)?atoi(args[1]):0, x=atoi(args[0]);
  int j=(nargs == 3)?atoi(args[2]):1;
  char *s=buff;

  *buff=0;
  for(a=0;a<x;a++)
    ADDWORD(buff, s, sprintf(s, "%d", i+a*j));
}

FUNCTION(left) {
  char *s=ansi_strip(args[0]);
  int i=strlen(s), len=atoi(args[1]);

  if(len < 1)
    *buff='\0';
  else {
    if(len > i)
      len=i;
    strncpy(buff, s, len);
    buff[len]='\0';
  }
}

FUNCTION(right) {
  char *s=ansi_strip(args[0]);
  int i=strlen(s), len=atoi(args[1]);

  if(len < 1)
    *buff='\0';
  else {
    if(len > i)
      len=i;
    strcpy(buff, s+i-len);
  }
}

FUNCTION(fill) {
  char *s=ansi_strip(args[0]);
  int i=strlen(s), j=0, len=atoi(args[1]);

  if(len < 0 || len > 8000) {
    strcpy(buff, ERR_RANGE);
    return;
  }

  if(len > i)
    while(j < len-i)
      buff[j++]='0';
  strcpy(buff+j, s);
}

FUNCTION(ljust) {
  DELIM(fill, 2);
  char *s=ansi_strip(args[0]);
  int i=strlen(s), len=atoi(args[1]);

  if(len < 0 || len > 8000) {
    strcpy(buff, ERR_RANGE);
    return;
  }

  /* if string is smaller than length, pad with fill character */
  if(len > i) {
    strcpy(buff, s);
    while(i < len)
      buff[i++]=fill;
    buff[i]='\0';
  } else {
    strncpy(buff, s, len);
    buff[len]='\0';
  }
}

FUNCTION(rjust) {
  DELIM(fill, 2);
  char *s=ansi_strip(args[0]);
  int i=strlen(s), j=0, len=atoi(args[1]);

  if(len < 0 || len > 8000) {
    strcpy(buff, ERR_RANGE);
    return;
  }

  if(len > i) {
    while(j < len-i)
      buff[j++]=fill;
    strcpy(buff+j, s);
  } else
    strcpy(buff, s+i-len);
}

FUNCTION(center) {
  DELIM(fill, 2);
  int len=atoi(args[1]);

  if(len < 0 || len > 8000) {
    strcpy(buff, ERR_RANGE);
    return;
  }

  strcpy(buff, center(ansi_strip(args[0]), len, fill));
}

FUNCTION(switch) {
  int i;

  for(i=1;i+1 < nargs;i+=2)
    if(wild_match(args[0], args[i], 1, 0)) {
      strcpy(buff, args[i+1]);
      return;
    }

  if(i < nargs)
    strcpy(buff, args[i]);
  else
    *buff='\0';
}

FUNCTION(repeat) {
  char *s=buff;
  int a;

  *buff=0;
  if(!*args[0])
    return;

  for(a=0;a<atoi(args[1]);a++) {
    strcpy(s, args[0]);
    s=strchr(s, 0);
    if(s-buff+strlen(args[0]) > 8000)
      return;
  }
}

FUNCTION(foreach) {
  DELIM(delim, 2);
  DELIMCH(odelim, 3, delim);
  char *p=buff, *r, *s=args[0], *sptr, temp[8192];

  *buff=0;
  if(!*args[0])
    return;

  SPTR(sptr, wptr[0]);

  while((r=parse_perfect(&s, delim))) {
    strcpy(wptr[0], r);
    pronoun_substitute(temp, privs, doer, args[1]);
    if(!*temp)
      continue;
    if(p != buff) {
      if(p-buff+strlen(temp) > 8000)
        break;
      *p++=odelim;
    }
    strcpy(p, temp);
    p=strchr(p, 0);
    if(func_total > MAX_FUNCTIONS)
      break;
  }

  strcpy(wptr[0], sptr);
  free(sptr);
}

FUNCTION(oper) {
  DELIM(delim, 3);
  char *r, *s=args[0], *sptr[2];

  *buff=0;
  if(!*args[0]) {
    strcpy(buff, args[1]);
    return;
  }

  SPTR(sptr[0], wptr[0]);
  SPTR(sptr[1], wptr[1]);
  strcpy(wptr[1], args[1]);

  while((r=(char *)parse_perfect(&s, delim))) {
    strcpy(wptr[0], r);
    pronoun_substitute(buff, privs, doer, args[2]);
    strcpy(wptr[1], buff);
    if(func_total > MAX_FUNCTIONS)
      break;
  }

  strcpy(wptr[0], sptr[0]);
  strcpy(wptr[1], sptr[1]);
  free(sptr[0]);
  free(sptr[1]);
}

FUNCTION(chr) {
  int chr=atoi(args[0]);

  if(!power(privs, POW_ANIMATION) && (chr < 32 || chr >= 127))
    chr=' ';
  buff[0]=chr;
  buff[1]='\0';
}

FUNCTION(esc) {
  if(!power(privs, POW_ANIMATION))
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "\e[%s", (nargs == 1)?args[0]:"");
}

FUNCTION(csi) {
  if(!power(privs, POW_ANIMATION))
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "\e%s", (nargs == 1)?args[0]:"");
}

FUNCTION(ansi) {
  int color=color_code(7, args[0]);

  if(color & 1024)
    strcpy(buff, rainbowmsg(args[1], color, 7960));
  else {
    sprintf(buff, "%s%s", textcolor(default_color, color), args[1]);
    strcat(buff, textcolor(color, default_color));
  }
}

FUNCTION(ansi_strip) { strcpy(buff, ansi_strip(args[0])); }

/* Set 10: Pueblo HTML Support ***********************************************/

FUNCTION(url_escape) {
  static char *esc_chars="<>#%{}|\\^~[]';/?:@=&\"+";
  char *s=buff, *t;

  for(t=args[0];*t && s-buff < 7998;t++)
    if(strchr(esc_chars, *t)) {
      sprintf(s, "%%%02x", *t);
      s+=3;
    } else
      *s++=(*t == ' ') ? '+' : *t;

  *s='\0';
}

FUNCTION(url_unescape) {
  char temp[3], *s=buff, *t;
  unsigned int c;

  for(t=args[0];*t;t++)
    if(*t == '%' && isxdigit(*(t+1)) && isxdigit(*(t+2))) {
      strncpy(temp, t+1, 2);
      temp[2]='\0';

      if(sscanf(temp, "%x", &c) == 1) {
        *s++=(c < 32 || c > 126) ? ' ':c;
        t+=2;
      }
    } else
      *s++=(*t == '+') ? ' ' : *t;

  *s='\0';
}


/* function flags */
#define F_MATH	0x1	// Math functions are stripped down to 100 characters.
#define F_EVAL	0x2	// Function arguments are automatically evaluated.
// FIX: F_EVAL not currently used.

struct fun {
  char *name;
  void (*fun)();
  int flags:16;		// Function flags (listed above).
  unsigned int loarg:8;	// Minimum #args
  unsigned int hiarg:8;	// Maximum #args (100 = Unlimited)
};

struct fun flist[]={
  {"ABS",	fun_abs,	1, 1},
  {"ACOS",	fun_acos,	1, 1},
  {"ADD",	fun_add,	0, 1,100},
  {"AGE",	fun_age,	0, 1},
  {"ANSI",	fun_ansi,	0, 2},
  {"ANSI_STRIP",fun_ansi_strip,	0, 1},
  {"APOSS",	fun_aposs,	0, 1},
  {"ASCII",	fun_ascii,	0, 1},
  {"ASIN",	fun_asin,	1, 1},
  {"ATAN",	fun_atan,	1, 1},
  {"ATTROPTS",	fun_attropts,	0, 1,2},
  {"BAND",	fun_band,	0, 2},
  {"BASE",	fun_base,	0, 3},
  {"BETWEEN",	fun_between,	1, 3},
  {"BNAND",	fun_bnand,	0, 2},
  {"BOR",	fun_bor,	0, 2},
  {"BXOR",	fun_bxor,	0, 2},
  {"CAPSTR",	fun_capstr,	0, 1},
  {"CEIL",	fun_ceil,	1, 1},
  {"CENTER",	fun_center,	0, 2,3},
  {"CHILDREN",	fun_children,	0, 1},
  {"CHR",	fun_chr,	0, 1},
  {"CLASS",	fun_class,	0, 1},
  {"CMDS",	fun_cmds,	0, 1},
  {"COLS",	fun_cols,	0, 1},
  {"COMP",	fun_comp,	0, 2},
  {"CON",	fun_con,	0, 1},
  {"CONCID",	fun_concid,	0, 1},
  {"CONFIG",	fun_config,	0, 1},
  {"CONTROLS",	fun_controls,	0, 2,3},
  {"COS",	fun_cos,	1, 1},
  {"CPU",	fun_cpu,	0, 0},
  {"CPUTIME",	fun_cputime,	0, 0},
  {"CREATE",	fun_create,	0, 1},
  {"CREDITS",	fun_pennies,	0, 1},
  {"CSI",	fun_csi,	0, 0,1},
  {"DATE",	fun_date,	0, 0,1},
  {"DBTOP",	fun_dbtop,	0, 0},
  {"DEC",	fun_dec,	0, 1},
  {"DEFAULT",	fun_default,	0, 2},
  {"DEGREES",	fun_degrees,	1, 2},
  {"DELETE",	fun_delete,	0, 3},
  {"DICE",	fun_dice,	1, 2},
  {"DIST2D",	fun_dist2d,	1, 4},
  {"DIST3D",	fun_dist3d,	1, 6},
  {"DIV",	fun_div,	0, 2},
  {"E",		fun_e,		1, 0},
  {"EDIT",	fun_edit,	0, 3},
  {"ENTRANCES",	fun_entrances,	0, 1},
  {"ESC",	fun_esc,	0, 0,1},
  {"EQ",	fun_eq,		1, 2},
  {"EXIT",	fun_exit,	0, 1},
  {"EXP",	fun_exp,	1, 1},
  {"EXTRACT",	fun_extract,	0, 3,4},
  {"FACTOR",	fun_factor,	1, 1},
  {"FADD",	fun_fadd,	1, 1,100},
  {"FDIV",	fun_fdiv,	1, 2},
  {"FILL",	fun_fill,	0, 2},
  {"FIRST",	fun_first,	0, 1,2},
  {"FLAGS",	fun_flags,	0, 1},
  {"FLIP",	fun_flip,	0, 1},
  {"FLOOR",	fun_floor,	1, 1},
  {"FMUL",	fun_fmul,	1, 1,100},
  {"FOREACH",	fun_foreach,	0, 2,4},
  {"FSUB",	fun_fsub,	1, 2},
  {"FULLNAME",	fun_fullname,	0, 1},
  {"GENDER",	fun_gender,	0, 1},
  {"GET",	fun_get,	0, 1,2},
  {"GT",	fun_gt,		1, 2},
  {"GTEQ",	fun_gteq,	1, 2},
  {"HASATTR",	fun_hasattr,	0, 2},
  {"HASFLAG",	fun_hasflag,	0, 2,3},
  {"HASPOW",	fun_haspow,	0, 2},
  {"HAS_A",	fun_has_a,	0, 2},
  {"HOME",	fun_link,	0, 1},
  {"HOST",	fun_host,	0, 1},
  {"IDLE",	fun_idle,	0, 1},
  {"IF",	fun_if,		0, 2},
  {"IFELSE",	fun_ifelse,	0, 3},
  {"IMMORTAL",	fun_immortal,	0, 1},
  {"INC",	fun_inc,	0, 1},
  {"INPUT",	fun_input,	0, 1},
  {"INSERT",	fun_insert,	0, 3,4},
  {"INT",	fun_int,	1, 1},
  {"INZONE",	fun_inzone,	0, 1},
  {"ISATTR",	fun_isattr,	0, 2},
  {"IS_A",	fun_is_a,	0, 2},
  {"LAG",	fun_lag,	0, 0},
  {"LAND",	fun_land,	0, 2},
  {"LAST",	fun_last,	0, 1},
  {"LASTOFF",	fun_lastoff,	0, 1},
  {"LATTR",	fun_lattr,	0, 1},
  {"LATTRDEF",	fun_lattrdef,	0, 1},
  {"LCON",	fun_lcon,	0, 1},
  {"LCSTR",	fun_lcstr,	0, 1},
  {"LEFT",	fun_left,	0, 2},
  {"LEXITS",	fun_lexits,	0, 1},
  {"LINK",	fun_link,	0, 1},
  {"LINKUP",	fun_linkup,	0, 1},
  {"LJUST",	fun_ljust,	0, 2,3},
  {"LN",	fun_ln,		1, 1},
  {"LNUM",	fun_lnum,	0, 1,3},
  {"LOC",	fun_loc,	0, 1},
  {"LOG",	fun_log,	1, 1,2},
  {"LOR",	fun_lor,	0, 2},
  {"LPOS",	fun_lpos,	0, 2},
  {"LT",	fun_lt,		1, 2},
  {"LTEQ",	fun_lteq,	1, 2},
  {"LWHO",	fun_lwho,	0, 0},
  {"LXOR",	fun_lxor,	0, 2},
  {"MATCH",	fun_match,	0, 2},
  {"MATCHALL",	fun_matchall,	0, 2,3},
  {"MAX",	fun_max,	0, 1,100},
  {"MID",	fun_mid,	0, 3},
  {"MIN",	fun_min,	0, 1,100},
  {"MOD",	fun_mod,	1, 2},
  {"MODIFY",	fun_modify,	0, 1},
  {"MONEY",	fun_pennies,	0, 1},
  {"MUDNAME",	fun_mudname,	0, 0},
  {"MUL",	fun_mul,	0, 1,100},
  {"NAME",	fun_name,	0, 1},
  {"NEARBY",	fun_nearby,	0, 2},
  {"NEG",	fun_neg,	1, 1},
  {"NEQ",	fun_neq,	1, 2},
  {"NETSTAT",	fun_netstat,	0, 1},
  {"NEXT",	fun_next,	0, 1},
  {"NOT",	fun_not,	1, 1},
  {"NUM",	fun_num,	0, 1},
  {"OBJLIST",	fun_objlist,	0, 1},
  {"OBJMEM",	fun_objmem,	0, 1},
  {"OBJN",	fun_objn,	0, 1},
  {"OEXIT",	fun_oexit,	0, 1},
  {"OKITEM",	fun_okitem,	0, 1},
  {"ONFOR",	fun_onfor,	0, 1},
  {"OPER",	fun_oper,	0, 3,4},
  {"OUTPUT",	fun_output,	0, 1},
  {"OWNER",	fun_owner,	0, 1,2},
  {"PARENTS",	fun_parents,	0, 1},
  {"PARSE_LOCK",fun_parse_lock,	0, 3},
  {"PENNIES",	fun_pennies,	0, 1},
  {"PI",	fun_pi,		1, 0},
  {"PLANE",	fun_plane,	0, 1},
  {"PLAYMEM",	fun_playmem,	0, 1},
  {"PLCON",	fun_plcon,	0, 1},
  {"PNUM",	fun_pnum,	0, 1},
  {"PORT",	fun_port,	0, 1},
  {"POS",	fun_pos,	0, 2},
  {"POSS",	fun_poss,	0, 1},
  {"POW",	fun_pow,	1, 2},
  {"POWERS",	fun_powers,	0, 1},
  {"RAND",	fun_rand,	1, 1},
  {"RANDEXIT",	fun_randexit,	0, 1},
  {"RANDWORD",	fun_randword,	0, 1,2},
  {"RANK",	fun_rank,	0, 1},
  {"REGMATCH",	fun_regmatch,	0, 2,3},
  {"RELAY",	fun_relay,	0, 1},
  {"REMOVE",	fun_remove,	0, 3,4},
  {"REPEAT",	fun_repeat,	0, 2},
  {"REST",	fun_rest,	0, 1,2},
  {"RESTARTS",	fun_restarts,	0, 0},
  {"RIGHT",	fun_right,	0, 2},
  {"RJUST",	fun_rjust,	0, 2,3},
  {"RLOC",	fun_rloc,	0, 2},
  {"RMATCH",	fun_rmatch,	0, 2},
  {"ROOM",	fun_room,	0, 1},
  {"ROUND",	fun_round,	1, 1,2},
  {"ROWS",	fun_rows,	0, 1},
  {"RTIME",	fun_rtime,	0, 0,1},
  {"S",		fun_s,		0, 1},
  {"SCRAMBLE",	fun_scramble,	0, 1},
  {"SECURE",	fun_secure,	0, 1},
  {"SEES",	fun_sees,	0, 2},
  {"SESSIONS",	fun_sessions,	0, 1},
  {"SETQ",	fun_setq,	0, 2},
  {"SETR",	fun_setr,	0, 2},
  {"SHL",	fun_shl,	0, 2},
  {"SHR",	fun_shr,	0, 2},
  {"SIGN",	fun_sign,	1, 1},
  {"SIN",	fun_sin,	1, 1},
  {"SPC",	fun_spc,	0, 1},
  {"SQRT",	fun_sqrt,	1, 1},
  {"STATS",	fun_stats,	0, 0,1},
  {"STEPS",	fun_steps,	0, 1},
  {"STRCAT",	fun_strcat,	0, 2,100},
  {"STRLEN",	fun_strlen,	0, 1},
  {"SUB",	fun_sub,	0, 2},
  {"SUBJ",	fun_subj,	0, 1},
  {"SWITCH",	fun_switch,	0, 2,100},
  {"S_AS",	fun_s_as,	0, 3},
  {"S_AS_WITH",	fun_s_as_with,	0, 3,13},
  {"TAN",	fun_tan,	1, 1},
  {"TERM",	fun_term,	0, 2},
  {"TIME",	fun_time,	0, 0,1},
  {"TIMEOFDAY",	fun_timeofday,	0, 0},
  {"TMA",	fun_tma,	0, 1},
  {"TMF",	fun_tmf,	0, 1},
  {"TML",	fun_tml,	0, 1},
  {"TMS",	fun_tms,	0, 1},
  {"TRUTH",	fun_truth,	1, 1},
  {"TYPE",	fun_type,	0, 1},
  {"UCSTR",	fun_ucstr,	0, 1},
  {"UNPARSE",	fun_unparse,	0, 1},
  {"UPFRONT",	fun_upfront,	0, 0},
  {"UPTIME",	fun_uptime,	0, 0},
  {"URL_ESCAPE",fun_url_escape, 0, 1},
  {"URL_UNESCAPE",fun_url_unescape,0,1},
  {"V",		fun_v,		0, 1},
  {"VERSION",	fun_version,	0, 0},
  {"WCOUNT",	fun_wcount,	0, 1,2},
  {"WDATE",	fun_wdate,	0, 0,1},
  {"WMATCH",	fun_wmatch,	0, 2,3},
  {"WORDFLIP",	fun_wordflip,	0, 1,2},
  {"WTIME",	fun_wtime,	0, 0,1},
  {"XTIME",	fun_xtime,	0, 0,1},
  {"ZONE",	fun_zone,	0, 1},
  {"ZONEATTR",	fun_zoneattr,	0, 1,2},
  {"ZWHO",	fun_zwho,	0, 1},
  {0}};

void list_functions(dbref player)
{
  char buf[512];
  int a, col=get_cols(NULL, player);

  notify(player, "List of hardcoded functions in alphabetical order:");
  notify(player, "");
  strcpy(buf, "");
  for(a=0;flist[a].name;a++) {
    if(*buf && strlen(buf)+strlen(flist[a].name) >= col-2) {
      notify(player, buf);
      strcpy(buf, "");
    }
    if(*buf)
      strcat(buf, ", ");
    strcat(buf, flist[a].name);
  }
  if(*buf)
    notify(player, buf);
  notify(player, "");
}

/* search for a user-defined function */
int user_defined(str, buff, privs, doer)
char **str, *buff;
dbref privs, doer;
{
  ATTR *attr;
  char temp[8192], *saveptr[10], *writeptr[10], *s=buff;
  dbref thing=NOTHING, *ptr;
  int a, b;

  if(*s == '!')
    s++;

  /* check for function defined on a specific object.  e.g. #123:func() */
  if(*s == '#') {
    thing=atoi(s+1);
    if(Invalid(thing) || !(s=strchr(s, ':')))
      return 0;
    if(!(attr=atr_str(thing, s+1)) || !(attr->flags & AF_FUNC) ||
       !can_see_atr(privs, thing, attr))
      return 0;
  } else if((attr=atr_str(privs, s)) && (attr->flags & AF_FUNC))
    thing=privs;
  else {
    /* Scan zones for matching function */
    for(ptr=db[mainroom(privs)].zone;ptr && *ptr != NOTHING;ptr++)
      if((attr=atr_str(*ptr, s)) && (attr->flags & AF_FUNC)) {
        thing=*ptr;
        break;
      }

    /* If still nothing, try universal zones */
    if(thing == NOTHING)
      for(ptr=uzo;ptr && *ptr != NOTHING;ptr++)
        if((attr=atr_str(*ptr, s)) && (attr->flags & AF_FUNC)) {
          thing=*ptr;
          break;
        }
  }

  /* Nothing found. */
  if(thing == NOTHING)
    return 0;

  /* get the 10 arguments to the function */
  for(a=0;**str;a++) {
    if(a >= 10) {  /* Skip over remaining arguments */
      while(**str && **str != ')' && **str != ']')
        (*str)++;
      break;
    }
    if(**str == ')' || **str == ']') {
      SPTR(writeptr[a], "");
      continue;
    }
    evaluate(str, buff, privs, doer, 1);
    SPTR(writeptr[a], buff);
    if(**str == ',')
      (*str)++;
  }
  for(b=0;b<10;b++) {
    SPTR(saveptr[b], wptr[b]);
    strcpy(wptr[b], b<a?writeptr[b]:"");
    if(b<a)
      free(writeptr[b]);
  }

  if(**str)
    (*str)++;
  strcpy(temp, atr_get_obj(thing, attr->obj, attr->num));
  pronoun_substitute(buff, privs, doer, temp);

  for(b=0;b<10;b++) {
    strcpy(wptr[b], b<a?saveptr[b]:"");
    free(saveptr[b]);
  }
  return 1;
}


void func_zerolev()
{
  func_recur=0; func_total=0;
}

char *func_printlev()
{
  return tprintf("%d %d", func_total, func_recur);
}

int function(str, buff, privs, doer)
char **str, *buff;
dbref privs, doer;
{
  static hash *funchash=NULL;
  struct fun *fp;
  char **args=NULL, *s=buff;
  int a, argc, inv=0;

  /* function names cannot have spaces or symbols in them */
  if(!ok_attribute_name(buff))
    return 0;

  if(++func_recur > FUNC_RECURSION) {
    strcpy(buff, "#-1 Too many levels of recursion.");
    func_recur--;
    return 1;
  } if(++func_total > MAX_FUNCTIONS) {
    strcpy(buff, "#-1 Too many functions.");
    func_recur--;
    return 1;
  }

  /* determine the function name we wish to find */
  if(*s == '!') {
    inv=1;
    s++;
  }
  if(!funchash)
    funchash=make_hashtab(flist, sizeof(struct fun));
  fp=lookup_hash(funchash, s);

  /* Not a built-in function */
  if(!fp) {
    if(user_defined(str, buff, privs, doer)) {  // User-defined function found
      if(inv)
        sprintf(buff, "%d", !ISTRUE(buff));
      func_recur--;
      return 1;
    }
    func_recur--;
    return 0;
  }

  /* get the arguments to the function */
  for(argc=0;**str && **str != ')' && **str != ']';argc++) {
    evaluate(str, buff, privs, doer, 1);

    if(!(argc%10))
      args=realloc(args, (argc+10)*sizeof(char *));  /* resize arg table */
    if(!args) {
      perror("argument allocation");
      argc=0;
      break;
    }

    SPTR(args[argc], buff);
    if(**str == ',')
      (*str)++;

    /* prep arguments for mathematical functions */
    if(fp->flags & F_MATH) {
      /* prevent system hangup with parameters > 100 chars */
      if(strlen(args[argc]) > 100)
        args[argc][100]=0;

      /* remove exponential notation */
      if((s=strchr(args[argc], 'e')))
        *s=0;
      if((s=strchr(args[argc], 'E')))
        *s=0;
    }
  }

  if(**str)
    (*str)++;

  if(argc < fp->loarg || argc > (fp->hiarg?:fp->loarg)) {
    sprintf(buff, "#-1 Function %s requires %s", fp->name,
            fp->hiarg >= 100?"at least ":"");
    if(!fp->hiarg || fp->hiarg >= 100)
      sprintf(buff+strlen(buff), "%d argument%s.", fp->loarg, fp->loarg==1?
              "":"s");
    else if(fp->loarg == fp->hiarg-1)
      sprintf(buff+strlen(buff), "%d or %d arguments.", fp->loarg, fp->hiarg);
    else
      sprintf(buff+strlen(buff), "between %d and %d arguments.", fp->loarg,
              fp->hiarg);
  } else {
    make_quiet=1;
    fp->fun(buff, args, privs, doer, argc);
    make_quiet=0;
    if(inv)
      sprintf(buff, "%d", !ISTRUE(buff));
  }

  for(a=0;a<argc;a++)
    free(args[a]);
  free(args);
  func_recur--;
  return 1;
}

/* type=0: used in command.c to parse function() types
        1: executed above to parse argument list */
void evaluate(str, buff, privs, doer, type)
char **str, *buff;
dbref privs, doer;
int type;
{
  char *r, *s=*str, *e=buff;
  int deep, chr, func=0;

  /* Skip preceding spaces */
  while(*s && isspace(*s))
    s++;

  while(*s) {
    switch(*s) {
    case '{':
      /* If this is first character, eliminate surrounding braces */
      if(e == buff) {
        chr=1;
        s++;
      } else
        chr=0;

      deep=0;
      while(*s && (deep || *s != '}')) {
        if(*s == '{')
          deep++;
        else if(*s == '}')
          deep--;
        *e++=*s++;
      }

      if(chr && *s == '}') {
        s++;
        continue;
      }
      break;
    case '[':  /* Skip over nested function calls */
      *e++=*s++;
      deep=0;
      while(*s && (deep || *s != ']')) {
        if(*s == '[')
          deep++;
        else if(*s == ']')
          deep--;
        *e++=*s++;
      }
      break;
    case ',': case ')': case ']':  /* End of function scope */
      if(!type || func)  /* Not in function */
        break;
      *e='\0';
      *str=s;
      return;
    case '(':  /* Function Header */
      if(func)
        break;

      *e='\0';
      *str=s+1;
      r=s;
      if(function(str, buff, privs, doer) || type)
        return;
      func=1;
      s=r;
    }

    if(*s)
      *e++=*s++;
  }

  *e='\0';
  *str=s;
}
