/* prog/hash.c */
/* dynamic hash table management */

#include "externs.h"

#define HASH_SIZE 239	// Size of table array. Relatively prime.


hash *make_hashtab(void *entry, int size)
{
  hash *op, *ptr;
  int i, sizes[HASH_SIZE]={0};
  
  op=calloc(HASH_SIZE, sizeof(hash *));

  /* we assume the name is the first attribute of the array structure */
  for(ptr=entry;ptr->ptr;ptr+=size/sizeof(hash *))
    sizes[hash_name((char *)ptr->ptr)]++;

  for(i=0;i<HASH_SIZE;i++) {
    op[i].ptr=malloc(sizeof(hash *)*(sizes[i]+1));
    op[i].ptr[sizes[i]].ptr=NULL;
  }

  for(ptr=entry;ptr->ptr;ptr+=size/sizeof(hash *)) {
    i=hash_name((char *)ptr->ptr);
    op[i].ptr[--sizes[i]].ptr=ptr;
  }

  return op;
}

int hash_name(char *name)
{
  unsigned short op=0;
  int j, shift=0;
  char *s;
  
  for(s=name;*s;s++) {
    j = tolower(*s);  /* case insensitive */
    op ^= j<<(shift);
    op ^= j>>(16-shift);
    shift = (shift+11)%16;
  }

  return op % HASH_SIZE;
}

void *lookup_hash(hash *tab, char *name)
{
  int i=hash_name(name), j;

  for(j=0;tab[i].ptr[j].ptr;j++)
    if(!strcasecmp((char *)tab[i].ptr[j].ptr->ptr, name))
      return tab[i].ptr[j].ptr;
  return NULL;
}

/* Sizable Hash Table Manipulation */

/* creates an empty hash table */
hash *create_hash()
{
  hash *op;
  int i;

  op=calloc(HASH_SIZE, sizeof(hash *));

  for(i=0;i<HASH_SIZE;i++) {
    op[i].ptr=malloc(sizeof(hash *));
    op[i].ptr[0].ptr=NULL;
  }

  return op;
}

void add_hash(hash *tab, void *entry)
{
  hash *ptr=entry;
  int j, i=hash_name((char *)ptr->ptr);

  /* count size of hash pointer */
  for(j=0;tab[i].ptr[j].ptr;j++);  /* count size of hash pointer */

  tab[i].ptr=realloc(tab[i].ptr, (j+2) * sizeof(hash *));
  tab[i].ptr[j].ptr=entry;
  tab[i].ptr[j+1].ptr=NULL;
}

void del_hash(hash *tab, char *name)
{
  int j, loc=-1, i=hash_name(name);

  for(j=0;tab[i].ptr[j].ptr;j++)
    if(!strcasecmp((char *)tab[i].ptr[j].ptr->ptr, name))
      loc=j;
  if(loc == -1)
    return;

  free(tab[i].ptr[loc].ptr);

  if(j > 1)
    tab[i].ptr[loc].ptr=tab[i].ptr[j-1].ptr;
  tab[i].ptr[j-1].ptr=NULL;
  tab[i].ptr=realloc(tab[i].ptr, j * sizeof(hash *));
}
